Package: ghclass
Title: Tools for Managing Classes on GitHub
Version: 0.4.0
Authors@R: 
    c(person(given = "Colin",
             family = "Rundel",
             role = c("aut", "cre"),
             email = "rundel@gmail.com"),
      person(given = "Mine",
             family = "Cetinkaya-Rundel",
             role = "aut",
             email = "mine@rstudio.com"),
      person(given = "Therese",
             family = "Anders",
             role = "ctb",
             email = "tanders@usc.edu"))
Description: Interface for the GitHub API that enables efficient 
    management of courses on GitHub. It has a functionality for 
    managing organizations, teams, repositories, and users on GitHub 
    and helps automate most of the tedious and repetitive tasks 
    around creating and distributing assignments.
License: GPL-3
URL: https://github.com/rundel/ghclass
BugReports: https://github.com/rundel/ghclass/issues
Depends: R (>= 3.4.0)
Imports: base64enc, fs, gh, glue, httr, lubridate, purrr, rlang,
        tibble, whisker, withr, dplyr, cli (>= 3.0.0), lifecycle
Suggests: here, knitr, rmarkdown, sodium, styler, usethis, gert, readr,
        gitcreds
Encoding: UTF-8
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-05-08 01:00:45 UTC; rundel
Author: Colin Rundel [aut, cre],
  Mine Cetinkaya-Rundel [aut],
  Therese Anders [ctb]
Maintainer: Colin Rundel <rundel@gmail.com>
Repository: CRAN
Date/Publication: 2026-05-08 02:11:15 UTC
Built: R 4.6.0; ; 2026-05-08 06:41:20 UTC; unix

Package: ghclass
Title: Tools for Managing Classes on GitHub
Version: 0.4.1
Authors@R: 
    c(person(given = "Colin",
             family = "Rundel",
             role = c("aut", "cre"),
             email = "rundel@gmail.com"),
      person(given = "Mine",
             family = "Cetinkaya-Rundel",
             role = "aut",
             email = "mine@rstudio.com"),
      person(given = "Therese",
             family = "Anders",
             role = "ctb",
             email = "tanders@usc.edu"))
Description: Interface for the GitHub API that enables efficient 
    management of courses on GitHub. It has a functionality for 
    managing organizations, teams, repositories, and users on GitHub 
    and helps automate most of the tedious and repetitive tasks 
    around creating and distributing assignments.
License: GPL-3
URL: https://github.com/rundel/ghclass,
        https://rundel.github.io/ghclass/
BugReports: https://github.com/rundel/ghclass/issues
Depends: R (>= 3.4.0)
Imports: base64enc, fs, gh (>= 1.4.0), glue, httr, lubridate, purrr,
        rlang, stringr, tibble, whisker, withr, dplyr, cli (>= 3.0.0),
        lifecycle, readr, tidyselect
Suggests: here, knitr, rmarkdown, sodium, styler, testthat (>= 3.0.0),
        usethis, gert, gitcreds
Encoding: UTF-8
Config/roxygen2/version: 8.0.0
Config/testthat/edition: 3
NeedsCompilation: no
Packaged: 2026-07-25 23:08:29 UTC; rundel
Author: Colin Rundel [aut, cre],
  Mine Cetinkaya-Rundel [aut],
  Therese Anders [ctb]
Maintainer: Colin Rundel <rundel@gmail.com>
Repository: CRAN
Date/Publication: 2026-07-25 23:20:02 UTC

# Generated by roxygen2: do not edit by hand

S3method(filter_results,data.frame)
S3method(filter_results,default)
export(action_add_badge)
export(action_artifact_delete)
export(action_artifact_download)
export(action_artifacts)
export(action_remove_badge)
export(action_run_logs)
export(action_runs)
export(action_runtime)
export(action_status)
export(action_workflows)
export(arg_is_chr)
export(arg_is_chr_scalar)
export(arg_is_lgl)
export(arg_is_lgl_scalar)
export(arg_is_pos_int)
export(arg_is_pos_int_scalar)
export(branch_create)
export(branch_delete)
export(branch_remove)
export(cli_glue)
export(cli_stop)
export(cli_warn)
export(error)
export(failed)
export(format_repo)
export(get_repo_name)
export(get_repo_owner)
export(ghclass_api_v3_req)
export(github_api_repo_get_file)
export(github_api_repo_tree)
export(github_get_api_limit)
export(github_get_max_rate)
export(github_get_max_wait)
export(github_get_token)
export(github_graphql_rate_limit)
export(github_orgs)
export(github_rate_limit)
export(github_reset_token)
export(github_set_api_limit)
export(github_set_max_rate)
export(github_set_max_wait)
export(github_set_token)
export(github_test_token)
export(github_token_scopes)
export(github_whoami)
export(issue_close)
export(issue_create)
export(issue_edit)
export(local_pat)
export(local_repo_add)
export(local_repo_anonymize)
export(local_repo_branch)
export(local_repo_clone)
export(local_repo_commit)
export(local_repo_log)
export(local_repo_pull)
export(local_repo_push)
export(local_repo_rename)
export(match_repo)
export(org_admins)
export(org_allows_forking)
export(org_create_assignment)
export(org_exists)
export(org_grade_assignment)
export(org_invite)
export(org_members)
export(org_pending)
export(org_remove)
export(org_repo_forking)
export(org_repo_search)
export(org_repo_stats)
export(org_repos)
export(org_set_repo_permission)
export(org_set_workflow_permissions)
export(org_sitrep)
export(org_team_details)
export(org_teams)
export(org_user_repos)
export(org_workflow_permissions)
export(pages_create)
export(pages_delete)
export(pages_enabled)
export(pages_status)
export(pr_create)
export(read_bin_file)
export(repo_add_file)
export(repo_add_team)
export(repo_add_user)
export(repo_allows_forking)
export(repo_branches)
export(repo_clone)
export(repo_clone_url)
export(repo_collaborators)
export(repo_commits)
export(repo_contributors)
export(repo_create)
export(repo_delete)
export(repo_delete_file)
export(repo_exists)
export(repo_files)
export(repo_get_file)
export(repo_get_readme)
export(repo_ignore)
export(repo_is_template)
export(repo_issues)
export(repo_ls)
export(repo_mirror)
export(repo_mirror_template)
export(repo_modify_file)
export(repo_n_commits)
export(repo_prs)
export(repo_pushes)
export(repo_put_file)
export(repo_remove_team)
export(repo_remove_user)
export(repo_rename)
export(repo_set_forking)
export(repo_set_template)
export(repo_style)
export(repo_team_permission)
export(repo_unwatch)
export(repo_user_permission)
export(repo_watch)
export(repo_watching)
export(result)
export(status_msg)
export(succeeded)
export(team_create)
export(team_delete)
export(team_invite)
export(team_members)
export(team_pending)
export(team_remove)
export(team_rename)
export(team_repos)
export(team_roster)
export(user_exists)
export(user_repos)
export(user_type)
export(with_pat)
importFrom(dplyr,"%>%")
importFrom(rlang,".data")

# ghclass <img src='man/figures/logo.png' align="right" height="140" />

<!-- badges: start -->
[![R-CMD-check](https://github.com/rundel/ghclass/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/rundel/ghclass/actions/workflows/R-CMD-check.yaml)
[![CRAN status](https://www.r-pkg.org/badges/version/ghclass)](https://CRAN.R-project.org/package=ghclass)
<!-- badges: end -->

## Tools for managing GitHub class organization accounts

This R package is designed to enable instructors to efficiently manage
their courses on GitHub. It has a wide range of functionality for
managing organizations, teams, repositories, and users on GitHub and
helps automate most of the tedious and repetitive tasks around creating
and distributing assignments.

Install ghclass from CRAN:

``` r
install.packages("ghclass")
```

Install the development version package from GitHub:

``` r
# install.packages("remotes")
remotes::install_github("rundel/ghclass")
```

See package
[vignette](https://rundel.github.io/ghclass/articles/ghclass.html)
for details on how to use the package.

## Peer Review

The peer review functionality currently lives on the `peer_review` branch
and is not part of the CRAN release. If you need it you can install that
branch using:

``` r
remotes::install_github("rundel/ghclass@peer_review")
```

## GitHub & default branches

GitHub now uses `main` as the default branch for new repositories (see
[here](https://github.com/github/renaming) for background). `ghclass`
supports alternative default branch names across the entire package, so
for the vast majority of use cases you will not need to do anything
differently. See the FAQ in the Getting Started vignette for more details.

github_api_team_invite = function(org, team_slug, username, role = c("member", "maintainer")) {
  role = match.arg(role)

  ghclass_api_v3_req(
    endpoint = "PUT /orgs/:org/teams/:team_slug/memberships/:username",
    org = org,
    team_slug = team_slug,
    username = username,
    role = role
  )
}

#' @rdname team_members
#' @export
team_invite = function(org, user, team, team_type = c("name", "slug")) {
  arg_is_chr_scalar(org)
  arg_is_chr(user, team)
  team_type = match.arg(team_type)


  if (team_type == "name")
    slug = team_slug_lookup(org, team)
  else
    slug = team

  check_team_slug(slug)

  res = purrr::pmap(
    unique(tibble::tibble(user, team, slug)),
    function(user, team, slug) {
      if (is.na(slug)) {
        cli::cli_alert_danger("Team {.val {team}} does not exist in org {.val {org}}.")
        return(NULL)
      }

      res = purrr::safely(github_api_team_invite)(org, slug, user)

      status_msg(
        res,
        "Added user {.val {user}} to team {.val {team}}.",
        "Failed to add user {.val {user}} to team {.val {team}}."
      )

      result(res)
    }
  )

  invisible(res)
}

github_api_team_members = function(org, team_slug, role = c("all", "member", "maintainer")) {
  role = match.arg(role)

  ghclass_api_v3_req(
    endpoint = "GET /orgs/:org/teams/:team_slug/members",
    org = org,
    team_slug = team_slug,
    role = role
  )
}

# TODO - good candidate for v4 upgrade

#' @rdname team_members
#' @export
#'
team_members = function(org, team = org_teams(org), team_type = c("name", "slug")) {
  arg_is_chr_scalar(org)
  arg_is_chr(team)
  team_type = match.arg(team_type)

  slug = if (team_type == "name") team_slug_lookup(org, team) else team

  check_team_slug(slug)

  purrr::map2_dfr(
    team, slug,
    function(team, slug) {

      if (is.na(slug)) {
        res = NULL
      } else {
        res = purrr::safely(github_api_team_members)(org, slug)

        status_msg(
          res,
          fail = "Failed to retrieve team members for {.val {team}}."
        )
      }

      members = if (failed(res) | empty_result(res))
        character()
      else
        purrr::map_chr(result(res), "login")

      tibble::tibble(
        team = team,
        slug = slug,
        user = members
      )
    }
  )
}

github_api_download_artifact = function(repo, id, dest) {
  id = as.double(id)
  arg_is_chr_scalar(repo)
  #arg_is_pos_int(id)

  archive_format = "zip"
  if (missing(dest))
    dest =  tempfile(fileext = paste0(".", archive_format))

  owner = get_repo_owner(repo)
  repo = get_repo_name(repo)

  url = glue::glue("https://api.github.com/repos/{owner}/{repo}/actions/artifacts/{id}/{archive_format}")

  github_api_download_file(url, dest)
}


#' @name action
#' @rdname action
#'
#' @param keep_zip Logical. Should the downloaded zip files be retained (`TRUE`)
#'   or deleted after extraction (`FALSE`). Default `FALSE`.
#' @param nest Logical. If `TRUE`, place each repo's artifacts under a
#'   `dir/{repo_name}/` subdirectory; otherwise flatten into `dir/` with
#'   repo-prefixed names. Default `FALSE`.
#' @param overwrite Logical. Should existing zip files or extraction folders
#'   be overwritten. Default `FALSE`.
#'
#' @export
#'
action_artifact_download = function(
  repo, dir,
  filter = NULL, exclude = FALSE,
  filter_branch = NULL, exclude_branch = FALSE,
  keep_zip = FALSE, nest = FALSE, overwrite = FALSE,
  ids = action_artifacts(
    repo,
    filter = filter, exclude = exclude,
    filter_branch = filter_branch, exclude_branch = exclude_branch
  )
) {
  arg_is_chr(repo)
  arg_is_chr_scalar(dir)
  arg_is_lgl_scalar(keep_zip, nest, overwrite, exclude_branch, exclude)
  arg_is_chr_scalar(filter_branch, allow_null = TRUE)
  arg_is_chr_scalar(filter, allow_null = TRUE)

  dir.create(dir, showWarnings = FALSE, recursive = TRUE)

  if (is.numeric(ids))
    ids = tibble::tibble(repo = repo, id = ids, name = as.character(ids))
  arg_is_df(ids, allow_empty = TRUE)

  if (nrow(ids) == 0) {
    cli::cli_alert_danger("No artifacts available for the given repos.")
    return(invisible(character()))
  }

  df = dplyr::left_join(
    tibble::tibble(repo = repo),
    ids,
    by = "repo"
  ) %>%
    dplyr::select("repo", "id", "name")

  repo_groups = split(df, df[["repo"]])

  res = purrr::map(
    repo_groups,
    function(repo_df) {
      cur_repo = repo_df[["repo"]][1]
      cur_ids = repo_df[["id"]]
      cur_names = repo_df[["name"]]

      if (all(is.na(cur_ids))) {
        cli::cli_alert_danger(
          "No artifacts found for repo {.val {cur_repo}}.",
          wrap = FALSE
        )
        return(NA_character_)
      }

      not_na = !is.na(cur_ids)
      cur_ids = cur_ids[not_na]
      cur_names = cur_names[not_na]

      repo_name = get_repo_name(cur_repo)

      purrr::map2_chr(
        cur_ids, cur_names,
        function(id, name) {
          if (nest) {
            repo_dir = fs::path_norm(fs::path(dir, repo_name))
            zip_path = fs::path_norm(fs::path(repo_dir, name, ext = "zip"))
            extract_dir = fs::path_norm(fs::path(repo_dir, name))
          } else {
            base = paste0(repo_name, "_", name)
            zip_path = fs::path_norm(fs::path(dir, base, ext = "zip"))
            extract_dir = fs::path_norm(fs::path(dir, base))
          }

          if ((file.exists(zip_path) || dir.exists(extract_dir)) && !overwrite) {
            cli::cli_alert_danger(
              paste0(
                "Destination {.file {extract_dir}} or {.file {zip_path}} already exists, ",
                "set {.code overwrite = TRUE} to overwrite."
              ),
              wrap = FALSE
            )
            return(NA_character_)
          }

          if (overwrite) {
            if (dir.exists(extract_dir))
              unlink(extract_dir, recursive = TRUE, force = TRUE)
            if (file.exists(zip_path))
              file.remove(zip_path)
          }

          if (nest)
            dir.create(fs::path_dir(zip_path), showWarnings = FALSE, recursive = TRUE)

          dl = purrr::safely(github_api_download_artifact)(cur_repo, id, dest = zip_path)
          if (failed(dl)) {
            cli::cli_alert_danger(
              "Failed to download artifact with id {.val {id}} from repo {.val {cur_repo}}.",
              wrap = FALSE
            )
            return(NA_character_)
          }

          dir.create(extract_dir, showWarnings = FALSE, recursive = TRUE)
          unzipped = purrr::safely(
            function() withCallingHandlers(
              utils::unzip(zip_path, exdir = extract_dir),
              warning = function(w) cli_stop("unzip reported a problem extracting artifact {.val {id}}.")
            )
          )()
          if (failed(unzipped)) {
            cli::cli_alert_danger(
              "Failed to extract artifact with id {.val {id}} from repo {.val {cur_repo}} to {.file {extract_dir}}.",
              wrap = FALSE
            )
            unlink(extract_dir, recursive = TRUE, force = TRUE)
            if (!keep_zip && file.exists(zip_path))
              file.remove(zip_path)
            return(NA_character_)
          }

          if (!keep_zip)
            file.remove(zip_path)

          cli::cli_alert_success(
            "Downloaded artifact {.val {id}} from repo {.val {cur_repo}} to {.file {extract_dir}}.",
            wrap = FALSE
          )

          extract_dir
        }
      )
    }
  )

  invisible(unlist(res))
}

github_api_repo_remove_user = function(repo, username){
  ghclass_api_v3_req(
    endpoint = "DELETE /repos/:owner/:repo/collaborators/:username",
    owner = get_repo_owner(repo),
    repo = get_repo_name(repo),
    username = username
  )
}


#' @rdname repo_user
#' @export
#'
repo_remove_user = function(repo, user) {
  arg_is_chr(repo, user)

  res = purrr::map2(
    repo, user,
    function(repo, user) {
      res = purrr::safely(github_api_repo_remove_user)(
        repo = repo,
        username = user
      )

      status_msg(
        res,
        "Removed user {.val {user}} from repo {.val {repo}}.",
        "Failed to remove user {.val {user}} from repo {.val {repo}}."
      )
    }
  )

  invisible(res)
}

require_pkg = function(pkg, call = rlang::caller_call()) {
  if (requireNamespace(pkg, quietly = TRUE))
    return(invisible())

  fn = rlang::call_name(call)
  if (is.null(fn))
    cli_stop("This function requires the package {.pkg {pkg}} is installed.")
  else
    cli_stop("Function {.fun {fn}} requires the package {.pkg {pkg}} is installed.")
}

require_gert = function(call = rlang::caller_call()) {
  require_pkg("gert", call = call)
}

require_styler = function(call = rlang::caller_call()) {
  require_pkg("styler", call = call)
}

#' @rdname repo_core
#'
#' @param source_repo Character. Address of repository in "owner/name" format.
#' @param target_repo Character. One or more repository addresses in "owner/name" format.
#' @param overwrite Logical. Should the target repositories be overwritten. Default `FALSE`.
#' @param verbose Logical. Display verbose output. Default `FALSE`.
#' @param warn Logical. Warn the user about the function being deprecated. Default `TRUE`.
#'
#' @export
#'

repo_mirror = function(source_repo, target_repo, overwrite=FALSE, verbose=FALSE, warn=TRUE) {
  arg_is_chr_scalar(source_repo)
  arg_is_chr(target_repo)
  arg_is_lgl_scalar(overwrite, verbose)

  target_repo = unique(target_repo)

  if (warn)
    .Deprecated("repo_mirror_template", package = "ghclass")


  withr::local_dir(tempdir())

  tmpdir = getwd()

  dir = file.path(tmpdir, get_repo_name(source_repo))
  unlink(dir, recursive = TRUE) # Make sure the source repo local folder does not exist

  repos = repo_n_commits(target_repo, quiet = TRUE) %>%
    dplyr::select(.data$repo, .data$n)

  local_repo_clone(source_repo, tmpdir, mirror = TRUE, verbose = verbose)

  warned = FALSE

  res = purrr::pmap(
    repos,
    function(repo, n) {
      repo_url = cli_glue("https://github.com/{repo}.git")

      if (is.na(n)) {
        cli::cli_alert_danger("The repo {.val {repo}} does not exist")
      } else if (n > 1 & !overwrite) {
        msg = paste(
          "The repo {.val {repo}} has more than one commit",
          "(n_commit = {.val {n}})."
        )

        if (!warned) {
          msg = c(msg, paste(
            "Use {.code overwrite = TRUE} if you want to permanently",
            "overwrite this repository."
          ))
          warned <<- TRUE
        }

        cli::cli_alert_danger( msg )
      } else {
        local_repo_push(dir, remote = repo_url, force = TRUE, prompt = FALSE, mirror = TRUE, verbose = verbose)
      }
    }
  )

  unlink(dir, recursive = TRUE)
  cli::cli_alert_success("Removed local copy of {.val {source_repo}}")

  invisible(res)
}

github_api_repo_issues = function(
  repo,
  state	= c("open", "closed", "all"),
  assignee = NULL,
  creator	= NULL,
  mentioned	= NULL,
  labels = NULL,
  sort = c("created", "updated", "comments"),
  direction =	c("desc", "asc"),
  since	= NULL
) {
  state = match.arg(state)
  sort = match.arg(sort)
  direction = match.arg(direction)

  ghclass_api_v3_req(
    endpoint = "GET /repos/:owner/:repo/issues",
    owner = get_repo_owner(repo),
    repo = get_repo_name(repo),
    state = state,
    assignee = assignee,
    creator = creator,
    mentioned = mentioned,
    labels = labels,
    sort = sort,
    direction = direction,
    since = since
  )
}

#' @rdname repo_details
#'
#' @param state	    Character. State of the issues to return. Can be either "open", "closed", or "all".
#' @param assignee	Character. Return issues assigned to a particular username.
#'   Pass in "none" for issues with no assigned user, and "*" for issues assigned to any user.
#' @param creator	  Character. Return issues created the by the given username.
#' @param mentioned Character. Return issues that mentioned the given username.
#' @param labels	  Character. Return issues labeled with one or more of of the given label names.
#' @param sort	    Character. What to sort results by. Can be either "created", "updated", or "comments".
#' @param direction Character. The direction of the sort. Can be either "asc" or "desc".
#' @param since	    Character. Only issues updated at or after this time are returned.
#'
#' @export
#'
repo_issues = function(
  repo,
  state	= c("open", "closed", "all"),
  assignee = NULL,
  creator	= NULL,
  mentioned	= NULL,
  labels = NULL,
  sort = c("created", "updated", "comments"),
  direction =	c("desc", "asc"),
  since	= NULL
) {
  arg_is_chr(repo)
  arg_is_chr(labels, allow_null=TRUE)
  arg_is_chr_scalar(assignee, creator, mentioned, since, allow_null=TRUE)

  state = match.arg(state)
  sort = match.arg(sort)
  direction = match.arg(direction)

  purrr::map_dfr(
    repo,
    function(repo) {
      res = purrr::safely(github_api_repo_issues)(
        repo = repo,
        state = state, assignee = assignee, creator = creator,
        mentioned = mentioned, labels = labels, sort = sort,
        direction = direction, since = since
      )

      status_msg(
        res,
        "Retrieved issues for repo {.val {repo}}.",
        "Failed to retrieve issues for repo {.val {repo}}."
      )

      if (succeeded(res)) {
        iss = result(res)

        tibble::tibble(
          repo = repo,
          number = purrr::map_int(iss, "number"),
          title = purrr::map_chr(iss, "title"),
          state = purrr::map_chr(iss, "state"),
          comments = purrr::map_int(iss, "comments"),
          created = purrr::map_chr(iss, "created_at") %>% lubridate::ymd_hms(),
          updated = purrr::map_chr(iss, "updated_at") %>% lubridate::ymd_hms(),
          closed = purrr::map_chr(iss, "closed_at", .default = NA) %>% lubridate::ymd_hms(),
          created_by = purrr::map_chr(iss, c("user","login")),
          assignees = purrr::map(iss, ~ purrr::map_chr(.x$assignees, "login"))
        )
      } else {
        tibble::tibble(
          repo = character(),
          number = integer(),
          title = character(),
          state = character(),
          comments = integer(),
          created = as.POSIXct(character()),
          updated = as.POSIXct(character()),
          closed = as.POSIXct(character()),
          created_by = character(),
          assignees = list()
        )
      }
    }
  )
}

replace_nas = function(cur, rep) {
  cur[is.na(cur)] = rep[is.na(cur)]
  cur
}

warn_experimental = function(call = rlang::caller_call()) {
  fn = rlang::call_name(call)
  id = if (is.null(fn)) "unknown" else fn

  cli::cli_warn(
    c( "i" = paste0(
      "The function {.fun {id}} is currently ",
      "considered experimental. Its interface, implementation, and other ",
      "features may change significantly in future versions of the package. ",
      "Use with caution."
    ) ),
    .frequency = "once", .frequency_id = paste0("warn_experimental_", id)
  )
}

url_encode = function(urls) {
  purrr::map_chr(urls, utils::URLencode)
}

github_api_org_repos = function(org, type = NULL, sort = NULL, direction = NULL) {
  arg_is_chr_scalar(org)
  arg_is_chr_scalar(type, sort, direction, allow_null = TRUE)

  ghclass_api_v3_req(
    endpoint = "GET /orgs/:org/repos",
    org = org,
    type = type,
    sort = sort,
    direction = direction
  )
}

#' @rdname org_details
#' @param full_repo Logical. Should the full repository address be returned (e.g. `owner/repo` instead of just `repo`). Default `TRUE`.
#' @param sort Character. Sorting criteria to use, can be one of "created", "updated", "pushed", or "full_name".
#' @param direction Character. Sorting order to use.
#' @param type  Character. Specifies the type of repositories you want, can be one of
#' "all", "public", "private", "forks", "sources", "member", or "internal".
#'
#' @export
#'
org_repos = function(
  org, filter = NULL, exclude = FALSE, full_repo = TRUE,
  sort = c("full_name", "created", "updated", "pushed"),
  direction = c("asc", "desc"),
  type = c("all", "public", "private", "forks", "sources", "member", "internal")
) {
  sort = match.arg(sort)
  direction = match.arg(direction)
  type = match.arg(type)

  arg_is_chr_scalar(org)
  arg_is_chr_scalar(filter, allow_null = TRUE)
  arg_is_lgl_scalar(exclude, full_repo)
  arg_is_chr_scalar(sort, direction, type)

  res = purrr::safely(
    function() {
      user = purrr::safely(github_api_get_user)(org)

      if (failed(user)) {
        if (inherits(error(user), "http_error_404"))
          cli_stop("Organization {.val {org}} does not exist on GitHub.")
        else
          cli_stop("Unable to determine the account type for {.val {org}}; the request to GitHub failed.")
      }

      org_type = result(user)[["type"]]

      if (org_type == "Organization") {
        github_api_org_repos(org, type = type, sort = sort, direction = direction)
      } else if (org_type == "User") {
        cli_stop("{.val {org}} is a user not an organization. Use {.fun user_repos} instead.")
      } else {
        cli_stop("{.val {org}} has unknown type {.val {org_type}}.")
      }
    }
  )()

  status_msg(
    res,
    fail = "Failed to retrieve repos for org {.val {org}}."
  )

  if (failed(res) | empty_result(res))
    return(invisible(NULL))

  if (full_repo) {
    res = purrr::map_chr(result(res), "full_name")
  } else {
    res = purrr::map_chr(result(res), "name")
  }

  filter_results(res, filter, exclude)
}

#github_repo_pattern ="^([A-Za-z0-9]+[A-Za-z0-9-]*[A-Za-z0-9]+)/([A-Za-z0-9_.-]+)$"
#github_username_pattern = "^[A-Za-z\\d](?:[A-Za-z\\d]|-(?=[A-Za-z\\d])){0,38}$"

# Use a simplified pattern and let GitHub sort out the particulars
github_repo_pattern ="^([A-Za-z0-9-]+)/([A-Za-z0-9_.-]+)$"


valid_repo_error = function(repo) {
  cli_stop( paste(
    "Invalid repository name(s) {.val {repo}}.",
    "Repository names must be in {.val 'owner/name'} format."
  ) )
}


#' @rdname ghclass-internal
#' @export
match_repo = function(repo, index=1) {
  arg_is_chr(repo)

  m = regexec(github_repo_pattern, repo)
  m = regmatches(repo, m)

  l = purrr::map_int(m, length)

  if (any(l != 3))
    valid_repo_error(repo[l != 3])

  purrr::map_chr(m, index)
}


#' @rdname ghclass-internal
#' @export
get_repo_name = function(repo) {
  match_repo(repo, 3)
}

#' @rdname ghclass-internal
#' @export
get_repo_owner = function(repo) {
  match_repo(repo, 2)
}

#' @rdname ghclass-internal
#' @export
format_repo = function(repo, branch = NULL, path = NULL) {
  if (!is.null(branch)) {
    repo = paste(repo, branch, sep="@")
  }

  if (!is.null(path))
    repo = file.path(repo, path)

  repo
}

github_api_repo_unwatch = function(repo){
  ghclass_api_v3_req(
    endpoint = "DELETE /repos/:owner/:repo/subscription",
    owner = get_repo_owner(repo),
    repo = get_repo_name(repo)
  )

}

#' @rdname repo_notification
#' @export
#'
repo_unwatch = function(repo) {
  arg_is_chr(repo)

  res = purrr::map(
    repo,
    function(repo) {
      res = purrr::safely(github_api_repo_unwatch)(repo)

      status_msg(
        res,
        "Unwatched repo {.val {repo}}.",
        "Failed to unwatch repo {.val {repo}}."
      )
    }
  )

  invisible(res)
}

github_api_repo_subscribe = function(repo, subscribed, ignored){
  arg_is_chr_scalar(repo)
  arg_is_lgl_scalar(subscribed, ignored)

  if (subscribed == ignored)
    cli_stop("{.code subscribed != ignored} must be true")

  ghclass_api_v3_req(
    endpoint = "PUT /repos/:owner/:repo/subscription",
    owner = get_repo_owner(repo),
    repo = get_repo_name(repo),
    subscribed = subscribed,
    ignored = ignored
  )
}

#' @rdname repo_notification
#' @export
#'
repo_watch = function(repo) {
  arg_is_chr(repo)

  res = purrr::map(
    repo,
    function(repo, notifications) {
      res = purrr::safely(github_api_repo_subscribe)(
        repo,
        subscribed = TRUE,
        ignored = FALSE
      )

      status_msg(
        res,
        "Watched {.val {repo}}.",
        "Failed to watch {.val {repo}}."
      )
    }
  )

  invisible(res)
}

#' @rdname repo_notification
#' @export
#'
repo_ignore = function(repo) {
  arg_is_chr(repo)

  res = purrr::map(
    repo,
    function(repo, notifications) {
      res = purrr::safely(github_api_repo_subscribe)(
        repo,
        subscribed = FALSE,
        ignored = TRUE
      )

      status_msg(
        res,
        "Ignored repo {.val {repo}}.",
        "Failed to ignore repo {.val {repo}}."
      )
    }
  )

  invisible(res)
}

github_api_org_invite = function(org, user) {
  arg_is_chr_scalar(org, user)

  ghclass_api_v3_req(
    endpoint = "PUT /orgs/:org/memberships/:username",
    org = org,
    username = user,
    role = "member"
  )
}


#' @rdname org_members
#' @export
#'
org_invite = function(org, user) {
  arg_is_chr_scalar(org)
  arg_is_chr(user)

  user = unique(tolower(user))
  member = tolower(org_members(org))
  pending = tolower(org_pending(org))

  res = purrr::map(
    user,
    function(user) {
      if (user %in% member) {
        cli::cli_alert_info("User {.val {user}} is already a member of org {.val {org}}.")
      } else if (user %in% pending) {
        cli::cli_alert_info("User {.val {user}} is already a pending member of org {.val {org}}.")
      } else {
        res = purrr::safely(github_api_org_invite)(org, user)

        status_msg(
          res,
          "Invited user {.val {user}} to org {.val {org}}.",
          "Failed to invite user {.val {user}} to org {.val {org}}."
        )
      }
    }
  )

  invisible(res)
}

#' @name github_with_pat
#' @rdname github_with_pat
#'
#' @title `withr`-like functions for temporary personal access token
#'
#' @description
#' Temporarily change the `GITHUB_PAT` environmental variable for
#' GitHub authentication. Based on the `withr` interface.
#'
#' @param new Temporary GitHub access token
#' @param code Code to execute with the temporary token
#' @param .local_envir The environment to use for scoping.
#'
#' @details if `new = NA` is used the `GITHUB_PAT` environment variable will be unset.
#'
#' @return The results of the evaluation of the code argument.
#'
#' @examples
#' \dontrun{
#' with_pat("1234", print(github_get_token()))
#' }
#'
NULL

#' @rdname github_with_pat
#' @export
#'
with_pat = function(new, code) {
  arg_is_chr_scalar(new)
  names(new) = "GITHUB_PAT"

  withr::with_envvar(
    new = new,
    code = code,
    action = "replace"
  )
}

#' @rdname github_with_pat
#' @export
#'
local_pat = function(new, .local_envir = parent.frame()) {
  arg_is_chr_scalar(new)
  names(new) = "GITHUB_PAT"

  withr::local_envvar(
    new = new,
    action = "replace",
    .local_envir = .local_envir
  )
}

#' @rdname local_repo
#' @export
#'
local_repo_pull = function(repo_dir, verbose = FALSE) {
  # TODO - add support for remotes when added to gert
  require_gert()

  arg_is_chr(repo_dir)
  arg_is_lgl_scalar(verbose)

  dir = repo_dir_helper(repo_dir)

  res = purrr::map(
    dir,
    function(dir) {
      res = purrr::safely(gert::git_pull)(
        repo = dir
      )

      repo = fs::path_file(dir)
      status_msg(
        res,
        "Pulled to local repo {.val {repo}}.",
        "Failed to pull to local repo {.val {repo}}."
      )

      res
    }
  )

  invisible(res)
}

github_api_org_pending = function(owner){
  arg_is_chr_scalar(owner)
  ghclass_api_v3_req(
    endpoint = "GET /orgs/:owner/invitations",
    owner = owner
  )
}

#' @rdname org_members
#' @export
#'
org_pending = function(org, filter = NULL, exclude = FALSE) {
  arg_is_chr_scalar(org)
  arg_is_chr_scalar(filter, allow_null = TRUE)
  arg_is_lgl_scalar(exclude)

  res = purrr::safely(github_api_org_pending)(org)
  status_msg(
    res,
    fail = "Failed to retrieve pending members for org {.val {org}}"
  )

  invite = purrr::map(result(res), "login")
  invite = purrr::flatten_chr(invite)
  filter_results(invite, filter, exclude)
}

github_api_repo_edit = function(repo, ...) {
  arg_is_chr_scalar(repo)

  ghclass_api_v3_req(
    endpoint = "PATCH /repos/:owner/:repo",
    owner = get_repo_owner(repo),
    repo = get_repo_name(repo),
    ...,
    # Needed for template repos
    .send_headers = c(Accept = "application/vnd.github.baptiste-preview+json")
  )
}

#' @rdname repo_core
#' @param status Logical. Should the repository be set as a template repository. Default `TRUE`.
#' @export
#'
repo_set_template = function(repo, status = TRUE) {

  arg_is_chr(repo)
  arg_is_lgl(status)

  # Checking if repo exists
  res = purrr::map2(
    repo, status,
    function(repo, status) {
      res = purrr::safely(github_api_repo_edit)(repo, is_template = status)
      status_msg(
        res,
        "Changed the template status of repo {.val {repo}} to {.val {status}}.",
        "Failed to change template status of repo {.val {repo}}."
      )
    }
  )

  invisible(res)
}

github_api_repo_contributors = function(repo, max_retries = 5, retry_delay = 2) {
  owner = get_repo_owner(repo)
  name  = get_repo_name(repo)

  url = paste0("https://api.github.com/repos/", owner, "/", name, "/stats/contributors")

  for (i in seq_len(max_retries + 1)) {
    req = httr::GET(
      url,
      httr::add_headers(
        Authorization = paste("bearer", github_get_token()),
        "X-GitHub-Api-Version" = "2022-11-28",
        Accept = "application/vnd.github+json"
      )
    )

    code = httr::status_code(req)

    if (code == 202) {
      if (i <= max_retries) {
        Sys.sleep(retry_delay)
        next
      }
      cli::cli_warn(
        "GitHub is still computing contributor stats for {.val {repo}} after {max_retries} retries; try again shortly."
      )
      return(list())
    }

    if (code >= 300) {
      content = httr::content(req)
      cli_stop("GitHub API Error ({code}) - {content$message}")
    }

    return(httr::content(req))
  }
}

#' @rdname repo_user
#' @export
#'
repo_contributors = function(repo) {

  arg_is_chr(repo)
  repo = unique(repo)

  purrr::map_dfr(
    repo,
    function(repo) {
      res = purrr::safely(github_api_repo_contributors)(repo)

      status_msg(
        res,
        fail = "Failed to retrieve contributors for {.val {repo}}."
      )

      contribs = result(res)

      if (empty_result(contribs)) {
        d = tibble::tibble(
          repo = character(),
          username = character(),
          commits = integer(),
          weekly_stats = list()
        )
      } else {
        d = tibble::tibble(
          repo = repo,
          username = purrr::map_chr(contribs, c("author", "login")),
          commits = purrr::map_int(contribs, "total"),
          weekly_stats = purrr::map(
            contribs,
            function(cont) {
              dplyr::bind_rows(cont$weeks) %>%
              dplyr::transmute(
                week = lubridate::as_datetime(.data$w),
                additions = .data$a,
                deletions = .data$d,
                commits = .data$c
              )
            }
          )
        )
      }
    }
  )
}

team_slug_lookup = function(org, name) {
  dplyr::left_join(
    tibble::tibble(name = name),
    org_team_details_v4(org),
    by = "name"
  )[["slug"]]
}

is_slug = function(name) {
  # a slug consists of solely lowercase alphanumeric characters, dashes and underscores,
  # without 2 or more dashes in a row (sequences of underscores are allowed).
  # Furthermore, a slug cannot start or end with a hyphen.
  !grepl("[^a-z0-9_-]+|-{2,}|^-|-$", name)
}

check_team_slug = function(name, quiet = FALSE) {
  r = is_slug(name)

  if(!quiet & sum(r) != length(r))
    cli::cli_alert_warning(
      "The following team names do not appear to be valid slugs: {.val {name[!r]}}"
    )

  all(r)
}

#' @name issue
#' @rdname issue
#'
#' @title GitHub Issue related tools
#'
#' @description
#' * `issue_create` creates a new issue.
#'
#' * `issue_close` closes an existing issue.
#'
#' * `issue_edit` edits the properties of an existing issue.
#'
#' @param repo Character. Address of one or more repositories in `owner/name` format.
#' @param title Character. Title of the issue.
#' @param body Character. Content of the issue.
#' @param labels Character. Vector of the labels to associate with this issue
#' @param assignees Character. Vector of logins for users assigned to the issue.
#' @param number Integer. GitHub issue number.
#'
#' @return All functions invisibly return a list containing the results of the relevant GitHub API call.
#'
#' @seealso [repo_issues]
#'
#' @examples
#' \dontrun{
#' repo_create("ghclass-test","test_issue")
#'
#' issue_create(
#'   "ghclass-test/test_issue",
#'   title = "Issue 1",
#'   body = "This is an issue"
#'  )
#'
#' issue_create(
#'   "ghclass-test/test_issue",
#'   title = "Issue 2", body = "This is also issue",
#'   label = "Important"
#' )
#'
#' issue_create(
#'   "ghclass-test/test_issue",
#'   title = "Issue 3", body = "This is also issue",
#'   label = c("Important", "Super Important"),
#'   assignees = "rundel"
#' )
#'
#' issue_close("ghclass-test/test_issue", 1)
#'
#' issue_edit(
#'   "ghclass-test/test_issue", 2,
#'   title = "New issue 2 title!",
#'   body = "Replacement body text"
#' )
#'
#' ghclass::repo_issues("ghclass-test/test_issue")
#'
#' repo_delete("ghclass-test/test_issue", prompt=FALSE)
#' }


NULL

#' @name team
#' @rdname team
#'
#' @title Create, delete, and rename teams within an organization
#'
#' @description
#'
#' * `team_create()` - create teams in a GitHub organization
#'
#' * `team_delete()` - delete a team from a GitHub organization.
#'
#' * `team_rename()` - rename an existing team
#'
#' @param org Character. Name of the GitHub organization.
#' @param team Character. Name of teams.
#' @param team_type Character. Either "slug" if the team names are slugs or "name" if full team names are provided.
#' @param prefix Character. Shared prefix.
#' @param suffix Character. Shared suffix.
#' @param privacy Character. Level of privacy for team, "closed" (visible to all
#' members of the organization) or "secret" (only visible to organization owners
#' and members of a team), default is "secret"
#' @param prompt Logical. Should the user be prompted before deleting team. Default `true`.
#' @param new_team Character. New team name.
#'
#' @return All functions invisibly return a list containing
#' the results of the relevant GitHub API calls.
#'
#' @examples
#' \dontrun{
#' team_create("ghclass-test",c("hw1-team01","hw1-team02"))
#'
#' org_teams("ghclass-test", "hw1-")
#'
#' team_rename("ghclass-test", "hw1-team02", "hw1-team03")
#'
#' org_teams("ghclass-test", "hw1-")
#'
#' team_delete("ghclass-test", "hw1-team01", prompt = FALSE)
#'
#' org_teams("ghclass-test", "hw1-")
#'
#'
#' # Cleanup
#' team_delete("ghclass-test", org_teams("ghclass-test", "hw1-"), prompt = FALSE)
#' }
#'
NULL



#' @name team_members
#' @rdname team_members
#'
#' @title Tools for inviting, removing, and managing members of an organization team
#'
#' @description
#'
#' * `team_invite()` - add members to team(s).
#'
#' * `team_remove()` - remove members from team(s).
#'
#' * `team_members()` - returns a tibble of team members.
#'
#' * `team_pending()` - returns a tibble of pending team members.
#'
#' * `team_repos()` - returns a tibble of teams and their repos.
#'
#' @param org Character. Name of the GitHub organization.
#' @param user Character. One or more GitHub users to invite.
#' @param team Character. Name of teams.
#' @param team_type Character. Either "slug" if the team names are slugs or "name" if full team names are provided.
#'
#' @return `team_members()`, `team_pending()`, and `team_repos()` all return a tibble.
#' The `team` column echoes the supplied `team` values (names or slugs, per `team_type`)
#' and a `slug` column gives the corresponding team slug used for the API calls.
#'
#' `team_invite()` and `team_remove()` invisibly return a list containing the results
#' of the relevant GitHub API calls.
#'
#' @examples
#' \dontrun{
#' team_create("ghclass-test",c("hw1-team01","hw1-team02"))
#'
#' team_invite("ghclass-test", user = "rundel", team = c("hw1-team01", "hw1-team02", "missing_team"))
#'
#' team_remove("ghclass-test", user = "rundel", team = c("hw1-team01", "missing_team"))
#'
#' team_members("ghclass-test", org_teams("ghclass-test", "hw1-"))
#'
#' team_pending("ghclass-test", org_teams("ghclass-test", "hw1-"))
#'
#' # Add team repo
#' repo_create("ghclass-test", name = "hw1-team02")
#' repo_add_team("ghclass-test/hw1-team02", team = "hw1-team02")
#'
#' team_repos("ghclass-test", org_teams("ghclass-test", "hw1-"))
#'
#'
#' # Cleanup
#' repo_delete("ghclass-test/hw1-team02", prompt = FALSE)
#' team_delete("ghclass-test", org_teams("ghclass-test", "hw1-"), prompt = FALSE)
#' }
#'
NULL

github_api_team_remove = function(org, team_slug, username) {
  ghclass_api_v3_req(
    endpoint = "DELETE /orgs/:org/teams/:team_slug/memberships/:username",
    org = org,
    team_slug = team_slug,
    username = username
  )
}

#' @rdname team_members
#' @export
team_remove = function(org, user, team, team_type = c("name", "slug")) {
  arg_is_chr_scalar(org)
  arg_is_chr(user, team)
  team_type = match.arg(team_type)

  if (team_type == "name")
    slug = team_slug_lookup(org, team)
  else
    slug = team

  check_team_slug(slug)

  res = purrr::pmap(
    unique( tibble::tibble(user, team, slug) ),
    function(user, team, slug) {
      if (is.na(slug)) {
        cli::cli_alert_danger("Team {.val {team}} does not exist in org {.val {org}}.")
        return(NULL)
      }

      res = purrr::safely(github_api_team_remove)(org, slug, user)

      status_msg(
        res,
        "Removed user {.val {user}} from team {.val {team}}.",
        "Failed to remove user {.val {user}} from team {.val {team}}."
      )

      result(res)
    }
  )

  invisible(res)
}

github_api_org = function(org) {
  ghclass_api_v3_req(
    endpoint = "GET /orgs/:org",
    org = org,
    .send_headers = c(Accept = "application/vnd.github.surtur-preview+json")
  )
}

#' @rdname org_perm
#' @export
#'
org_sitrep = function(org) {
  res = purrr::safely(github_api_org)(org)

  if (failed(res)) {
    status_msg(res, fail = "Failed to find org {.val {org}}.")
    invisible(org)
  } else {
    admins = org_admins(org)
    workflow_perm = org_workflow_permissions(org)

    res_org = result(res)

    repo_perm = res_org$default_repository_permission
    if (repo_perm != "none")
      repo_perm_warn = "members can currently view {.emph all} repos in this org."
    else
      repo_perm_warn = NULL


    if (workflow_perm != "write")
      workflow_perm_warn = "this may prevent some GitHub actions from working correctly."
    else
      workflow_perm_warn = NULL

    allows_forking = res_org$members_can_fork_private_repositories
    if (allows_forking)
      forking_warn = "by default members can currently fork private repos in this org."
    else
      forking_warn = NULL

    cli::cli_h1("{.strong {res_org$login} sitrep:}")
    cli::cli_ul()
    cli::cli_li(cli_kv("Admins", admins))
    cli::cli_li(cli_kv("Members", res_org$collaborators))
    cli::cli_li(cli_kv("Public repos", res_org$public_repos))
    cli::cli_li(cli_kv("Private repos", res_org$total_private_repos))
    cli::cli_li(
      cli_kv(
        "Default repository permission",
        repo_perm,
        repo_perm_warn
      )
    )
    cli::cli_li(
      cli_kv(
        "Default workflow permissions",
        workflow_perm,
        workflow_perm_warn
      )
    )
    cli::cli_li(cli_kv("Members can create public repos", res_org$members_can_create_public_repositories))
    cli::cli_li(cli_kv("Members can create private repos", res_org$members_can_create_private_repositories))
    cli::cli_li(cli_kv("Members can fork private repos", allows_forking, forking_warn))
    cli::cli_end()

    invisible(res_org)
  }
}

#' @rdname ghclass-internal
#' @export
github_api_repo_get_file = function(repo, path, branch = NULL) {
  arg_is_chr_scalar(repo, path)
  arg_is_chr_scalar(branch, allow_null=TRUE)

  ghclass_api_v3_req(
    endpoint = "GET /repos/:owner/:repo/contents/:path",
    owner = get_repo_owner(repo),
    repo = get_repo_name(repo),
    path = path,
    ref = branch
  )

}


#' @rdname repo_file
#'
#' @export
#'
repo_get_file = function(repo, path, branch = NULL, quiet = FALSE, include_details = TRUE) {
  arg_is_chr_scalar(repo, path)
  arg_is_chr_scalar(branch, allow_null = TRUE)

  file = purrr::possibly(github_api_repo_get_file, NULL)(repo, path, branch)
  extract_content(repo = repo, path = path, file = file,
                  include_details = include_details, quiet = quiet)
}

github_api_repo_collaborators = function(repo) {
  ghclass_api_v3_req(
    endpoint = "GET /repos/:owner/:repo/collaborators",
    owner = get_repo_owner(repo),
    repo = get_repo_name(repo)
  )
}


#' @rdname repo_user
#' @export
#'
repo_collaborators = function(repo, include_admins = TRUE) {
  arg_is_chr(repo)
  arg_is_lgl_scalar(include_admins)

  purrr::map_dfr(
    repo,
    function(repo) {
      res = purrr::safely(github_api_repo_collaborators)(repo)
      status_msg(
        res,
        fail = "Failed to retrieve collaborators for {.val {repo}}."
      )

      collabs = result(res)

      d = if (empty_result(collabs)) {
        tibble::tibble(
          repo = character(),
          username = character(),
          pull  = logical(),
          push  = logical(),
          admin = logical()
        )
      } else {
        tibble::tibble(
          repo = repo,
          username = purrr::map_chr(collabs, "login"),
          pull  = purrr::map_lgl(collabs, c("permissions", "pull")),
          push  = purrr::map_lgl(collabs, c("permissions", "push")),
          admin = purrr::map_lgl(collabs, c("permissions", "admin"))
        )
      }

      if (!include_admins)
        d = dplyr::filter(d, !.data$admin)

      d
    }
  )
}

#' @rdname repo_file
#'
#' @param full_path Logical. Should the function return the full path of the files and directories. Default `FALSE`.
#'
#' @export
#'
repo_ls = function(repo, path = ".", branch = NULL, full_path = FALSE) {
  arg_is_chr_scalar(repo, path)
  arg_is_chr_scalar(branch, allow_null = TRUE)
  arg_is_lgl_scalar(full_path)

  res = purrr::safely(github_api_repo_get_file)(repo, path, branch)

  if (failed(res)) {
    status = error(res)[['headers']][['status']]

    cli_stop(
      "Failed to retrieve path {.val {path}} in repo {.val {repo}}.",
      " ({.val {status}})"
    )
  }
  files = purrr::map_chr(result(res), "path")
  if (!full_path)
    files = fs::path_file(files)

  files
}

#' @rdname local_repo
#' @export
#'
local_repo_push = function(repo_dir, remote = "origin", branch = NULL,
                           force = FALSE, prompt = TRUE, mirror = FALSE,
                           verbose = FALSE) {
  require_gert()

  arg_is_chr(repo_dir, remote)
  arg_is_chr(branch, allow_null = TRUE)
  arg_is_lgl_scalar(verbose)

  if (is.null(branch))
    branch = list(NULL)

  dir = repo_dir_helper(repo_dir)

  res = purrr::pmap(
    list(dir, remote, branch),
    function(dir, remote, branch) {
      if (is.null(branch))
        branch = gert::git_info(repo = dir)$shorthand

      repo = fs::path_file(dir)
      if (is.na(branch))
        ref = remote
      else
        ref = paste(remote, branch, sep="/")

      run = TRUE
      if (prompt && force) {
        remotes = gert::git_remote_list(repo = dir)
        remote_url = remotes$url[remotes$name == remote]

        run = cli_yeah("This command will overwrite the branch {.val {branch}} from repo {.val {remote_url}}.")
      }

      if (!run) {
        cli::cli_alert_danger("User canceled force push (overwrite) of {.val {ref}}")
        return(NULL)
      }

      refspec = if (is.na(branch))
        NULL
      else
        glue::glue("refs/heads/{branch}:refs/heads/{branch}")

      res = purrr::safely(gert::git_push)(
        remote = remote,
        refspec = refspec,
        repo = dir, force = force, mirror = mirror,
        verbose = verbose
      )

      status_msg(
        res,
        "Pushed from local repo {.val {repo}} to {.val {ref}}.",
        "Failed to push from local repo {.val {repo}} to {.val {ref}}."
      )

      res
    }
  )

  invisible(res)
}

#' @rdname local_repo
#'
#' @param repo GitHub repo address with the form `owner/name`.
#' @param local_path Local directory to store cloned repos.
#'
#' @aliases repo_clone
#'
#' @export
#'
local_repo_clone = function(repo, local_path=".", branch = NULL, mirror = FALSE, verbose = FALSE) {
  require_gert()

  arg_is_chr(repo)
  arg_is_chr(branch, allow_null=TRUE)
  arg_is_chr_scalar(local_path)
  arg_is_lgl_scalar(mirror, verbose)

  if (is.null(branch) || mirror) # If mirroring then branch should not be set
    branch = list(NULL)

  local_path = fs::path_expand(local_path)
  dir.create(local_path, showWarnings = FALSE, recursive = TRUE)

  dirs = purrr::map2_chr(
    repo, branch,
    function(repo, branch) {
      dir = fs::path(local_path, get_repo_name(repo))
      url = glue::glue("https://github.com/{repo}.git")

      res = purrr::safely(gert::git_clone)(
        url = url, path = dir, branch = branch, mirror = mirror, verbose = verbose
      )

      fmt_repo = format_repo(repo, branch)

      status_msg(
        res,
        "Cloned {.val {fmt_repo}}.",
        "Failed to clone {.val {fmt_repo}}."
      )

      ternary(succeeded(res), as.character(dir), NA_character_)
    }
  )

  names(dirs) = repo

  invisible(dirs)
}

#' @export
repo_clone = function(repo, local_path = ".", branch = NULL, mirror = FALSE, verbose = FALSE) {
  local_repo_clone(repo = repo, local_path = local_path, branch = branch, mirror = mirror, verbose = verbose)
}

filter_results = function(res, pattern = NULL, exclude = FALSE, col = NULL) {
  UseMethod("filter_results", res)
}

#' @exportS3Method
filter_results.default = function(res, pattern = NULL, exclude = FALSE, col = NULL) {
  if (is.null(pattern))
    return(res)

  subset = grepl(pattern, res)
  if (exclude)
    res[!subset]
  else
    res[subset]
}

#' @exportS3Method
filter_results.data.frame = function(res, pattern = NULL, exclude = FALSE, col = NULL) {
  if (is.null(pattern))
    return(res)

  stopifnot(!is.null(col))

  subset = grepl(pattern, res[[col]])
  if (exclude)
    res[!subset,]
  else
    res[subset,]
}

#' @name action
#' @rdname action
#'
#' @title Retrieve information about GitHub Actions workflows and their runs.
#'
#' @description
#' * `action_workflows()` - retrieve details on repo workflows.
#'
#' * `action_runs()` - retrieve details on repo workflow runs.
#'
#' * `action_status()` - DEPRECATED - retrieve details on most recent workflow runs.
#'
#' * `action_runtime()` - retrieves runtime durations for workflow runs.
#'
#' * `action_artifacts()` - retrieve details on available workflow artifacts.
#'
#' * `action_artifact_download()` - downloads artifact(s) into a local directory.
#'
#' * `action_artifact_delete()` - deletes artifact(s).
#'
#' * `action_run_logs()` - downloads workflow run log(s) into a local directory.
#'
#' @param repo Character. Address of repositories in `owner/name` format.
#' @param dir Character. Path to the directory where logs or artifacts will be saved.
#' @param ids Integer or data frame. Artifact ids to be downloaded or deleted.
#' If a data frame is passed then the `id` column will be used.
#'
#' @return
#'
#' `action_workflows()`, `action_runs()`, `action_runtime()`, and `action_artifacts`
#' all return tibbles containing information on requested repos' available workflows,
#' recent workflow runs, workflow runs runtimes, and generated artifacts
#' respectively.
#'
#' `action_artifact_download()` returns an invisible character vector containing the
#' paths of the extraction directories for each successfully downloaded artifact
#' (`NA` entries indicate failures).
#'
#' `action_artifact_delete()` returns an invisible data frame containing repository names and
#'  ids of the deleted artifacts.
#'
#' `action_run_logs()` returns an invisible character vector containing the paths of all
#' downloaded log files or directories.
#'
#' @examples
#' \dontrun{
#' action_workflows("rundel/ghclass")
#'
#' action_runs("rundel/ghclass")
#'
#' action_runtime(c("rundel/ghclass", "rundel/parsermd"))
#'
#' action_artifacts(c("rundel/ghclass", "rundel/parsermd"))
#' }
#'
NULL

#' @name action_badge
#' @rdname action_badge
#'
#' @title Add or remove GitHub Actions badges from a repository
#'
#' @description
#' * `action_add_badge()` - Add a GitHub Actions badge to a file.
#'
#' * `action_remove_badge()` - Remove one or more GitHub Action badges from a file.
#'
#' @param repo Character. Address of repository in `owner/name` format.
#' @param workflow Character. Name of the workflow.
#' @param workflow_pat Character. Name of the workflow to be removed, or a regex pattern
#' that matches the workflow name.
#' @param where Character. Regex pattern indicating where to insert the badge, defaults
#' to the beginning of the target file.
#' @param line_padding Character. What text should be added after the badge.
#' @param file Character. Target file to be modified, defaults to `README.md`.#'
#'
#' @return Both `action_add_badge()` and `action_remove_badge()` invisibly return a list
#' containing the results of the relevant GitHub API call.
#'
NULL

github_api_org_teams = function(org) {
  ghclass_api_v3_req(
    endpoint = "GET /orgs/:org/teams", 
    org = org
  )
}

#' @rdname org_details
#' @param team_type Character. Either "slug" if the team names are slugs or "name" if full team names are provided.
#' @export
#'
org_teams = function(org, filter = NULL, exclude = FALSE, team_type = c("name", "slug")) {
  arg_is_chr_scalar(org)
  arg_is_chr_scalar(filter, allow_null = TRUE)
  arg_is_lgl_scalar(exclude)
  team_type = match.arg(team_type)

  res = purrr::safely(github_api_org_teams)(org)

  status_msg(
    res,
    fail = "Failed to retrieve teams for org {.val {org}}."
  )

  if (failed(res) | empty_result(res)) {
    character()
  } else {
    teams = purrr::map_chr(result(res), team_type)
    filter_results(teams, filter, exclude)
  }
}

#' @rdname ghclass-internal
#' @export
cli_glue = function(..., .envir = parent.frame()) {
  txt = cli::cli_format_method(cli::cli_text(..., .envir = .envir))

  # cli_format_method does wrapping which we dont want at this stage
  # so glue things back together.
  paste(txt, collapse = " ")
}

#' @rdname ghclass-internal
#' @export
cli_stop = function(..., .envir = parent.frame()) {
  text = cli_glue(..., .envir = .envir)
  stop(paste(text, collapse = "\n"), call. = FALSE)
}

#' @rdname ghclass-internal
#' @export
cli_warn = function(..., .envir = parent.frame()) {
  text = cli_glue(..., .envir = .envir)
  warning(paste(text, collapse = "\n"), call. = FALSE)
}


# Simplified version of usethis::ui_yeah
cli_yeah = function (..., .envir = parent.frame()) {
  if (!rlang::is_interactive()) {
    cli_stop("User input required, but session is not interactive.")
  }

  cli::cli_text(..., .envir = .envir)
  opts = c("Yes", "No")
  s = utils::menu(opts)
  s != 0L && opts[[s]] == "Yes"
}

cli_kv = function(key, value, warning = NULL) {
  key = cli::col_silver(key)
  if (is.null(warning)) {
    cli_glue("{key}: {.val {value}}")
  } else {
    cli_glue("{key}: {.val {value}} <- {.strong {cli::col_red('Warning:')}} {cli::col_red(warning)}")
  }
}

github_api_repo = function(repo) {
  arg_is_chr_scalar(repo)

  ghclass_api_v3_req(
    endpoint = "GET /repos/:owner/:repo",
    owner = get_repo_owner(repo),
    repo = get_repo_name(repo),
    # Needed for template repos
    .send_headers = c(Accept = "application/vnd.github.baptiste-preview+json")
  )
}

#' @rdname repo_core
#' @param strict Logical. Should the old name of a renamed repositories be allowed. Default `FALSE`.
#' @param quiet Logical. Should details on renamed repositories be printed. Default `FALSE`.
#' @export
#'
repo_exists = function(repo, strict = FALSE, quiet = FALSE) {

  arg_is_chr(repo)
  arg_is_lgl_scalar(strict, quiet)

  # Checking if repo exists
  repo_details = purrr::map(repo, purrr::safely(github_api_repo))
  repo_exists = purrr::map_lgl(repo_details, succeeded)

  cur_names = purrr::map_chr(repo_details, c("result","full_name"), .default = NA)
  cur_names = replace_nas(cur_names, repo)

  renamed = cur_names != repo

  if (!quiet) {
    purrr::walk2(
      repo[renamed], cur_names[renamed],
      ~ cli::cli_alert_info("Repo {.val {.x}} has been renamed to {.val {.y}}.")
    )
  }

  if (strict)
    repo_exists = repo_exists & !renamed

  repo_exists
}

github_api_org_repo_stats = function(org, branch = NULL, filter = "", filter_type="in:name", inc_commits = TRUE, inc_issues = TRUE, inc_prs = TRUE) {
  filter_type = paste(filter_type, collapse = ",")

  org = graphql_escape(org)
  filter = graphql_escape(filter)
  filter_type = graphql_escape(filter_type)
  if (!is.null(branch))
    branch = graphql_escape(branch)

  query = '
    query {
      search(query: "org:{{{org}}} {{{filter}}} {{{filter_type}}}", type: REPOSITORY, first: 100, after: <graphql_quote(cursor)>) {
        repositoryCount
        pageInfo {
          hasNextPage
          endCursor
        }
        edges {
          node {
            ... on Repository {
              nameWithOwner
              isPrivate
              {{#inc_commits}}
              {{#branch}}
              object(expression: "{{{branch}}}") {
                ... on Commit {
                  history {
                    totalCount
                  }
                }
              }
              {{/branch}}
              {{^branch}}
              defaultBranchRef {
                name
                target {
                  ... on Commit {
                    history {
                      totalCount
                    }
                  }
                }
              }
              {{/branch}}
              pushedAt
              updatedAt
              {{/inc_commits}}
              {{#inc_issues}}
              closed_issues: issues(states: CLOSED) {
                totalCount
              }
              open_issues: issues(states: OPEN) {
                totalCount
              }
              {{/inc_issues}}
              {{#inc_prs}}
              open_prs: pullRequests(states: OPEN) {
                totalCount
              }
              merged_prs: pullRequests(states: MERGED) {
                totalCount
              }
              closed_prs: pullRequests(states: CLOSED) {
                totalCount
              }
              {{/inc_prs}}
            }
          }
        }
      }
    }
  '

  query = whisker::whisker.render(query)

  github_api_v4_graphql_paginated(query, page_info = c("search"))
}



#' @rdname org_details
#'
#' @param branch Character. The branch to use for counting commits, if `NULL` then each repo's default branch is used.
#' @param filter_type Character. One or more GitHub search `in` qualifiers.
#' See [documentation](https://help.github.com/en/articles/searching-for-repositories)
#' for more details.
#' @param inc_commits Logical. Include commit statistics (`branch`, `commits`, `last_update`). Default `TRUE`.
#' @param inc_issues Logical. Include issue statistics (`open_issues`, `closed_issues`). Default `TRUE`.
#' @param inc_prs Logical. Include pull request statistics (`open_prs`, `merged_prs`, `closed_prs`). Default `TRUE`.
#'
#' @export
#'
org_repo_stats = function(org, branch = NULL, filter = "", filter_type="in:name", inc_commits = TRUE, inc_issues = TRUE, inc_prs = TRUE) {
  warn_experimental()

  arg_is_chr_scalar(org, filter)
  arg_is_chr_scalar(branch, allow_null = TRUE)

  arg_is_chr(org, filter_type)
  arg_is_lgl_scalar(inc_commits, inc_issues, inc_prs)

  pages = github_api_org_repo_stats(org, branch, filter, filter_type, inc_commits, inc_issues, inc_prs)

  warn_if_search_capped(purrr::pluck(pages, 1, "data", "search", "repositoryCount"))

  purrr::map_dfr(
    pages,
    function(page) {
      repos = purrr::pluck(page, "data", "search", "edges")

      df = tibble::tibble(
        repo    = purrr::map_chr(repos, c("node", "nameWithOwner")),
        private = purrr::map_lgl(repos, c("node", "isPrivate"))
      )

      if (inc_commits) {
        if (is.null(branch)) {
          df$branch  = purrr::map_chr(repos, c("node", "defaultBranchRef", "name"), .default = NA_character_)
          df$commits = purrr::map_int(repos, c("node", "defaultBranchRef", "target", "history", "totalCount"), .default=NA_integer_)
        } else {
          df$branch  = branch
          df$commits = purrr::map_int(repos, c("node", "object", "history", "totalCount"), .default=NA_integer_)
        }

        df$last_update = lubridate::ymd_hms(purrr::map_chr(repos, c("node", "updatedAt")))
      }

      if (inc_issues) {
        df$open_issues   = purrr::map_int(repos, c("node", "open_issues", "totalCount"))
        df$closed_issues = purrr::map_int(repos, c("node", "closed_issues", "totalCount"))
      }

      if (inc_prs) {
        df$open_prs      = purrr::map_int(repos, c("node", "open_prs", "totalCount"))
        df$merged_prs    = purrr::map_int(repos, c("node", "merged_prs", "totalCount"))
        df$closed_prs    = purrr::map_int(repos, c("node", "closed_prs", "totalCount"))
      }

      df
    }
  ) %>%
    dplyr::distinct(.data$repo, .keep_all = TRUE)
}

github_api_repo_remove_team = function(
  org, team_slug, repo,
  permission = c("pull", "push", "admin", "maintain", "triage")
){
  permission = match.arg(permission)

  ghclass_api_v3_req(
    endpoint = "DELETE /orgs/:org/teams/:team_slug/repos/:owner/:repo",
    org = org,
    team_slug = team_slug,
    owner = get_repo_owner(repo),
    repo = get_repo_name(repo)
  )
}


#' @rdname repo_user
#' @export
repo_remove_team = function(
  repo, team,
  team_type = c("name", "slug")
) {
  arg_is_chr(repo, team)
  team_type = match.arg(team_type)

  org = unique(get_repo_owner(repo))

  if (length(org) != 1) {
    cli_stop("Teams can only be removed from repositories within a single organization. ",
             "Requested orgs: {.val {org}}")
  }

  d = tibble::tibble(team, repo)
  d = dplyr::distinct(d)

  if (team_type == "name")
    d[["team"]] = team_slug_lookup(org, d[["team"]])

  check_team_slug(d[["team"]])

  res = purrr::pmap(
    d,
    function(team, repo) {
      if (is.na(team))
        return()

      res = purrr::safely(github_api_repo_remove_team)(
        org = org,
        team_slug = team,
        repo = repo
      )

      status_msg(
        res,
        "Removed team {.val {team}} from repo {.val {repo}}.",
        "Failed to remove team {.val {team}} from repo {.val {repo}}."
      )
    }
  )

  invisible(res)
}

#' @rdname repo_core
#' @export
#'
repo_is_template = function(repo) {

  arg_is_chr(repo)

  # Checking if repo exists
  purrr::map_lgl(
    repo,
    function(repo) {
      res = purrr::safely(github_api_repo)(repo)
      status_msg(
        res,
        fail = "Failed to retrieve repo {.val {repo}}."
      )

      if (succeeded(res)) {
        result(res)[["is_template"]]
      } else {
        NA
      }
    }
  )
}

#' @name repo_core
#' @rdname repo_core
#'
#' @title GitHub Repository tools - core functions
#'
#' @description
#'
#' * `repo_create()` - create a GitHub repository.
#'
#' * `repo_delete()` - delete a GitHub repository.
#'
#' * `repo_rename()` - rename a repository, note that renamed repositories retain their
#' unique identifier and can still be accessed via their old names due to
#' GitHub re-directing.
#'
#' * `repo_exists()` - returns `TRUE` if the GitHub repository exists. It will also print
#'  a message if a repository has been renamed, unless `quiet = TRUE`.
#'
#' * `repo_mirror()` - mirror the content of a repository to another repository,
#' the target repo must already exist.
#'
#' * `repo_mirror_template()` - mirror the content of a source template repository to a new repository,
#' the target repo must *not* already exist.
#'
#' * `repo_is_template()` - returns `TRUE` if a repository is a template repo.
#'
#' * `repo_set_template()` - change the template status of a repository.
#'
#' * `repo_allows_forking()` - returns `TRUE` if forking is enabled for a repository.
#'
#' * `repo_set_forking()` - enable or disable forking for a repository.
#'
#' @param repo Character. Address of repository in `owner/repo` format.
#'
#' @return `repo_create()` returns a character vector of created repos (in `owner/repo` format)
#'
#' `repo_exists()`, `repo_is_template()`, and `repo_allows_forking()` all return a logical vector.
#'
#' `repo_set_forking()` invisibly returns a list containing the results of the relevant GitHub API calls.
#'
#' All other functions invisibly return a list containing the results of the relevant GitHub API calls.
#'
#' @examples
#' \dontrun{
#' repo_create("ghclass-test", "repo_test")
#'
#' repo_exists("ghclass-test/repo_test")
#'
#' repo_rename("ghclass-test/repo_test", "repo_test_new")
#'
#'
#' # The new repo exists
#' repo_exists("ghclass-test/repo_test_new")
#'
#' # The old repo forwards to the new repo
#' repo_exists("ghclass-test/repo_test")
#'
#'
#' # Check for the redirect by setting `strict = TRUE`
#' repo_exists("ghclass-test/repo_test", strict = TRUE)
#'
#'
#' # The prefered way of copying a repo is by making the source a template
#' repo_is_template("ghclass-test/repo_test_new")
#'
#' repo_set_template("ghclass-test/repo_test_new")
#'
#' repo_is_template("ghclass-test/repo_test_new")
#'
#'
#' # Given a template repo we can then directly copy the repo on GitHub
#' repo_mirror_template("ghclass-test/repo_test_new", "ghclass-test/repo_test_copy")
#'
#' repo_exists("ghclass-test/repo_test_copy")
#'
#' # Cleanup
#' repo_delete(
#'  c("ghclass-test/repo_test_new",
#'    "ghclass-test/repo_test_copy"),
#'  prompt = FALSE
#' )
#'
#' }
#'
#'
NULL


#' @name repo_details
#' @rdname repo_details
#'
#' @title GitHub Repository tools - repository details
#'
#' @description
#'
#' * `repo_clone_url()` - Returns the url, for cloning, a GitHub repo (either ssh or https)
#'
#' * `repo_branches()` - Returns a (filtered) vector of branch names.
#'
#' * `repo_commits()` - Returns a tibble of commits to a GitHub repository.
#'
#' * `repo_issues()` - Returns a tibble of issues for a GitHub repository.
#'
#' * `repo_n_commits()` - Returns a tibble of the number of commits in a GitHub repository (and branch).
#'
#' * `repo_prs()` - Returns a tibble of pull requests for a GitHub repository.
#'
#' * `repo_pushes()` - Returns a tibble of push activity to a GitHub repository.
#'
#' @param repo Character. Address of repository in `owner/repo` format.
#'
#' @details
#'
#' For `repo_commits()`, GitHub returns two timestamps for each commit: the
#' author date (`date`) and the committer date (`committer_date`). These
#' differ for rebased, cherry-picked, or amended commits. The `since` and
#' `until` arguments are filtered by GitHub against the committer date.
#'
#' The `login` column is `NA` when the commit's author email is not linked
#' to a GitHub account. The git author's `name` and `email` are still
#' populated in those cases.
#'
#' @return `repo_clone_url()` and `repo_branches()` both return a character vector.
#'
#' `repo_commits()`, `repo_issues()`, `repo_n_commits()`, and `repo_prs()` all return a tibble.
#'
#' @examples
#' \dontrun{
#' repo_clone_url("rundel/ghclass")
#'
#' repo_branches("rundel/ghclass")
#'
#' repo_commits("rundel/ghclass")
#'
#' repo_issues("rundel/ghclass")
#'
#' repo_n_commits("rundel/ghclass", branch = "master")
#'
#' repo_prs("rundel/ghclass")
#' }
#'
NULL


#' @name repo_file
#' @rdname repo_file
#'
#' @title GitHub Repository tools - file functions
#'
#' @description
#'
#' * `repo_add_file()` - Add / update files in a GitHub repository.
#' Note that due to delays in caching, files that have been added
#' very recently might not yet be displayed as existing and might accidentally
#' be overwritten.
#'
#' * `repo_delete_file()` - Delete a file from a GitHub repository
#'
#' * `repo_modify_file()` - Modify an existing file within a GitHub repository.
#'
#' * `repo_ls()` - Low level function for listing the files in a GitHub Repository
#'
#' * `repo_put_file()` - Low level function for adding a file to a GitHub repository
#'
#' * `repo_get_file()` - Low level function for retrieving the content of a file from a GitHub Repository
#'
#' * `repo_get_readme()` - Low level function for retrieving the content of the `README.md` of a GitHub Repository
#'
#'
#' @param repo Character. Address of repository in `owner/name` format.
#' @param path Character. File's path within the repository.
#' @param message Character. Commit message.
#' @param branch Character. Name of branch to use.
#' @param quiet Logical. Should status messages be printed. Default `FALSE`.
#' @param include_details Logical. Should file details be attached as attributes. Default `TRUE`.
#'
#' @return
#' `repo_delete_file()`, `repo_modify_file()`, and `repo_put_file()` all invisibly
#' return a list containing the results of the relevant GitHub API calls.
#'
#' `repo_ls()` returns a character vector of repo files in the given path.
#'
#' `repo_get_file()` and `repo_get_readme()` return a character vector with API results
#' attached as attributes if `include_details = TRUE`
#'
#' @examples
#' \dontrun{
#' repo = repo_create("ghclass-test", "repo_file_test", auto_init=TRUE)
#'
#' repo_ls(repo, path = ".")
#'
#' repo_get_readme(repo, include_details = FALSE)
#'
#' repo_get_file(repo, ".gitignore", include_details = FALSE)
#'
#' repo_modify_file(
#'   repo, path = "README.md", pattern = "repo_file_test",
#'   content = "\n\nHello world!\n", method = "after"
#' )
#'
#' repo_get_readme(repo, include_details = FALSE)
#'
#' repo_add_file(repo, file = system.file("DESCRIPTION", package="ghclass"))
#'
#' repo_get_file(repo, "DESCRIPTION", include_details = FALSE)
#'
#' repo_delete_file(repo, "DESCRIPTION")
#'
#' repo_delete(repo, prompt=FALSE)
#' }
#'
NULL


#' @name repo_notification
#' @rdname repo_notification
#'
#' @title GitHub Repository tools - notification functions
#'
#' @description
#'
#' * `repo_ignore()` - Ignore a GitHub repository.
#'
#' * `repo_unwatch()` - Unwatch / unsubscribe from a GitHub repository.
#'
#' * `repo_watch()` - Watch / subscribe to a GitHub repository.
#'
#' * `repo_watching()` - Returns a vector of your watched repositories. This should
#' match the list at [github.com/watching](https://github.com/watching).
#'
#' @param repo repository address in `owner/repo` format
#'
#' @return `repo_ignore()`, `repo_unwatch()`, and `repo_watch()` all invisibly return a
#' list containing the results of the relevant GitHub API call.
#'
#' `repo_watching()` returns a character vector of watched repos.
#'
#' @examples
#' \dontrun{
#' repo_ignore("Sta323-Sp19/hw1")
#'
#' repo_unwatch("rundel/ghclass")
#'
#' repo_watch("rundel/ghclass")
#' }
#'
NULL


#' @name repo_user
#' @rdname repo_user
#'
#' @title GitHub Repository tools - user functions
#'
#' @description
#'
#' * `repo_add_user()` - Add a user to a repository
#' * `repo_remove_user()` - Remove a user from a repository
#' * `repo_add_team()` - Add a team to a repository
#' * `repo_remove_team()` - Remove a team from a repository
#' * `repo_user_permission()` - Change a collaborator's permissions for a repository
#' * `repo_team_permission()` - Change a team's permissions for a repository
#' * `repo_collaborators()` - Returns a data frame of repos, their collaborators, and their permissions.
#' * `repo_contributors()` - Returns a data frame containing details on repository contributor(s).
#'
#' @param repo Character. Address of repository in `owner/repo` format.
#' @param user Character. One or more GitHub usernames.
#' @param permission Character. Permission to be granted to a user or team for repo, defaults to "push".
#' @param team Character. Slug or name of team to add.
#' @param team_type Character. Either "slug" if the team names are slugs or "name" if full team names are provided.
#' @param include_admins Logical. If `FALSE`, user names of users with Admin rights are not included, defaults to `TRUE`.
#'
#' @return `repo_collaborators()` and `repo_contributors()` return a tibble.
#'
#' All other functions invisibly return a list containing the results of the relevant GitHub API calls.
#'
#' @details
#'
#' Permissions can be set to any of the following:
#' * `"pull"` - can pull, but not push to or administer this repository.
#' * `"push"` - can pull and push, but not administer this repository.
#' * `"admin"` - can pull, push and administer this repository.
#' * `"maintain"` - Recommended for project managers who need to manage the repository without access to sensitive or destructive actions.
#' * `"triage"` - Recommended for contributors who need to proactively manage issues and pull requests without write access.
#'
#' `repo_contributors()` is built on GitHub's `/stats/contributors` endpoint,
#' which groups commits by GitHub user. Commits whose author email is not
#' linked to a GitHub account are not included in the result.
#'
#'
#' @examples
#' \dontrun{
#' repo = repo_create("ghclass-test", "hw1")
#'
#' team_create("ghclass-test", "team_awesome")
#'
#' repo_add_user(repo, "rundel")
#'
#' repo_add_team(repo, "team_awesome")
#'
#' repo_remove_team(repo, "team_awesome")
#'
#' repo_collaborators(repo)
#'
#' repo_contributors(repo)
#' repo_contributors("rundel/ghclass")
#'
#' # Cleanup
#' repo_delete(repo, prompt=FALSE)
#' }
#'
NULL

github_api_repo_watching = function() {
  ghclass_api_v3_req(
    endpoint = "GET /user/subscriptions"
  )
}

#' @rdname repo_notification
#'
#' @param filter character, regex pattern for matching (or excluding) repositories.
#' @param exclude logical, should entries matching the regex be excluded or included.
#'
#' @export
#'
repo_watching = function(filter = NULL, exclude = FALSE){
  arg_is_chr_scalar(filter, allow_null = TRUE)
  arg_is_lgl_scalar(exclude)

  res = purrr::map_chr(github_api_repo_watching(), "full_name")
  filter_results(res, filter, exclude)
}

github_api_repo_search = function(q, sort = NULL, order = c("desc", "asc")) {
  if (!is.null(sort))
    stopifnot(sort %in% c("stars", "forks", "help-wanted-issues", "updated"))

  order = match.arg(order)

  arg_is_chr_scalar(q)
  arg_is_chr_scalar(sort, order, allow_null = TRUE)

  ghclass_api_v3_req(
    endpoint = "GET /search/repositories",
    q = q,
    sort = sort,
    order = order
  )
}

#' @rdname org_details
#' @param name Character. Full or partial repo name to search for within the org
#' @param extra Character. Any additional search qualifiers, see
#' [Searching for repositories](https://docs.github.com/en/free-pro-team@latest/github/searching-for-information-on-github/searching-for-repositories)
#' for details.
#' @export
#'
org_repo_search = function(org, name, extra = "", full_repo = TRUE) {
  arg_is_chr_scalar(org, name, extra)
  arg_is_lgl_scalar(full_repo)

  q = glue::glue("org:{org} {name} in:name {extra}")

  res = purrr::safely(github_api_repo_search)(q)


  status_msg(
    res,
    fail = "Repo search failed for org {.val {org}}."
  )

  if (failed(res) | empty_result(res))
    return(character())

  warn_if_search_capped(result(res)[["total_count"]])

  if (full_repo) {
    purrr::map_chr(result(res)[["items"]], "full_name")
  } else {
    purrr::map_chr(result(res)[["items"]], "name")
  }
}

github_api_delete_artifact = function(repo, id) {
  arg_is_chr_scalar(repo)
  arg_is_pos_int(id)

  ghclass_api_v3_req(
    endpoint = "DELETE /repos/{owner}/{repo}/actions/artifacts/{artifact_id}",
    owner = get_repo_owner(repo),
    repo = get_repo_name(repo),
    artifact_id = id
  )

}


#' @name action
#' @rdname action
#'
#' @export
#'
action_artifact_delete = function(repo, ids) {

  arg_is_chr(repo)

  if (is.numeric(ids))
    ids = tibble::tibble(repo = repo, id = ids)
  arg_is_df(ids)

  df = dplyr::left_join(
    tibble::tibble(repo = repo), ids,
    by = "repo"
  ) %>%
    dplyr::select("repo", "id")

  purrr::pwalk(
    df,
    function(repo, id) {
      res = purrr::safely(github_api_delete_artifact)(repo, id)

      status_msg(
        res,
        "Deleted artifact {.val {id}} from repo {.val {repo}}.",
        "Failed to delete artifact with id {.val {id}} from repo {.val {repo}}."
      )
    }
  )

  invisible(df)
}

#' @name user
#' @rdname user
#'
#' @title GitHub user related tools
#'
#' @description
#'
#' * `user_exists()` - returns `TRUE` if the username(s) (or organization) exist on GitHub
#' and `FALSE` otherwise. Note that GitHub considers organizations to be a type of user.
#'
#' * `user_repos()` - returns a (filtered) vector of repositories belonging to the user.
#'
#' * `user_type()` - returns a vector of the accounts' types.
#'
#' @param user Character. GitHub username(s).
#' @param type Character. Can be one of "all", "owner", "public", "private", "member".
#' @param filter Character. Regular expression pattern for matching (or excluding) repositories.
#' @param exclude Logical. Should entries matching the regular expression in `filter` be excluded or included? Default `FALSE`.
#' @param full_repo Logical. Should the full repository address be returned
#' (e.g. `owner/repo` instead of just `repo`)? Default `TRUE`.
#'
#' @return `user_exists()` returns a logical vector.
#'
#' `user_repos()` and `user_type()` retrun a character vector.
#'
#' @examples
#' \dontrun{
#' user_exists(c("rundel", "ghclass-test", "hopefullydoesnotexist"))
#'
#' user_repos("rundel", type = "public", filter = "ghclass")
#'
#' user_repos("ghclass-test")
#'
#' org_repos("ghclass-test")
#'
#' user_type(c("rundel", "ghclass-test"))
#' }

NULL

github_api_team_add = function(
  org, team_slug, repo,
  permission = c("pull", "push", "admin", "maintain", "triage")
){
  permission = match.arg(permission)

  ghclass_api_v3_req(
    endpoint = "PUT /orgs/:org/teams/:team_slug/repos/:owner/:repo",
    org = org,
    team_slug = team_slug,
    owner = get_repo_owner(repo),
    repo = get_repo_name(repo),
    permission = permission
  )
}


#' @rdname repo_user
#' @param team Character. Slug or name of team to add.
#' @param team_type Character. Either "slug" if the team names are slugs or "name" if full team names are provided.
#' @export
repo_add_team = function(
  repo, team,
  permission = c("push", "pull", "admin", "maintain", "triage"),
  team_type = c("name", "slug")
) {
  arg_is_chr(repo, team, allow_empty = FALSE)
  permission = match.arg(permission)
  team_type = match.arg(team_type)

  org = unique(get_repo_owner(repo))

  if (length(org) != 1) {
    cli_stop("Permissions can only be changed for one organization at a time. ",
             "Requested orgs: {.val {org}}")
  }

  repo = unique(repo)
  team = unique(team)

  d = tibble::tibble(team, repo)
  d = dplyr::distinct(d)

  if (team_type == "name")
    d[["team"]] = team_slug_lookup(org, d[["team"]])

  check_team_slug(d[["team"]])

  res = purrr::pmap(
    d,
    function(team, repo) {
      if (is.na(team))
        return()

      res = purrr::safely(github_api_team_add)(
        org = org,
        team_slug = team,
        repo = repo,
        permission = permission
      )

      status_msg(
        res,
        "Team {.val {team}} given {.val {permission}} access to repo {.val {repo}}",
        "Failed to give team {.val {team}} {.val {permission}} access to repo {.val {repo}}."
      )
    }
  )

  invisible(res)
}

#' @rdname repo_user
#' @export
#'
repo_team_permission = function(
  repo, team,
  permission = c("push", "pull", "admin", "maintain", "triage"),
  team_type = c("name", "slug")
) {
  arg_is_chr(repo, team, allow_empty = FALSE)
  permission = match.arg(permission)
  team_type = match.arg(team_type)

  repo_add_team(
    repo = repo, team = team, permission = permission, team_type = team_type
  )
}

# github_api_repo_commits_v4 = function(repo, branch) {
#   arg_is_chr_scalar(repo, branch)
#
#   owner = get_repo_owner(repo)
#   name  = get_repo_name(repo)
#
#   query = '
#     query {
#       repository(owner: <graphql_quote(owner)>, name: <graphql_quote(name)>) {
#         object(expression: <graphql_quote(branch)>) {
#           ... on Commit {
#             history(first:100, after: <graphql_quote(cursor)>) {
#               totalCount
#               pageInfo {
#                 hasNextPage
#                 endCursor
#               }
#               nodes {
#                 author {
#                   email
#                   name
#                   user {
#                     login
#                   }
#                 }
#                 changedFiles
#                 message
#                 additions
#                 deletions
#                 committedDate
#               }
#             }
#           }
#         }
#       }
#     }
#   '
#
#   github_api_v4_graphql_paginated(
#     query,
#     vars = list(owner = owner, name = name, branch = branch),
#     page_info = c("repository", "object", "history")
#   )
# }


github_api_repo_commits = function(repo, sha=NULL, path=NULL, author=NULL, since=NULL, until=NULL) {
  ghclass_api_v3_req(
    endpoint = "GET /repos/:owner/:repo/commits",
    owner = get_repo_owner(repo),
    repo = get_repo_name(repo),
    sha = sha,
    path = path,
    author = author,
    since = since,
    until = until
  )
}

#' @rdname repo_details
#'
#' @param branch Character.	Branch to list commits from.
#' @param sha	   Character.	SHA to start listing commits from.
#' @param path	 Character.	Only commits containing this file path will be returned.
#' @param author Character.	GitHub login or email address by which to filter commit author.
#' @param since	 Character.	Only commits after this date will be returned, expects `YYYY-MM-DDTHH:MM:SSZ` format. Filtered against the `committer_date` column, see Details.
#' @param until	 Character.	Only commits before this date will be returned, expects `YYYY-MM-DDTHH:MM:SSZ` format. Filtered against the `committer_date` column, see Details.
#' @param quiet  Logical. Should an error message be printed if the repo does not exist. Default `FALSE`.
#'
#' @export
#'

repo_commits = function(repo, branch = NULL, sha = branch, path = NULL,
                        author = NULL, since = NULL, until = NULL, quiet = FALSE) {

  arg_is_chr(repo)
  arg_is_chr_scalar(branch, sha, path, author, since, until, allow_null = TRUE)
  arg_is_lgl_scalar(quiet)

  purrr::map_dfr(
    repo,
    function(repo) {
      res = purrr::safely(github_api_repo_commits)(
        repo, sha, path, author, since, until
      )

      # API gives an error if the repo has 0 commits
      res = allow_error(res, message = "Git Repository is empty")

      if (!quiet) {
        status_msg(
          res,
          fail = "Failed to retrieve commits from {.val {repo}}."
        )
      }

      commits = result(res)

      if (empty_result(commits)) {
        tibble::tibble(
          repo           = character(),
          login          = character(),
          name           = character(),
          email          = character(),
          date           = as.POSIXct(character()),
          committer_date = as.POSIXct(character()),
          msg            = character()
        )
      } else {
        tibble::tibble(
          repo  = repo,
          login = purrr::map_chr(commits, c("author", "login"), .default = NA),
          name  = purrr::map_chr(commits, c("commit", "author","name"), .default = NA),
          email = purrr::map_chr(commits, c("commit", "author","email"), .default = NA),
          date  = lubridate::ymd_hms(
            purrr::map_chr(commits, c("commit", "author", "date"), .default = NA)
          ),
          committer_date = lubridate::ymd_hms(
            purrr::map_chr(commits, c("commit", "committer", "date"), .default = NA)
          ),
          msg = purrr::map_chr(commits, c("commit", "message"), .default = NA)
        )
      }
    }
  )
}

#' @title Rename local directories using a vector of patterns and replacements.
#'
#' @description  This function is meant to help with renaming local student repos to include
#' something more useful like `Last, First name` or a unique identifier
#' for the purposes of ordering repository folders.
#'
#' @param repo_dir Character. A single directory containing one or more repos.
#' @param pattern Character. One or more regexp patterns to match to directory names.
#' @param replacement Character.  One or more text strings containing the replacement value for matched patterns.
#'
#' @return Returns a character vector of the new repo directory paths, or `NA` if the rename
#' failed.
#'
#' @examples
#' \dontrun{
#' # Prefix each cloned repo directory with a roster identifier
#' local_repo_rename(
#'   "/path/to/cloned/repos",
#'   pattern = "hw1-(.*)",
#'   replacement = "\\1"
#' )
#' }
#'
#' @export
#'
local_repo_rename = function(repo_dir, pattern, replacement) {
  arg_is_chr_scalar(repo_dir)
  arg_is_chr(pattern, replacement)

  # If there isn't a regex group, add one for the whole pattern
  if (!any(grepl(pattern = "\\(.*\\)", x = pattern))) {
    pattern = paste0("(", pattern, ")")
  }

  repos = repo_dir_helper(repo_dir)
  cur_repos = repos

  for(i in seq_along(pattern)) {
    repos = sub(pattern[i], replacement[i], repos)
  }

  sub = repos != cur_repos
  res = purrr::map2_chr(
    cur_repos[sub], repos[sub],
    function(cur, new) {
      res = purrr::safely(fs::file_move)(cur, new)

      status_msg(
        res,
        "Renaming {.val {fs::path_file(cur)}} to {.val {fs::path_file(new)}}.",
        "Failed to rename {.val {fs::path_file(cur)}} to {.val {fs::path_file(new)}}."
      )

      ifelse(succeeded(res), new, NA)
    }
  )

  invisible(res)
}

#' @title Setup grading for an assignment
#'
#' @description
#' This is a higher level function that automates the common steps for
#' setting up grading of an assignment:
#'
#' * Find repos matching a filter pattern in an organization
#' * Clone all matched repos locally
#' * Download GitHub Actions artifacts (e.g. rendered html, md files)
#' * Report any missing artifacts
#' * Create comment template files for each repo
#'
#' @param path Character. Root directory for the grading folder (created if it doesn't exist).
#' @param org Character. Name of the GitHub organization.
#' @param repo_filter Character. Regex pattern passed to [org_repos()]'s `filter`
#'   argument to select repos (e.g. `"hw01_"`).
#' @param artifacts Named character vector. Names are subfolder names, values are
#'   regex patterns passed as `filter` to [action_artifact_download()] (matched
#'   against artifact names). E.g. `c("html" = "html-output", "report" = "pdf-report")`.
#' @param comment_template Character. Template text used to populate each comment
#'   markdown file. Defaults to `""`.
#' @param key_repo Character. Optional repository address in `owner/name` format
#'   to clone into the root of the grading folder as an answer key.
#'
#' @return An invisible list containing:
#'   * `repos` — character vector of matched repo addresses
#'   * `cloned` — result from [local_repo_clone()]
#'   * `artifacts` — named list of results from each [action_artifact_download()] call
#'   * `comments` — character vector of comment file paths created
#'   * `key` — result from cloning the key repo (if provided)
#'
#' @export
#'
org_grade_assignment = function(
  path,
  org,
  repo_filter,
  artifacts = character(),
  comment_template = "",
  key_repo = NULL
) {
  arg_is_chr_scalar(path, org, repo_filter, comment_template)
  arg_is_chr_scalar(key_repo, allow_null = TRUE)
  arg_is_chr(artifacts)

  if (length(artifacts) > 0 && is.null(names(artifacts))) {
    cli_stop("{.arg artifacts} must be a named character vector.")
  }

  if (dir.exists(path)) {
    cli_stop("Destination directory {.file {path}} already exists, please remove it or choose a different path.")
  }

  if (!is.null(key_repo) && !repo_exists(key_repo, quiet = TRUE)) {
    cli_stop("Key repo {.val {key_repo}} does not exist.")
  }

  repos = org_repos(org, repo_filter)

  if (length(repos) == 0) {
    cli_stop("No repos found in {.val {org}} matching {.val {repo_filter}}.")
  }

  res = list(repos = repos)

  repos_dir = file.path(path, "repos")
  cli::cli_alert_info("Cloning {.val {length(repos)}} student repo{?s} matching {.val {repo_filter}}.")
  res[["cloned"]] = local_repo_clone(repos, repos_dir)

  res[["artifacts"]] = list()
  repo_names = get_repo_name(repos)

  if (length(artifacts) > 0) {
    cli::cli_alert_info("Gathering artifacts.")

    for (nm in names(artifacts)) {
      art_dir = file.path(path, nm)
      res[["artifacts"]][[nm]] = action_artifact_download(
        repos, art_dir, filter = artifacts[[nm]]
      )
    }

    for (nm in names(artifacts)) {
      art_dir = file.path(path, nm)
      if (!dir.exists(art_dir)) next

      subdirs = list.dirs(art_dir, recursive = FALSE, full.names = FALSE)

      missing = purrr::keep(repo_names, function(rn) {
        !any(subdirs == rn | startsWith(subdirs, paste0(rn, "_")))
      })

      if (length(missing) > 0) {
        cli::cli_alert_warning(
          "Missing {.val {nm}} artifacts for {.val {length(missing)}} repo{?s}: {.val {missing}}."
        )
      }
    }
  }

  comments_dir = file.path(path, "comments")
  cli::cli_alert_info("Creating comment files.")
  dir.create(comments_dir, showWarnings = FALSE, recursive = TRUE)

  comment_files = file.path(comments_dir, paste0(repo_names, ".md"))

  purrr::walk(
    comment_files,
    writeLines,
    text = comment_template
  )

  cli::cli_alert_success("Created {.val {length(comment_files)}} comment file{?s} in {.file {comments_dir}}.")
  res[["comments"]] = comment_files

  if (!is.null(key_repo)) {
    cli::cli_alert_info("Cloning key repo {.val {key_repo}}.")
    res[["key"]] = local_repo_clone(key_repo, path)
  }

  invisible(res)
}

github_api_org_repo_forking = function(org) {
  org = graphql_escape(org)

  query = '
    query {
      organization(login: "{{{org}}}") {
        repositories(first: 100, after: <graphql_quote(cursor)>, privacy: PRIVATE) {
          totalCount
          pageInfo {
            hasNextPage
            endCursor
          }
          nodes {
            nameWithOwner
            forkingAllowed
            forkCount
          }
        }
      }
    }
  '

  query = whisker::whisker.render(query)

  github_api_v4_graphql_paginated(query, page_info = c("organization", "repositories"))
}


#' @rdname org_details
#' @export
#'
org_repo_forking = function(org) {
  arg_is_chr_scalar(org)

  pages = github_api_org_repo_forking(org)

  purrr::map_dfr(
    pages,
    function(page) {
      nodes = purrr::pluck(page, "data", "organization", "repositories", "nodes")

      tibble::tibble(
        repo          = purrr::map_chr(nodes, "nameWithOwner"),
        allow_forking = purrr::map_lgl(nodes, "forkingAllowed"),
        n_forks       = purrr::map_int(nodes, "forkCount")
      )
    }
  )
}

github_api_org_workflow_permissions = function(org) {
  ghclass_api_v3_req(
    endpoint = "GET /orgs/:org/actions/permissions/workflow",
    org = org,
    .send_headers = c(Accept = "application/vnd.github+json")
  )
}

#' @rdname org_perm
#' @export
#'
org_workflow_permissions = function(org) {
  arg_is_chr_scalar(org)

  res = purrr::safely(github_api_org_workflow_permissions)(org)

  if (failed(res)) {
    cli::cli_abort("Failed to find org {.val {org}}.")
  }

  result(res)$default_workflow_permissions
}


github_api_org_set_workflow_permissions = function(org, default_workflow_permissions = "read", can_approve_pull_request_reviews = FALSE) {
  ghclass_api_v3_req(
    endpoint = "PUT /orgs/:org/actions/permissions/workflow",
    org = org,
    default_workflow_permissions = default_workflow_permissions,
    can_approve_pull_request_reviews = can_approve_pull_request_reviews,
    .send_headers = c(Accept = "application/vnd.github+json")
  )
}

#' @rdname org_perm
#' @export
#'
org_set_workflow_permissions = function(org, workflow_permission = c("read","write")) {
  workflow_permission = match.arg(workflow_permission)
  arg_is_chr_scalar(org, workflow_permission)

  res = purrr::safely(github_api_org_set_workflow_permissions)(
    org, default_workflow_permissions = workflow_permission
  )

  status_msg(
    res,
    "Set org {.val {org}}'s default workflow permissions to {.val {workflow_permission}}.",
    "Failed to set org {.val {org}}'s default workflow permissions."
  )

  invisible(result(res))
}

github_api_repo_mirror_template = function(source_repo, repo, private){
  ghclass_api_v3_req(
    endpoint = "POST /repos/:template_owner/:template_repo/generate",
    template_owner = get_repo_owner(source_repo),
    template_repo = get_repo_name(source_repo),
    owner = get_repo_owner(repo),
    name = get_repo_name(repo),
    private = private,
    .send_headers = c(Accept = "application/vnd.github.baptiste-preview+json")
  )
}

#' @rdname repo_core
#'
#' @param source_repo Character. Address of template repository in `owner/name` format.
#' @param target_repo Character. One or more repository addresses in `owner/name` format.
#' Note when using template repos these new repositories must *not* exist.
#' @param private Logical. Should the new repository be private or public. Default `TRUE`.
#'
#' @export
#'
repo_mirror_template = function(source_repo, target_repo, private = TRUE) {
  arg_is_chr_scalar(source_repo)
  arg_is_chr(target_repo)
  arg_is_lgl_scalar(private)

  if (length(repo_ls(source_repo)) == 0) {
    cli_stop("Cannot mirror {.val {source_repo}} because the repository is empty.")
  }

  target_repo = unique(target_repo)
  exists = repo_exists(target_repo)

  res = purrr::map2(
    target_repo, exists,
    function(repo, exists) {
      res = purrr::safely(
        function() {
          if (exists) {
            cli_stop("Cannot mirror (template) to repo {.val {repo}} because this reposistory already exists.")
          }

          github_api_repo_mirror_template(source_repo, repo, private)
        }
      )()

      status_msg(
        res,
        "Mirrored repo {.val {source_repo}} to repo {.val {repo}}.",
        "Failed to mirror repo {.val {source_repo}} to repo {.val {repo}}."
      )

      res
    }
  )

  invisible(res)
}

github_api_branch_delete = function(repo, branch) {
  ghclass_api_v3_req(
    endpoint = "DELETE /repos/:owner/:repo/git/refs/:ref",
    owner = get_repo_owner(repo),
    repo = get_repo_name(repo),
    ref = paste0("heads/", branch)
  )
}


#' @rdname branch
#' @export
#'
branch_delete = function(repo, branch) {
  arg_is_chr(repo, branch)

  invisible( purrr::pmap(
    list(repo, branch),
    function(repo, branch) {
      res = purrr::safely(github_api_branch_delete)(repo, branch)

      status_msg(
        res,
        "Removed branch {.val {format_repo(repo, branch)}}.",
        "Failed to remove branch {.val {format_repo(repo, branch)}}."
      )

      res
    }
  ) )
}

#' @rdname branch
#' @export
#'
branch_remove = function(repo, branch) {
  .Deprecated("branch_delete")
  branch_delete(repo, branch)
}

github_api_action_runs = function(
  repo, actor = NULL, branch = NULL, event = NULL, status = NULL,
  created = NULL,
  limit = 1
) {
  arg_is_chr_scalar(repo)
  arg_is_chr_scalar(actor, branch, event, status, allow_null = TRUE)

  ghclass_api_v3_req(
    endpoint = "GET /repos/:owner/:repo/actions/runs",
    owner = get_repo_owner(repo),
    repo = get_repo_name(repo),
    actor	= actor,
    branch = branch,
    event = event,
    status = status,
    created = created,
    limit = limit
  )
}

github_api_action_workflow_runs = function(
    repo, workflow_id,
    actor = NULL, branch = NULL, event = NULL,
    status = NULL, created = NULL,
    limit = 1
) {
  arg_is_chr_scalar(repo)
  arg_is_pos_int_scalar(workflow_id)
  arg_is_chr_scalar(actor, branch, event, status, allow_null = TRUE)

  ghclass_api_v3_req(
    endpoint = "GET /repos/{owner}/{repo}/actions/workflows/{workflow_id}/runs",
    owner = get_repo_owner(repo),
    repo = get_repo_name(repo),
    workflow_id = workflow_id,
    actor	= actor,
    branch = branch,
    event = event,
    status = status,
    created = created,
    limit = limit
  )
}


#' @name action
#' @rdname action
#'
#' @param repo Character. Address of repository in `owner/name` format.
#' @param branch Character. Filter runs associated with a particular branch.
#' @param event Character. Filter runs for triggered by a specific event. See
#' [here](https://docs.github.com/en/actions/automating-your-workflow-with-github-actions/events-that-trigger-workflows)
#' for possible event names.
#' @param status Character. Filter runs for a particular status or conclusion (e.g. `completed` or `success`).
#' @param created Character. Filter runs for a given creation date.
#' See [here](https://docs.github.com/en/search-github/getting-started-with-searching-on-github/understanding-the-search-syntax#query-for-dates)
#' for date query syntax.
#' @param limit Numeric. Maximum number of workflow runs to return. Default `1`. Note results
#' are chronologically ordered, so `limit = 1` will return the most recent action run for a repository.
#'
#' @export
#'
action_runs = function(
  repo,
  branch = NULL,
  event = NULL,
  status = NULL,
  created = NULL,
  limit = 1
) {
  arg_is_chr(repo)
  arg_is_chr_scalar(branch, event, status, created, allow_null = TRUE)
  arg_is_pos_int_scalar(limit)

  purrr::map_dfr(
    repo,
    function(repo) {
      purrr::pmap_dfr(
        action_workflows(repo)[,c("name", "id")],
        function(name, id) {
          res = purrr::safely(github_api_action_workflow_runs)(
            repo, workflow_id = id,
            branch = branch,
            event = event, status = status,
            created = created, limit = limit
          )

          status_msg(
            res,
            fail = "Failed to retrieve workflow runs for repo {.val {repo}}."
          )

          if (failed(res) || empty_result(res) || result(res)[["total_count"]] == 0) {
            tibble::tibble(
              repo = character(),
              workflow = character(),
              run_id = double(),
              branch = character(),
              commit = character(),
              actor = character(),
              event = character(),
              status = character(),
              conclusion = character(),
              created = lubridate::ymd_hms()
            )
          } else {
            runs = result(res)[["workflow_runs"]]
            run_df = tibble::tibble(
              repo   = repo,
              workflow = name,
              run_id = purrr::map_dbl(runs, "id", .default = NA),
              branch = purrr::map_chr(runs, "head_branch", .default = NA),
              commit = purrr::map_chr(runs, "head_sha", .default = NA),
              actor  = purrr::map_chr(runs, c("actor", "login"), .default = NA),
              event  = purrr::map_chr(runs, "event", .default = NA),
              status = purrr::map_chr(runs, "status", .default = NA),
              conclusion = purrr::map_chr(runs, "conclusion", .default = NA),
              created = purrr::map_chr(runs, "created_at", .default = NA) %>% lubridate::ymd_hms()
            )
          }
        }
      )
    }
  )
}


#' @name action
#' @rdname action
#'
#' @export
#'
action_status = function(
    repo,
    branch = NULL,
    event = NULL,
    status = NULL,
    created = NULL,
    limit = 1
) {
  lifecycle::deprecate_warn("0.2.2", "action_status()", "action_runs()")

  action_runs(repo = repo, branch = branch, event = event,
              status = status, created = created, limit = limit)
}

github_api_repo_delete_file = function(repo, path, message, sha = NULL, branch = NULL) {

  if (is.null(sha)) {
    cur_file = repo_get_file(repo, path, branch, quiet = TRUE)
    sha = attr(cur_file, "sha")
  }

  if (is.null(sha)) {
    cli_stop("Unable to locate file using the given path.")
  }

  ghclass_api_v3_req(
    "DELETE /repos/:owner/:repo/contents/:path",
    owner = get_repo_owner(repo),
    repo = get_repo_name(repo),
    path = path,
    message = message,
    sha = sha,
    branch = branch
  )
}


#' @rdname repo_file
#'
#' @export
#'
repo_delete_file = function(repo, path, message = NULL, branch = NULL) {
  arg_is_chr(repo, path)
  arg_is_chr(branch, allow_null = TRUE)
  arg_is_chr_scalar(message, allow_null = TRUE)

  if (is.null(branch))
    branch = list(NULL)

  res = purrr::pmap(
    list(repo, path, branch),
    function(repo, path, branch) {
      if (is.null(message))
        message = cli_glue("Deleting file {path}")

      res = purrr::safely(
        function() {
          github_api_repo_delete_file(repo, path, message, branch = branch)
        }
      )()

      status_msg(
        res,
        "Deleted file {.file {path}} from repo {.val {repo}}.",
        "Failed to delete file {.file {path}} from repo {.val {repo}}."
      )

      res
    }
  )

  invisible(res)
}

github_api_issue_create = function(repo, title, body, labels=character(), assignees=character()){
  arg_is_chr_scalar(repo, title, body)
  arg_is_chr(labels, assignees)

  ghclass_api_v3_req(
    endpoint = "POST /repos/:owner/:repo/issues",
    owner = get_repo_owner(repo),
    repo = get_repo_name(repo),
    title = title,
    body = body,
    labels = I(labels),
    assignees = I(assignees)
  )
}


#' @rdname issue
#'
#' @param delay Numeric. Delay between each API request. Issue creation has a secondary rate limit (~ 20/min).
#'
#' @export
#'
issue_create = function(repo, title, body, labels = character(), assignees = character(), delay=0) {

  arg_is_chr(repo, title, body)

  if (!is.list(labels))
    labels = list(labels)

  if (!is.list(assignees))
    assignees = list(assignees)

  # Handle any necessary recycling
  df = tibble::tibble(
    repo, title, body, labels, assignees
  )

  n = nrow(df)

  if (n > 20 & delay == 0) {
    cli::cli_alert_info(c(
      "Attempting to create {.val {n}} issues - this is likely to trigger GitHub's secondary rate limit.",
      "Setting a delay of 5 seconds per request to avoid this, override this using {.field delay}."
    ), wrap=TRUE)

    delay = 5
  }

  res = purrr::pmap(
    df,
    function(repo, title, body, labels, assignees) {
      res = purrr::safely(github_api_issue_create)(
        repo, title, body, labels, assignees
      )

      status_msg(
        res,
        "Created issue {.val {title}} for repo {.val {repo}}.",
        "Failed to create issue {.val {title}} for repo {.val {repo}}."
      )

      Sys.sleep(delay)

      res
    }
  )

  invisible(res)
}

github_api_org_repo_create = function(repo, private, auto_init, gitignore_template){
  ghclass_api_v3_req(
    endpoint = "POST /orgs/:owner/repos",
    owner = get_repo_owner(repo),
    name = get_repo_name(repo),
    private = private,
    auto_init = auto_init,
    gitignore_template = gitignore_template
  )
}

#' @rdname repo_core
#'
#' @param org Character. GitHub organization that will own the repository
#' @param name Character. Repository name
#' @param prefix Character. Common repository name prefix
#' @param suffix Character. Common repository name suffix
#' @param private Logical. Create private repositories? Default `TRUE`.
#' @param auto_init Logical. Should the repository be initialized with a `README.md`. Default `FALSE`.
#' @param gitignore_template Character. `.gitignore` language template to use.
#'
#' @export
#'
repo_create = function(
  org, name,
  prefix = "", suffix = "",
  private = TRUE, auto_init = FALSE,
  gitignore_template = "R"
) {

  arg_is_chr(name)
  arg_is_chr_scalar(org, prefix, suffix, gitignore_template)
  arg_is_lgl_scalar(private, auto_init)

  name = unique(name)

  repo = paste0(prefix, name, suffix)
  repo = paste0(org, "/", repo)

  if (length(repo) != length(unique(repo)))
    cli_stop("Not all repo names are unique: {.val {repo}}.")

  exists = repo_exists(repo)

  res = purrr::map2(
    repo, exists,
    function(repo, exists) {
      if (exists) {
        cli::cli_alert_info("Skipping repo {.val {repo}}, it already exists.")
        return(repo)
      }
      res = purrr::safely(github_api_org_repo_create)(
        repo,
        private = private,
        auto_init = auto_init,
        gitignore_template = gitignore_template
      )

      status_msg(
        res,
        "Created repo {.val {repo}}.",
        "Failed to create repo {.val {repo}}."
      )

      ternary(succeeded(res), repo, NULL)
    }
  )

  invisible(purrr::flatten_chr(res))
}

github_api_get_org = function(org) {
  arg_is_chr_scalar(org)

  ghclass_api_v3_req(
    endpoint = "/orgs/:org",
    org = org
  )
}


#' @rdname org_details
#' @export
#'
org_exists = function(org) {
  arg_is_chr(org)

  res = purrr::map(org, purrr::safely(github_api_get_org))
  purrr::map_lgl(res, succeeded)
}

github_api_org_teams_v4 = function(org) {
  org = graphql_escape(org)

  query = '
    query {
      organization(login: "{{{org}}}") {
        teams(first: 100, after: <graphql_quote(cursor)>) {
          pageInfo {
             hasNextPage
             endCursor
          }
          edges {
            node {
              id
              name
              slug
              privacy
            }
          }
        }
      }
    }
  '

  query = whisker::whisker.render(query)

  github_api_v4_graphql_paginated(query, page_info = c("organization", "teams"))
}


org_team_details_v4 = function(org) {
  arg_is_chr_scalar(org)

  pages = github_api_org_teams_v4(org)

  purrr::map_dfr(
    pages,
    function(page) {
      teams = purrr::pluck(page, "data", "organization", "teams", "edges")

      tibble::tibble(
        org  = org,
        id   = purrr::map_chr(teams, c("node", "id")),
        name = purrr::map_chr(teams, c("node", "name")),
        slug   = purrr::map_chr(teams, c("node", "slug")),
        privacy = purrr::map_chr(teams, c("node", "privacy")) %>% tolower(),
      )
    }
  )
}

github_api_repo_activity = function(repo,  ref = NULL, actor = NULL, time_period = NULL, activity_type = NULL) {
  arg_is_chr_scalar(repo, allow_null = FALSE)
  arg_is_chr_scalar(ref, actor, time_period, activity_type, allow_null = TRUE)

  if (!is.null(time_period))
    stopifnot(time_period %in% c("day", "week", "month", "quarter", "year"))

  if (!is.null(activity_type))
    stopifnot(activity_type %in% c("push", "force_push", "branch_creation", "branch_deletion", "pr_merge", "merge_queue_merge"))

  ghclass_api_v3_req(
    endpoint = "GET /repos/:owner/:repo/activity",
    owner = get_repo_owner(repo),
    repo = get_repo_name(repo),
    ref = ref,
    actor = actor,
    time_period = time_period,
    activity_type = activity_type
  )
}

#' @rdname repo_details
#'
#' @param time_period  Character. The time period to filter by.
#' Options are  "all time", "day", "week", "month", "quarter", "year".
#'
#' @export
#'

repo_pushes = function(repo, branch = NULL, author = NULL, time_period = c("all time", "day", "week", "month", "quarter", "year"), quiet = FALSE) {

  time_period = match.arg(time_period)

  if (time_period == "all time")
    time_period = NULL

  arg_is_chr(repo)
  arg_is_chr_scalar(branch, author, time_period, allow_null = TRUE)
  arg_is_lgl_scalar(quiet)

  purrr::map_dfr(
    repo,
    function(repo) {
      res = purrr::safely(github_api_repo_activity)(
        repo, ref = branch, actor = author, time_period = time_period, activity_type = "push"
      )

      if (!quiet) {
        status_msg(
          res,
          fail = "Failed to retrieve activity from {.val {repo}}."
        )
      }

      pushes = result(res)

      if (empty_result(pushes)) {
        tibble::tibble(
          repo  = character(),
          login = character(),
          ref  = character(),
          activity = character(),
          date  = as.POSIXct(character()),
          before = character(),
          after = character()
        )
      } else {
        tibble::tibble(
          repo   = repo,
          login  = purrr::map_chr(pushes, c("actor", "login"), .default = NA),
          ref    = purrr::map_chr(pushes, c("ref"), .default = NA),
          activity = purrr::map_chr(pushes, c("activity_type"), .default = NA),
          date   = lubridate::ymd_hms(
            purrr::map_chr(pushes, c("timestamp"), .default = NA)
          ),
          before = purrr::map_chr(pushes, c("before"), .default = NA),
          after  = purrr::map_chr(pushes, c("after"), .default = NA)
        )
      }
    }
  )
}

github_api_repo_n_commits = function(repo) {
  org = graphql_escape(get_repo_owner(repo))
  repo = graphql_escape(get_repo_name(repo))

  query = '
    query {
      repository(owner:"{{{org}}}", name:"{{{repo}}}") {
        nameWithOwner
        defaultBranchRef {
          name
          target {
            ... on Commit {
              history {
                totalCount
              }
            }
          }
        }
      }
    }
  '

  query = whisker::whisker.render(query)

  github_api_v4_graphql(query)
}



#' @rdname repo_details
#'
#' @param repo   Character. Address of repository in `owner/name` format.
#' @param quiet  Logical. Should an error message be printed if a repo does not exist. Default `FALSE`.
#'
#' @export
#'
repo_n_commits = function(repo, quiet = FALSE) {
  arg_is_chr(repo)
  arg_is_lgl_scalar(quiet)

  purrr::map_dfr(
    repo,
    function(repo) {
      res = purrr::safely(github_api_repo_n_commits)(repo)

      if (failed(res)) {
        if (!quiet)
          cli::cli_alert_danger("Failed to retrieve commits for {.val {repo}}.")

        tibble::tibble(
          repo = repo,
          branch = NA_character_,
          n = NA_integer_
        )
      } else {
        res = result(res)
        tibble::tibble(
          repo = purrr::pluck(res, "data", "repository", "nameWithOwner"),
          branch = purrr::pluck(res, "data", "repository", "defaultBranchRef", "name", .default=NA_character_),
          n = purrr::pluck(res, "data", "repository", "defaultBranchRef", "target", "history", "totalCount", .default=0)
        )
      }
    }
  )
}

#' @rdname org_details
#'
#' @param user Character. One or more GitHub usernames.
#' @param prefix Character. String to prepend to each user name.
#' @param suffix Character. String to append to each user name.
#'
#' @export
#'
org_user_repos = function(org, user, prefix = "", suffix = "") {
  arg_is_chr_scalar(org, prefix, suffix)
  arg_is_chr(user)

  paste0(org, "/", prefix, user, suffix)
}

github_api_repo_get_readme = function(repo, branch = NULL) {
  arg_is_chr_scalar(repo)
  arg_is_chr_scalar(branch, allow_null=TRUE)

  ghclass_api_v3_req(
    endpoint = "GET /repos/:owner/:repo/readme",
    owner = get_repo_owner(repo),
    repo = get_repo_name(repo),
    ref = branch
  )
}

#' @rdname repo_file
#'
#' @export
#'
repo_get_readme = function(repo, branch = NULL, include_details = TRUE) {
  arg_is_chr_scalar(repo)
  arg_is_chr_scalar(branch, allow_null=TRUE)

  file = purrr::possibly(github_api_repo_get_readme, NULL)(repo, branch)
  extract_content(repo, path = "README.md", file = file, include_details = include_details)
}

github_api_repo_rename = function(repo, new_name){
  ghclass_api_v3_req(
    endpoint = "PATCH /repos/:owner/:repo",
    owner = get_repo_owner(repo),
    repo = get_repo_name(repo),
    name = new_name
  )
}

#' @rdname repo_core
#' @param new_repo Character. New name of repository without the owner.
#' @export
#'
repo_rename = function(repo, new_repo) {
  arg_is_chr(repo, new_repo)

  res = purrr::map2(
    repo, new_repo,
    function(repo, new_repo) {
      res = purrr::safely(github_api_repo_rename)(repo, new_repo)
      new_full = paste0(get_repo_owner(repo), "/", new_repo)

      status_msg(
        res,
        "Renamed repo {.val {repo}} to {.val {new_full}}.",
        "Failed to rename repo {.val {repo}} to {.val {new_full}}."
      )
    }
  )

  invisible(res)
}

github_api_team_delete = function(org, team_slug) {
  ghclass_api_v3_req(
    endpoint = "DELETE /orgs/:org/teams/:team_slug",
    org = org,
    team_slug = team_slug
  )
}

#' @rdname team
#' @export
#'
team_delete = function(org, team, team_type = c("name", "slug"), prompt = TRUE) {

  arg_is_chr_scalar(org)
  arg_is_chr(team)
  arg_is_lgl_scalar(prompt)
  team_type = match.arg(team_type)

  if (prompt) {
    delete = cli_yeah("This command will delete the following teams permanently: {.val {team}}.")
    if (!delete) {
      return(invisible())
    }
  }

  if (team_type == "name")
    team = team_slug_lookup(org, team)

  check_team_slug(team)

  res = purrr::map(
    team,
    function(team) {

      if (is.na(team)) {
        cli::cli_alert_danger("Team {.val {team}} does not exist in org {.val {org}}.")
        return()
      }

      res = purrr::safely(github_api_team_delete)(org, team)

      status_msg(
        res,
        "Deleted team {.val {team}} from org {.val {org}}.",
        "Failed to delete team {.val {team}} from org {.val {org}}."
      )
    }
  )

  invisible(res)
}

#' @rdname ghclass-internal
#' @export
read_bin_file = function(x) {
  f = file(x, "rb")
  size = file.size(x)
  raw = readBin(f, "raw", size)
  close(f)

  raw
}


#' @rdname repo_file
#'
#' @param file Character. Local file path(s) of file or files to be added.
#' @param repo_folder Character. Name of folder on repository to save the file(s) to. If the folder does not exist on the repository, it will be created.
#' @param preserve_path Logical. Should the local relative path be preserved. Default `FALSE`.
#' @param overwrite Logical. Should existing file or files with same name be overwritten. Default `FALSE`.
#'
#' @export
#'
repo_add_file = function(repo, file, message = NULL, repo_folder = NULL, branch = NULL,
                         preserve_path = FALSE, overwrite = FALSE) {

  arg_is_chr(repo, file)
  arg_is_chr(branch, allow_null=TRUE)
  arg_is_chr_scalar(repo_folder, message, allow_null = TRUE)
  arg_is_lgl_scalar(preserve_path, overwrite)

  missing_files = file[!fs::file_exists(file)]
  if (length(missing_files) != 0)
    cli_stop("Unable to locate the following file{?s}: {.val {missing_files}}")

  if (is.character(file) & (length(file) > 1))
    file = list(file)

  if (is.null(branch))
    branch = list(NULL)

  res = purrr::pmap(
    list(repo, file, branch),
    function(repo, file, branch) {
      purrr::map(
        file,
        function(file){
          gh_path = file

          if (!preserve_path)
            gh_path = fs::path_file(file)

          if(!is.null(repo_folder))
            gh_path = fs::path(repo_folder, gh_path)

          if (!file_exists(repo, gh_path, branch) | overwrite) {
            repo_put_file(
              repo = repo,
              path = gh_path,
              content = read_bin_file(file),
              message = message,
              branch = branch,
              verbose = TRUE
            )
          } else {
            cli::cli_alert_danger( c(
              "Failed to add file {.val {gh_path}} to repo {.val {repo}}, this file already exists. ",
              "If you want to force add this file, re-run the command with {.code overwrite = TRUE}."
            ) )

            NULL
          }
        }
      )
    }
  )

  invisible(res)
}

github_api_repo_add_user = function(repo, username, permission){
  ghclass_api_v3_req(
    endpoint = "PUT /repos/:owner/:repo/collaborators/:username",
    owner = get_repo_owner(repo),
    repo = get_repo_name(repo),
    username = username,
    permission = permission
  )
}



#' @rdname repo_user
#' @export
#'
repo_add_user = function(repo, user, permission = c("push", "pull", "admin", "maintain", "triage")) {
  permission = match.arg(permission)
  arg_is_chr(repo, user)

  res = purrr::map2(
    repo, user,
    function(repo, user) {
      res = purrr::safely(github_api_repo_add_user)(
        repo = repo,
        username = user,
        permission = permission
      )

      status_msg(
        res,
        "User {.val {user}} given {.val {permission}} access to repo {.val {repo}}",
        "Failed to give user {.val {user}} {.val {permission}} access to repo {.val {repo}}."
      )
    }
  )

  invisible(res)
}

#' @rdname repo_user
#' @export
#'
repo_user_permission = function(
  repo, user,
  permission = c("push", "pull", "admin", "maintain", "triage")
) {
  permission = match.arg(permission)
  arg_is_chr(repo, user)

  repo_add_user(repo = repo, user = user, permission = permission)
}

github_api_team_create = function(org, name, privacy) {
  ghclass_api_v3_req(
    endpoint = "POST /orgs/:org/teams",
    org=org, name=name, privacy=privacy
  )
}

#' @rdname team
#' @export
#'
team_create = function(
  org, team, prefix = "", suffix = "",
  privacy = c("secret","closed")
) {
  arg_is_chr_scalar(org, prefix, suffix)
  arg_is_chr(team)

  team = unique(as.character(team))
  privacy = match.arg(privacy)

  team = paste0(prefix, team, suffix)

  org_teams = org_teams(org)

  new_teams = setdiff(team, org_teams)
  existing_teams = intersect(team, org_teams)

  if (length(existing_teams) > 0)
    cli::cli_alert_info("Skipping existing teams: {.val {existing_teams}}.")

  res = purrr::map(
    new_teams,
    function(team) {
      res = purrr::safely(github_api_team_create)(
        org = org, name = team, privacy = privacy
      )

      status_msg(
        res,
        "Created team {.val {team}} in org {.val {org}}.",
        "Failed to create team {.val {team}} in org {.val {org}}."
      )
    }
  )

  invisible(res)
}

#' @rdname local_repo
#' @export
#'
local_repo_add = function(repo_dir, files = ".") {
  require_gert()
  arg_is_chr(repo_dir, files)
  repo_dir = repo_dir_helper(repo_dir)

  res = purrr::map(
    repo_dir,
    function(dir) {
      if (identical(files, ".")) {
        files = gert::git_status(repo = dir)[["file"]]
      }

      res = purrr::safely(gert::git_add)(
        files = files, repo = dir
      )

      # TODO - handle the case where there are 0 additions

      repo = fs::path_file(dir)
      status_msg(
        res,
        "Added files {.val {files}} to {.val {repo}}.",
        "Failed to add files {.val {files}} to {.val {repo}}."
      )

      res
    }
  )

  invisible(res)
}

#' @rdname org_details
#' @export
#'
org_team_details = function(org) {
  arg_is_chr_scalar(org)

  res = purrr::safely(github_api_org_teams)(org)

  status_msg(
    res,
    fail = "Failed to retrieve teams for org {.val {org}}."
  )

  if (failed(res) | empty_result(res)) {
    tibble::tibble(
      slug = character(),
      id   = integer(),
      name = character(),
      privacy = character(),
      permission = character()
    )
  } else {
    r = result(res)
    tibble::tibble(
      slug       = purrr::map_chr(r, "slug"),
      id         = purrr::map_int(r, "id"),
      name       = purrr::map_chr(r, "name"),
      privacy    = purrr::map_chr(r, "privacy"),
      permission = purrr::map_chr(r, "permission")
    )
  }
}

#' @name github_token
#' @rdname github_token
#'
#' @title Tools for handling GitHub personal access tokens (PAT)
#'
#' @description
#' * `github_get_token` - returns the user's GitHub personal access token (PAT).
#'
#' * `github_set_token` - defines the user's GitHub PAT by setting the `GITHUB_PAT` environmental variable.
#' This value will persist until the session ends or `gihub_reset_token()` is called.
#'
#' * `github_reset_token` - removes the value stored in the `GITHUB_PAT` environmental variable.
#'
#' * `github_test_token` - checks if a PAT is valid by attempting to authenticate with the GitHub API.
#'
#' * `github_token_scopes` - returns a vector of scopes granted to the token.
#'
#' @param token Character. Either the literal token, or the path to a file containing the token.
#'
#' @details
#' This package looks for the personal access token (PAT) in the following places (in order):
#' * Value of `GITHUB_PAT` environmental variable.
#' * Any GitHub PAT token(s) stored with `gitcreds` via `gitcreds_set()`.
#'
#' For additional details on creating a GitHub PAT see the usethis vignette on
#' [Managing Git(Hub) Credentials](https://usethis.r-lib.org/articles/articles/git-credentials.html).
#' For those who do not wish to read the entire article, the quick start method is to use:
#' * `usethis::create_github_token()` - to create the token and then,
#' * `gitcreds::gitcreds_set()` - to securely cache the token.
#'
#' @return `github_get_token()` returns the current PAT as a character string with the `gh_pat`
#' class. See [gh::gh_token()] for additional details.
#'
#' `github_set_token()` and `github_reset_token()` return the result of `Sys.setenv()` and
#' `Sys.unsetenv()` respectively.
#'
#' `github_test_token()` invisibly returns a logical value, `TRUE` if the test passes,
#' `FALSE` if not.
#'
#' `github_token_scopes()` returns a character vector of granted scopes.
#'
#' @examples
#' \dontrun{
#' github_test_token()
#'
#' github_token_scopes()
#'
#' (pat = github_get_token())
#'
#' github_set_token("ghp_BadTokenBadTokenBadTokenBadTokenBadToken")
#' github_get_token()
#' github_test_token()
#'
#' github_set_token(pat)
#' }
#'
NULL


# Modified from usethis
# https://github.com/r-lib/usethis/blob/7c8e0049a1e40e6dcabbde069bb29576215a11b6/R/github_token.R
scold_for_renviron = function () {
  renviron_path = fs::path_expand(Sys.getenv("R_ENVIRON_USER", unset = ""))
  if (renviron_path == "")
    renviron_path = fs::path_home_r()

  renviron_path = fs::path(renviron_path, ".Renviron")

  if (!fs::file_exists(renviron_path))
    return(invisible())

  renviron_lines = readLines(renviron_path, encoding = "UTF-8", warn = FALSE)
  fishy_lines = grep("^GITHUB_(PAT|TOKEN).*=.+", renviron_lines, value = TRUE)

  if (length(fishy_lines) == 0)
    return(invisible())

  fishy_keys = gsub("=.*", "", fishy_lines)

  cli::cli_warn(
    c(
      "{.path {renviron_path}} defines environment {cli::qty(fishy_keys)} variable{?s}: {.val {fishy_keys}}.",
      i = paste(
        "This is no longer considered a best practice, and it is recommended that you instead use {.fun gitcreds::gitcreds_set}",
        "to securely store your PAT. Existing environement variables can be removed from {.path .Renviron} using",
        "{.fun usethis::edit_r_environ}."
      )
    ),
    .frequency = "once",
    .frequency_id = "warn_renviron_pat"
  )

  invisible()
}




#' @rdname github_token
#' @export
#'
github_get_token = function() {
  scold_for_renviron()

  # Give priority to env GITHUB_PAT if set (so we use PAT from github_set_token preferentially)
  token = Sys.getenv("GITHUB_PAT", "")
  class(token) = "gh_pat"

  if (token == "")
    token = gh::gh_token()

  if (file.exists("~/.github/token") & token == "") {
    cli::cli_abort( c(
      "File based token storage is no longer directly supported by ghclass:",
      "*" = "You can temporarily load this PAT using {.code github_set_token(\"~/.github/token\")}.",
      "*" = "or you can securely store your PAT using {.fun gitcreds::gitcreds}."
    ) )
  }

  if (token == "")
    cli::cli_abort( c(
      "Unable to locate a github token:",
      "*" = "You can temporarily set the token using {.fun github_set_token}",
      "*" = "or you can securely store your PAT using {.fun gitcreds::gitcreds}."
    ) )

  token
}

# Resolve a `token` argument that may be either a literal PAT or a path to a
# file containing one. Only an existing file is read, taking the first
# non-empty, whitespace-trimmed line.
read_token = function(token) {
  arg_is_chr_scalar(token)

  if (file.exists(token)) {
    lines = trimws(readLines(token, warn = FALSE))
    token = lines[nzchar(lines)][1]
  }

  token
}

#' @rdname github_token
#' @export
#'
github_set_token = function(token) {
  Sys.setenv(GITHUB_PAT = read_token(token))
}

#' @rdname github_token
#' @export
#'
github_reset_token = function() {
  Sys.unsetenv("GITHUB_PAT")
}

#' @rdname github_token
#' @export
#'
github_test_token = function(token = github_get_token()) {
  token = read_token(token)

  res = purrr::safely(gh::gh)("/user", .token = token)

  status_msg(
    res,
    "Your GitHub PAT authenticated correctly.",
    "Your GitHub PAT failed to authenticate.",
  )

  invisible(succeeded(res))
}

#' @rdname github_token
#' @export
#'
github_token_scopes = function(token = github_get_token()) {
  token = read_token(token)

  res = purrr::safely(gh::gh)("/user", .token = token)

  status_msg(
    res,
    fail = "Your GitHub PAT failed to authenticate."
  )

  scopes = attr(result(res), "response")[["x-oauth-scopes"]]

  if (is.null(scopes))
    return(character())

  strsplit(scopes, ", ")[[1]]
}

github_api_get_user = function(user) {
  arg_is_chr_scalar(user)

  ghclass_api_v3_req(
    endpoint = "/users/:username",
    username = user
  )
}

#' @rdname user
#' @export
#'
user_exists = function(user) {
  arg_is_chr(user)

  res = purrr::map(user, purrr::safely(github_api_get_user))
  purrr::map_lgl(res, succeeded)
}

#' @rdname local_repo
#' @export
local_repo_commit = function(repo_dir, message) {
  require_gert()
  arg_is_chr(repo_dir, message)
  repo_dir = repo_dir_helper(repo_dir)

  res = purrr::map2(
    repo_dir, message,
    function(dir, message) {
      res = purrr::safely(gert::git_commit)(
        message = message, repo = dir
      )

      repo = fs::path_file(dir)
      status_msg(
        res,
        "Committed changes to {.val {repo}}.",
        "Failed to commit changes to {.val {repo}}."
      )

      res
    }
  )

  invisible(res)
}

#' @rdname github_token
#' @export
#'
github_rate_limit = function() {
  gh::gh_rate_limit(.token = github_get_token())
}


#' @rdname github_token
#' @export
#'
github_graphql_rate_limit = function() {
  query = '
    query {
      viewer {
        login
      }
      rateLimit {
        limit
        cost
        remaining
        resetAt
      }
    }
  '

  github_api_v4_graphql(query)[[c("data","rateLimit")]]
}

#' @rdname org_perm
#' @export
#'
org_allows_forking = function(org) {
  arg_is_chr_scalar(org)

  res = purrr::safely(github_api_org)(org)

  if (failed(res)) {
    cli::cli_abort("Failed to find org {.val {org}}.")
  }

  result(res)$members_can_fork_private_repositories
}

github_api_repo_branches = function(repo) {
  ghclass_api_v3_req(
    endpoint = "GET /repos/:owner/:repo/branches",
    owner = get_repo_owner(repo),
    repo = get_repo_name(repo)
  )
}

#' @rdname repo_details
#' @export
#'
repo_branches = function(repo) {
  arg_is_chr_scalar(repo)

  res = purrr::safely(github_api_repo_branches)(repo)

  status_msg(
    res,
    fail = "Failed to retrieve branches for repo {.val {repo}}."
  )

  if (failed(res) | empty_result(res)) {
    character()
  } else {
    purrr::map_chr(result(res), "name")
  }
}

#' @name pages
#' @rdname pages
#'
#' @title Retrieve information about GitHub Pages sites and builds.
#'
#' @description
#' * `pages_enabled()` - returns `TRUE` if a Pages site exists for the repo.
#'
#' * `pages_status()` - returns more detailed information about a repo's Pages site.
#'
#' * `pages_create()` - creates a Pages site for the provided repos.
#'
#' * `pages_delete()` - deletes the Pages site for the provided repos.
#'
#' @param repo Character. Address of repositories in `owner/name` format.
#'
#' @return
#'
#' `pages_enabled()` returns a named logical vector - `TRUE` if a Pages site exists, `FALSE` otherwise.
#'
#' `pages_status()` returns a tibble containing details on Pages sites.
#'
#' `pages_create()` & `pages_delete()` return an invisible list containing the API responses.
#'
#' @examples
#' \dontrun{
#' pages_enabled("rundel/ghclass")
#'
#' pages_status("rundel/ghclass")
#' }
#'
NULL


github_api_pages = function(repo) {
  arg_is_chr_scalar(repo)

  ghclass_api_v3_req(
    endpoint = "GET /repos/:owner/:repo/pages",
    owner = get_repo_owner(repo),
    repo = get_repo_name(repo)
  )
}



#' @name pages
#' @rdname pages
#'
#' @export
#'
pages_enabled = function(repo) {
  arg_is_chr(repo)

  purrr::map_lgl(
    repo,
    function(repo) {
      purrr::safely(github_api_pages)(repo) %>%
        succeeded()
    }
  ) %>%
    stats::setNames(repo)
}




#' @name pages
#' @rdname pages
#'
#' @export
#'
pages_status = function(repo) {
  arg_is_chr(repo)

  purrr::map_dfr(
    repo,
    function(repo) {
      res = purrr::safely(github_api_pages)(repo)

      status_msg(
        res,
        fail = "Failed find Pages information for repo {.val {repo}}."
      )

      if (failed(res) || empty_result(res)) {
        tibble::tibble(
          repo = character(),
          status = character(),
          url    = character(),
          build_type = character(),
          branch = character(),
          path   = character(),
          public = logical(),
          cname  = character(),
          custom_404 = logical(),
          https_enforced = logical()
        )
      } else {
        page = list(result(res))
        tibble::tibble(
          repo   = repo,
          status = purrr::map_chr(page, "status", .default = NA),
          url    = purrr::map_chr(page, "html_url", .default = NA),
          build_type = purrr::map_chr(page, "build_type", .default = NA),
          branch = purrr::map_chr(page, c("source", "branch"), .default = NA),
          path   = purrr::map_chr(page, c("source", "path"), .default = NA),
          public = purrr::map_lgl(page, "public", .default = NA),
          cname  = purrr::map_chr(page, "cname", .default = NA),
          custom_404 = purrr::map_lgl(page, "custom_404", .default = NA),
          https_enforced = purrr::map_lgl(page, "https_enforced", .default = NA)
        )
      }
    }
  )
}




github_api_pages_create = function(repo, build_type, branch, path) {
  arg_is_chr_scalar(repo, build_type, branch, path)

  ghclass_api_v3_req(
    endpoint = "POST /repos/:owner/:repo/pages",
    owner = get_repo_owner(repo),
    repo = get_repo_name(repo),
    build_type = build_type,
    source = list(
      branch = branch, path = path
    )
  )
}



#' @name pages
#' @rdname pages
#'
#' @param build_type Character. Either `"workflow"` or `"legacy"` - the former uses GitHub actions to
#' build and publish the site (requires a workflow file to achieve this).
#'
#' @param branch Character. Repository branch to publish.
#'
#' @param path Character. Repository path to publish.
#'
#' @export
#'
pages_create = function(
    repo,
    build_type = c("legacy", "workflow"),
    branch = "main",
    path = "/"
) {
  build_type = match.arg(build_type)
  arg_is_chr(repo)
  arg_is_chr_scalar(build_type, branch, path)

  purrr::map(
    repo,
    function(repo) {
      res = purrr::safely(github_api_pages_create)(repo, build_type, branch, path)

      status_msg(
        res,
        "Created Pages site for {.val {repo}}.",
        "Failed to create Pages site for repo {.val {repo}}."
      )

      res
    }
  ) %>%
    invisible()
}



github_api_pages_delete = function(repo) {
  arg_is_chr_scalar(repo)

  ghclass_api_v3_req(
    endpoint = "DELETE /repos/:owner/:repo/pages",
    owner = get_repo_owner(repo),
    repo = get_repo_name(repo)
  )
}



#' @name pages
#' @rdname pages
#'
#' @param prompt Logical. Should the user be prompted before deleting Pages sites. Default `TRUE`.
#'
#' @export
#'
pages_delete = function(repo, prompt = TRUE) {
  arg_is_chr(repo)
  arg_is_lgl_scalar(prompt)

  if (prompt) {
    delete = cli_yeah("This command will delete Pages sites for the following repositories permanently: {.val {repo}}.")
    if (!delete) {
      return(invisible())
    }
  }

  purrr::map(
    repo,
    function(repo) {
      res = purrr::safely(github_api_pages_delete)(repo)

      status_msg(
        res,
        "Deleted Pages site for repo {.val {repo}}.",
        "Failed to delete Pages site for repo {.val {repo}}."
      )

      res
    }
  ) %>%
    invisible()
}

#' @name github_api_limit
#' @rdname github_api_limit
#'
#' @title Tools for limiting gh's GitHub api requests.
#'
#' @description
#' * `github_get_api_limit()` - returns the current limit on results returned by gh.
#'
#' * `github_set_api_limit()` - sets a limit on results returned by gh.
#'
#' * `github_get_max_wait()` / `github_set_max_wait()` - get or set the maximum number
#' of seconds gh will wait when it hits a rate limit. `NULL` (the default) uses gh's
#' own default.
#'
#' * `github_get_max_rate()` / `github_set_max_rate()` - get or set the maximum request
#' rate, in requests per second, used to proactively throttle requests and avoid
#' secondary rate limits. `NULL` (the default) disables throttling.
#'
#' @param limit The maximum number of records to return from an API request.
#' @param max_wait The maximum number of seconds to wait when rate limited, or `NULL`
#' to use gh's default. Passed to [gh::gh()]'s `.max_wait`.
#' @param max_rate The maximum request rate in requests per second, or `NULL` for no
#' throttling. Passed to [gh::gh()]'s `.max_rate`.
#'
#' @details These values are stored in the `"ghclass.api.limit"`, `"ghclass.max.wait"`,
#' and `"ghclass.max.rate"` options globally.
#'
#' @return `github_get_api_limit()` returns a single integer value.
#'
#' `github_get_max_wait()` and `github_get_max_rate()` return their stored value or `NULL`.
#'
#' `github_set_*()` functions invisibly return the value of their argument.
#'
#' @examples
#' github_get_api_limit()
#'
#' github_set_api_limit(500)
#'
#' github_get_api_limit()
NULL



#' @rdname github_api_limit
#' @export
#'
github_get_api_limit = function() {
  getOption("ghclass.api.limit", 10000L)
}

#' @rdname github_api_limit
#' @export
#'
github_set_api_limit = function(limit = 10000L) {
  arg_is_pos_int(limit)
  options("ghclass.api.limit" = limit)

  invisible(limit)
}

#' @rdname github_api_limit
#' @export
#'
github_get_max_wait = function() {
  getOption("ghclass.max.wait", NULL)
}

#' @rdname github_api_limit
#' @export
#'
github_set_max_wait = function(max_wait = NULL) {
  arg_is_pos_int_scalar(max_wait, allow_null = TRUE)
  options("ghclass.max.wait" = max_wait)

  invisible(max_wait)
}

#' @rdname github_api_limit
#' @export
#'
github_get_max_rate = function() {
  getOption("ghclass.max.rate", NULL)
}

#' @rdname github_api_limit
#' @export
#'
github_set_max_rate = function(max_rate = NULL) {
  if (!is.null(max_rate) &&
      (!is.numeric(max_rate) || length(max_rate) != 1L || is.na(max_rate) || max_rate <= 0)) {
    cli_stop("{.arg max_rate} must be a single positive number (requests per second) or {.val NULL}.")
  }

  options("ghclass.max.rate" = max_rate)

  invisible(max_rate)
}

is_git_repo = function(path) {
  arg_is_chr_scalar(path)

  res = purrr::safely(gert::git_open)(path)
  succeeded(res)
}

# If we are given a single repo directory check if it is a repo or a directory of repos
repo_dir_helper = function(repo_dir) {

  exists = fs::dir_exists(repo_dir)
  if(!all(exists))
    cli_stop("Unable to locate the repo{?s}: {.val {repo_dir[!exists]}}.")

  if (length(repo_dir) == 1 && !is_git_repo(repo_dir[1])) {
    dir = fs::dir_ls(repo_dir, type="directory")
  } else {
    dir = repo_dir
  }

  fs::path_real(dir)
}

github_api_team_update = function(
  org, team_slug,
  name = NULL, description = NULL,
  privacy = NULL, permission = NULL, parent_team_id = NULL
) {
  ghclass_api_v3_req(
    endpoint = "PATCH /orgs/:org/teams/:team_slug",
    org = org,
    team_slug = team_slug,
    name = name,
    description = description,
    privacy = privacy,
    permission = permission,
    parent_team_id = parent_team_id
  )
}


#' @rdname team
#' @param new_team character, new team name.
#' @export
team_rename = function(org, team, new_team, team_type = c("name", "slug")) {
  arg_is_chr_scalar(org)
  arg_is_chr(team, new_team)
  team_type = match.arg(team_type)

  if (team_type == "name") {
    slug = team_slug_lookup(org, team)
  } else {
    slug = team
  }

  check_team_slug(slug)

  res = purrr::pmap(
    tibble::tibble(team, slug, new_team),
    function(team, slug, new_team) {
      if (is.na(slug)) {
        cli::cli_alert_danger("Team {.val {team}} does not exist.")
        return()
      }

      res = purrr::safely(github_api_team_update)(org, slug, name = new_team)

      status_msg(
        res,
        "Renamed team {.val {team}} to {.val {new_team}}.",
        "Failed to rename team {.val {team}} to {.val {new_team}}."
      )
    }
  )

  invisible(res)
}

github_api_repo = function(repo) {
  arg_is_chr_scalar(repo)

  ghclass_api_v3_req(
    endpoint = "GET /repos/:owner/:repo",
    owner = get_repo_owner(repo),
    repo = get_repo_name(repo)
  )
}

#' @rdname repo_details
#'
#' @param type Character. Clone url type, either "https" or "ssh".
#'
#' @export
#'
repo_clone_url = function(repo, type = c("https", "ssh")) {
  arg_is_chr(repo)
  type = match.arg(type)

  details = purrr::map(repo, purrr::safely(github_api_repo))

  if (type == "https") {
    lookup = c("result", "clone_url")
  } else {
    lookup = c("result", "ssh_url")
  }

  purrr::map_chr(details, lookup, .default = NA)
}

github_api_branch_get_ref = function(repo, branch) {
  ghclass_api_v3_req(
    endpoint = "GET /repos/:owner/:repo/commits/:ref",
    owner = get_repo_owner(repo),
    repo = get_repo_name(repo),
    ref = paste0("heads/", branch)
  )

}

get_branch_ref = function(repo, branch) {
  arg_is_chr_scalar(repo, branch)

  res = purrr::safely(github_api_branch_get_ref)(repo, branch)
  repo_txt = format_repo(repo, branch)
  if (failed(res))
    cli_stop("Unable to locate branch {.val {repo_txt}}.")

  result(res)
}

github_api_branch_create = function(repo, branch, new_branch) {
  head = get_branch_ref(repo, branch)

  ghclass_api_v3_req(
    "POST /repos/:owner/:repo/git/refs",
    owner = get_repo_owner(repo),
    repo = get_repo_name(repo),
    ref = paste0("refs/heads/", new_branch),
    sha = head[["sha"]]
  )
}





#' @rdname branch
#' @export
#'
branch_create = function(repo, branch, new_branch) {
  arg_is_chr(repo, branch, new_branch)

  invisible( purrr::pmap(
    list(repo, branch, new_branch),
    function(repo, branch, new_branch) {

      cur_repo = format_repo(repo, branch)
      new_repo = format_repo(repo, new_branch)

      branches = repo_branches(repo)

      if (!branch %in% branches) {
        cli::cli_alert_danger("Failed to create branch, {.val {cur_repo}} does not exist.")
        return()
      }

      if (new_branch %in% branches) {
        cli::cli_alert_danger("Skipping creation of branch {.val {new_repo}}, it already exists.")
        return()
      }

      res = purrr::safely(github_api_branch_create)(repo, branch, new_branch)

      status_msg(
        res,
        "Created branch {.val {new_branch}} in repo {.val {repo}}.",
        "Failed to create branch {.val {new_branch}} in repo {.val {repo}}."
      )

      res
    }
  ) )
}

github_api_issue_comment = function(repo, number, body) {
  arg_is_chr_scalar(repo, body)

  ghclass_api_v3_req(
    endpoint = "POST /repos/:owner/:repo/issues/:issue_number/comments",
    owner = get_repo_owner(repo),
    repo = get_repo_name(repo),
    issue_number = number,
    body = body
  )
}


#' @rdname issue
#'
#' @param comment Character. Optional comment to post on the issue before
#' closing it. If posting the comment fails the issue is not closed. Recycled
#' against `repo` / `number`.
#'
#' @export
#'
issue_close = function(repo, number, comment = NULL) {

  arg_is_chr(repo)
  arg_is_chr(comment, allow_null = TRUE)
  arg_is_pos_int(number)

  if (is.null(comment))
    comment = list(NULL)

  res = purrr::pmap(
    tibble::tibble(repo, number, comment),
    function(repo, number, comment) {
      num_text = paste0("#", number)

      if (!is.null(comment)) {
        comment_res = purrr::safely(github_api_issue_comment)(
          repo, number, comment
        )

        status_msg(
          comment_res,
          "Commented on issue {.val {num_text}} for repo {.val {repo}}.",
          "Failed to comment on issue {.val {num_text}} for repo {.val {repo}}."
        )

        if (failed(comment_res))
          return(comment_res)
      }

      close_res = purrr::safely(github_api_issue_edit)(
        repo, number, state = "closed"
      )

      status_msg(
        close_res,
        "Closed issue {.val {num_text}} for repo {.val {repo}}.",
        "Failed to close issue {.val {num_text}} for repo {.val {repo}}."
      )

      close_res
    }
  )

  invisible(res)
}

github_api_v4_graphql = function(query, vars = list(), max_retries = 3L,
                                 max_wait = github_get_max_wait()) {
  arg_is_chr_scalar(query)

  query = graphql_glue(query, vars)
  cap = if (is.null(max_wait)) 600 else max_wait

  attempt = 0
  repeat {
    req = httr::POST(
      "https://api.github.com/graphql",
      httr::add_headers(
        Authorization = paste("bearer", github_get_token())
      ),
      encode = "json",
      body = list(query = query)
    )

    res = httr::content(req)
    code = httr::status_code(req)

    wait = graphql_rate_limit_wait(req, res, code)

    if (is.null(wait) || attempt >= max_retries)
      break

    attempt = attempt + 1
    wait = min(max(wait, 0), cap)
    cli::cli_alert_info(
      "GitHub API v4 rate limit reached; waiting {round(wait)}s before retry {attempt}/{max_retries}."
    )
    Sys.sleep(wait)
  }

  if (code >= 300) {
    cli_stop("GitHub API v4 error code ({code}) - {res[['message']]}")
  }

  if (!is.null(res$errors)) {
    msgs = unlist(purrr::map(res$errors, "message"))
    cli_stop(paste(msgs, collapse = "\n"))
  }

  res
}

# Number of seconds to wait before retrying a rate-limited v4 request, or NULL
# if the response is not a rate-limit response. GitHub signals GraphQL rate
# limits either with a 403/429 status (secondary limits, usually carrying a
# Retry-After header) or a 200 with a RATE_LIMITED error (the points budget,
# with x-ratelimit-reset).
graphql_rate_limit_wait = function(req, res, code) {
  err_types = purrr::map_chr(res[["errors"]], "type", .default = NA_character_)
  rate_limited = code %in% c(403L, 429L) || any(err_types == "RATE_LIMITED", na.rm = TRUE)

  if (!rate_limited)
    return(NULL)

  hdr = httr::headers(req)

  retry_after = hdr[["retry-after"]]
  if (!is.null(retry_after))
    return(as.numeric(retry_after))

  remaining = hdr[["x-ratelimit-remaining"]]
  reset = hdr[["x-ratelimit-reset"]]
  if (!is.null(remaining) && remaining == "0" && !is.null(reset))
    return(as.numeric(reset) - as.numeric(Sys.time()))

  5
}

github_api_v4_graphql_paginated = function(query, page_info, cursor_var = "cursor") {
  arg_is_chr_scalar(query)
  arg_is_chr(page_info)

  vars = list()
  vars[cursor_var] = list(NULL)

  page_info = as.list(page_info)
  if (page_info[[1]] != "data")
    page_info = c("data",page_info)
  if (page_info[[length(page_info)]] != "pageInfo")
    page_info = c(page_info, "pageInfo")


  res = list()
  i = 1

  repeat {
    res[[i]] = github_api_v4_graphql(query, vars)
    page = purrr::pluck(res[[i]], !!!page_info)

    if (is.null(page)) {
      cli_stop("Unable to locate page info for this query.")
    }

    if (!page$hasNextPage)
      break

    vars[[cursor_var]] = page$endCursor
    i = i + 1
  }

  res
}


graphql_glue = function(query, vars) {
  glue::glue_data(vars, query, .open = "<", .close = ">")
}

graphql_quote = function(x) {
  if (is.null(x)) {
    "null"
  } else if (is.na(x)) {
    cli_stop("{.val NA} values are not supported by graphql.")
  } else {
    glue::double_quote(x)
  }
}

graphql_escape = function(x) {
  if (is.null(x))
    return(x)

  x = gsub("\\", "\\\\", x, fixed = TRUE)
  gsub('"', '\\"', x, fixed = TRUE)
}

# Magical ... and substitute voodoo based on
# http://adv-r.had.co.nz/Computing-on-the-language.html#substitute

handle_arg_list = function(..., tests) {
  values = list(...)
  names = eval(substitute(alist(...)))
  names = purrr::map_chr(names, ~ paste(deparse(.x), collapse = " "))

  purrr::walk2(names, values, tests)
}

#' @rdname ghclass-internal
#' @export
arg_is_chr_scalar = function(..., allow_null = FALSE, allow_na = FALSE) {
  arg_is_chr(..., allow_null = allow_null, allow_na = allow_na)
  arg_is_scalar(..., allow_null = allow_null, allow_na = allow_na)
}

#' @rdname ghclass-internal
#' @export
arg_is_lgl_scalar = function(..., allow_null = FALSE, allow_na = FALSE) {
  arg_is_lgl(..., allow_null = allow_null, allow_na = allow_na)
  arg_is_scalar(..., allow_null = allow_null, allow_na = allow_na)
}

#' @rdname ghclass-internal
#' @export
arg_is_pos_int_scalar = function(..., allow_null = FALSE, allow_na = FALSE) {
  arg_is_pos_int(..., allow_null = allow_null)
  arg_is_scalar(..., allow_null = allow_null, allow_na = allow_na)
}

#' @rdname ghclass-internal
#' @export
arg_is_chr = function(..., allow_null = FALSE, allow_na = FALSE, allow_empty = TRUE) {
  handle_arg_list(
    ...,
    tests = function(name, value) {
      if (!(is.character(value) | (is.null(value) & allow_null)))
        cli_stop("Argument {.val {name}} must be of character type.")

      if (any(is.na(value)) & !allow_na)
        cli_stop("Argument {.val {name}} must not contain any missing values ({.val {NA}}).")

      if (length(value) == 0 & !allow_empty)
        cli_stop("Argument {.val {name}} must have length >= 1.")
    }
  )
}

#' @rdname ghclass-internal
#' @export
arg_is_lgl = function(..., allow_null = FALSE, allow_na = FALSE, allow_empty = TRUE) {
  handle_arg_list(
    ...,
    tests = function(name, value) {
      if (!(is.logical(value) | (is.null(value) & allow_null)))
        cli_stop("Argument {.val {name}} must be of logical type.")

      if (any(is.na(value)) & !allow_na)
        cli_stop("Argument {.val {name}} must not contain any missing values ({.val {NA}}).")

      if (length(value) == 0 & !allow_empty)
        cli_stop("Argument {.val {name}} must have length >= 1.")
    }
  )
}

arg_is_scalar = function(...,  allow_null = FALSE, allow_na = FALSE) {
  handle_arg_list(
    ...,
    tests = function(name, value) {
      if (length(value) > 1 | (!allow_null & length(value) == 0)) {
        cli_stop("Argument {.val {name}} must be of length 1.")
      }

      if (!is.null(value)) {
        if (is.na(value) & !allow_na)
          cli_stop("Argument {.val {name}} must not be a missing value ({.val {NA}}).")
      }
    }
  )
}

arg_is_raw = function(..., allow_null = FALSE) {
  handle_arg_list(
    ...,
    tests = function(name, value) {
      if (!(is.raw(value) | (is.null(value) & allow_null)))
        cli_stop("Argument {.val {name}} must be of type raw.")
    }
  )
}

#' @rdname ghclass-internal
#' @export
arg_is_pos_int = function(..., allow_null = FALSE) {
  handle_arg_list(
    ...,
    tests = function(name, value) {
      if (is.null(value)) {
        if (!allow_null)
          cli_stop("Argument {.val {name}} must be a whole positive non-zero number.")
      } else if (!(is.numeric(value) && all(value > 0 & value %% 1 == 0))) {
        cli_stop("Argument {.val {name}} must be a whole positive non-zero number.")
      }
    }
  )
}

arg_is_df = function(..., allow_null = FALSE, allow_empty = FALSE) {
  handle_arg_list(
    ...,
    tests = function(name, value) {
      if (!inherits(value, "data.frame") | (is.null(value) & allow_null))
        cli_stop("Argument {.val {name}} must be a data frame.")
      if (!is.null(value)) {
        if (nrow(value) == 0 & !allow_empty)
          cli_stop("Argument {.val {name}} must be a data frame with nrows >= 1.")
        if (ncol(value) == 0)
          cli_stop("Argument {.val {name}} must be a data frame with ncols >= 1.")
      }
    }
  )
}

# Remove org and team membership
github_api_org_remove = function(org, user) {
  ghclass_api_v3_req(
    endpoint = "DELETE /orgs/:org/members/:username",
    org = org,
    username = user
  )
}

# Only remove membership
github_api_org_remove_membership = function(org, user) {
  ghclass_api_v3_req(
    "DELETE /orgs/:org/memberships/:username",
    org = org,
    username = user
  )
}

# TODO - think about removing (or not) team membership

#' @rdname org_members
#' @param prompt Logical. Prompt before removing member from organization. Default `TRUE`.
#' @export
#'
org_remove = function(org, user, prompt = TRUE) {

  arg_is_chr_scalar(org)
  arg_is_chr(user, allow_null = TRUE)
  arg_is_lgl_scalar(prompt)

  if (prompt) {
    delete = cli_yeah(
      "Org {.val {org}} will have the following {cli::qty(user)} user{?s} removed: ",
      "{.val {user}}"
    )
    if (!delete) {
      return(invisible())
    }
  }

  pending = user %in% org_pending(org)

  res = purrr::map2(
    user, pending,
    function(user, pending) {
      if (pending)
        res = purrr::safely(github_api_org_remove_membership)(org, user)
      else
        res = purrr::safely(github_api_org_remove)(org, user)

      type = ternary(pending, "pending user", "user")
      status_msg(
        res,
        "Removed {type} {.val {user}} from org {.val {org}}.",
        "Failed to remove {type} {.val {user}} from org {.val {org}}."
      )
    }
  )

  invisible(res)
}

#' @title Create a team or individual assignment
#'
#' @description
#' This is a higher level function that combines the following steps:
#'
#' * Create repos
#' * Create teams and invite students if necessary
#' * Add teams or individuals to the repositories
#' * Mirror a template repository to assignment repositories
#'
#' @param org Character. Name of the GitHub organization.
#' @param repo Character. Name of the repo(s) for the assignment.
#' @param user Character. GitHub username(s).
#' @param team Character. Team names, if not provided an individual assignment will be created.
#' @param source_repo Character. Address of the repository to use as a template for all created repos.
#' @param private Logical. Should the created repositories be private. Default `TRUE`.
#' @param add_badges Logical. Should GitHub action badges be added to the README. Default `FALSE`.
#' @param ignore_existing Logical. If `TRUE`, any requested repos that already
#' exist are skipped (with a warning) and the remaining repos are still
#' created; useful for resuming a partially failed run. Default `FALSE`.
#'
#' @return An invisible list containing the results of each step.
#'
#' @export
#'

org_create_assignment = function(org, repo, user, team = NULL, source_repo = NULL,
                                 private = TRUE, add_badges = FALSE,
                                 ignore_existing = FALSE) {

  arg_is_chr_scalar(org)
  arg_is_chr(repo, user)
  arg_is_chr(team, allow_null = TRUE)
  arg_is_chr_scalar(source_repo, allow_null = TRUE)
  arg_is_lgl_scalar(private, add_badges, ignore_existing)

  if (length(repo) != length(user))
    cli_stop("{.arg repo} and {.arg user} must have the same length.")
  if (!is.null(team) && length(team) != length(user))
    cli_stop("{.arg team} and {.arg user} must have the same length.")

  repo_full = paste0(org, "/", repo)

  if (!is.null(source_repo) && !repo_is_template(source_repo)) {
    cli_stop(
      "{.val {source_repo}} is not a template repo - ",
      "this can be set using {.code repo_set_template({.val {source_repo}})}."
    )
  }

  existing = repo_exists(repo_full)
  if (any(existing)) {
    if (!ignore_existing) {
      cli_stop(
        "The following repo{?s} already exist{?s/}: {.val {repo_full[existing]}}. ",
        "Delete these repo{?s} or use an alternative method to create the assignment, ",
        "or set {.code ignore_existing = TRUE} to skip them."
      )
    }

    cli::cli_warn(
      "Skipping {sum(existing)} existing repo{?s}: {.val {repo_full[existing]}}."
    )

    if (all(existing))
      return(invisible(list()))

    repo = repo[!existing]
    user = user[!existing]
    if (!is.null(team))
      team = team[!existing]
    repo_full = paste0(org, "/", repo)
  }

  res = list()

  if (!is.null(source_repo)) {
    res[["mirror"]] = repo_mirror_template(source_repo, repo_full, private = private)
  } else {
    res[["create"]] = repo_create(org, repo, private = private)
  }

  if (!is.null(team)) {
    # Assume team assignment
    res[["team_create"]] = team_create(org, unique(team))
    res[["team_invite"]] = team_invite(org, user, team)
    res[["team_add"]] = repo_add_team(repo_full, team)
  } else {
    # Assume individual assignment
    res[["user_invite"]] = repo_add_user(repo_full, user)
  }

  if (add_badges)
    res[["add_badge"]] = action_add_badge(unique(repo_full))

  invisible(res)
}

github_api_action_run_usage = function(repo, run_id) {
  arg_is_chr_scalar(repo)
  arg_is_pos_int_scalar(run_id)

  ghclass_api_v3_req(
    endpoint = "GET /repos/{owner}/{repo}/actions/runs/{run_id}/timing",
    owner = get_repo_owner(repo),
    repo = get_repo_name(repo),
    run_id = run_id
  )
}

#' @name action
#' @rdname action
#'
#' @export
#'
action_runtime = function(
    repo,
    branch = NULL,
    event = NULL,
    status = NULL,
    created = NULL,
    limit = 1
) {
  d = action_runs(repo = repo, branch = branch, event = event,
                  status = status, created = created, limit = limit)


  get_run_dur = function(repo, run_id) {
    res = purrr::safely(github_api_action_run_usage)(repo, run_id)

    status_msg(
      res,
      fail = "Failed to retrieve workflow runs for repo {.val {repo}}."
    )

    run_dur = result(res)$run_duration_ms
    if (is.null(run_dur))
      run_dur = NA

    run_dur
  }

  dplyr::mutate(
    d,
    run_dur = purrr::map2_dbl(.data$repo, .data$run_id, get_run_dur),
    run_dur = lubridate::duration(.data$run_dur/1000)
  )
}

github_api_org_update = function(
  org,
  default_repository_permission = NULL,
  members_can_create_repositories = NULL,
  members_can_create_private_repositories = NULL,
  members_can_create_public_repositories = NULL
) {
  ghclass_api_v3_req(
    endpoint = "PATCH /orgs/:org",
    org = org,
    default_repository_permission = default_repository_permission,
    members_can_create_repositories = members_can_create_repositories,
    members_can_create_private_repositories = members_can_create_private_repositories,
    members_can_create_public_repositories  = members_can_create_public_repositories,
    .send_headers = c(Accept = "application/vnd.github.surtur-preview+json")
  )
}

#' @rdname org_perm
#' @export
#'
org_set_repo_permission = function(org, repo_permission = c("none", "read", "write", "admin")) {
  arg_is_chr_scalar(org)
  repo_permission = match.arg(repo_permission)

  res = purrr::safely(github_api_org_update)(org, default_repository_permission = repo_permission)

  status_msg(
    res,
    "Set org {.val {org}}'s repo permissions to {.val {repo_permission}}.",
    "failed to set org {.val {org}}'s repo permissions."
  )

  invisible(result(res))
}

#' @rdname action_badge
#'
#' @note `action_add_badge()` is not idempotent; re-running it adds another badge
#' block, so use `action_remove_badge()` first if you need to avoid duplicates.
#'
#' @examples
#' \dontrun{
#' action_add_badge("ghclass-test/hw1-team01")
#'
#' action_remove_badge("ghclass-test/hw1-team01")
#' }
#'
#' @export
#'
action_add_badge = function(repo, workflow = NULL, where = "^.",
                            line_padding = "\n\n\n", file = "README.md") {
  arg_is_chr(repo)
  arg_is_chr(workflow, allow_null=TRUE)
  arg_is_chr_scalar(where, line_padding, file)

  if (is.null(workflow)) {
    d = purrr::map_dfr(
      repo,
      ~ tibble::tibble(
        repo = .x,
        workflow = action_workflows(.x)[["name"]]
      )
    )
  } else {
    d = tibble::tibble(
      repo = repo,
      workflow = workflow
    )
  }

  d[["url"]] =  glue::glue_data(d, "https://github.com/{repo}/workflows/{workflow}/badge.svg")
  d[["dest"]] = glue::glue_data(d, "https://github.com/{repo}/actions?query=workflow:\"{workflow}\"")
  d[["link"]] = glue::glue_data(d, "[![{workflow}]({url_encode(url)})]({url_encode(dest)})")

  # Collapse by repo to save multiple changes to a single file
  d = dplyr::group_by(d, repo) %>%
    dplyr::summarize(
      link = paste0(paste(.data$link, collapse = " "), line_padding),
      workflows = list(workflow)
    )

  res = purrr::pmap(
    d,
    function(repo, link, workflows) {
      repo_modify_file(
        repo = repo, path = file,
        pattern = where, content = link,
        method = "before"
      )
    }
  )

  invisible(res)
}


#' @rdname action_badge
#' @export
#'
action_remove_badge = function(repo, workflow_pat = ".*?", file = "README.md") {
  arg_is_chr(repo, workflow_pat)
  arg_is_chr_scalar(file)

  res = purrr::map2(
    repo, workflow_pat,
    function(repo, workflow_pat) {
      pattern = glue::glue(
        "\\[!\\[{workflow_pat}\\]\\(.*?\\)\\]\\(https://github.com/.*?/actions.*?\\)\\s*"
      )

      repo_modify_file(
        repo = repo, path = file,
        pattern = pattern, content = "",
        method = "replace", all = TRUE
      )
    }
  )

  invisible(res)
}

#' @name branch
#' @rdname branch
#'
#' @title Create and delete branches in a repository
#'
#' @description
#' * `branch_create()` - creates a new branch from an existing GitHub repo.
#'
#' * `branch_delete()` - deletes a branch from an existing GitHub repo.
#'
#' * `branch_remove()` - previous name of `branch_delete`, deprecated.
#'
#' @param repo GitHub repository address in `owner/repo` format.
#' @param branch Repository branch to use.
#' @param new_branch Name of branch to create.
#'
#' @return `branch_create()` and `branch_delete()` invisibly return a list containing
#' the results of the relevant GitHub API call.
#'
#' @seealso [repo_branches]
#'
#' @examples
#' \dontrun{
#' repo_create("ghclass-test", "test_branch", auto_init=TRUE)
#'
#' branch_create("ghclass-test/test_branch", branch = "main", new_branch = "test")
#' repo_branches("ghclass-test/test_branch")
#'
#' branch_delete("ghclass-test/test_branch", branch="test")
#' repo_branches("ghclass-test/test_branch")
#'
#' repo_delete("ghclass-test/test_branch", prompt = FALSE)
#' }
#'
NULL

purge_all = function(org) {
  repos = org_repos(org)
  teams = org_teams(org)
  users = c( org_members(org, include_admins = FALSE), org_pending(org))

  if (length(repos) != 0)
    repo_delete(repos, prompt = TRUE)

  if (length(teams) != 0)
    team_delete(org, teams, prompt = TRUE)

  if (length(users) != 0)
    org_remove(org, users, prompt = TRUE)
}

github_api_org_admins = function(owner){
  ghclass_api_v3_req(
    endpoint = "GET /orgs/:owner/members",
    owner = owner,
    role = "admin"
  )
}

#' @rdname org_members
#' @export
#'
org_admins = function(org) {
  arg_is_chr_scalar(org)

  res = purrr::safely(github_api_org_admins)(owner = org)

  if (succeeded(res))
    return(purrr::map_chr(result(res), "login"))

  # A user account has no org-members endpoint, so the call above fails and the
  # user is their own sole admin. Use user_type() rather than user_exists()
  # (which is also TRUE for organizations) so a transient failure on a real org
  # errors instead of being misreported as a user.
  if (identical(user_type(org), "User"))
    return(org)

  cli_stop("Failed to retrieve admins for org {.val {org}}.")
}

github_api_action_workflows = function(repo) {
  ghclass_api_v3_req(
    endpoint = "GET /repos/:owner/:repo/actions/workflows",
    owner = get_repo_owner(repo),
    repo = get_repo_name(repo)
  )
}

#' @name action
#' @rdname action
#'
#' @param repo Character. Address of repository in `owner/name` format.
#' @param full Logical. Should all workflow columns be returned. Default `FALSE`.
#'
#' @export
#'
action_workflows = function(repo, full = FALSE) {
  arg_is_chr_scalar(repo)

  res = purrr::safely(github_api_action_workflows)(repo)

  status_msg(
    res,
    fail = "Failed to retrieve workflows for repo {.val {repo}}."
  )

  if (failed(res) || empty_result(res) || result(res)[["total_count"]] == 0) {
    d = tibble::tibble(
      id         = integer(),
      node_id    = character(),
      name       = character(),
      path       = character(),
      state      = character(),
      created_at = character(),
      updated_at = character(),
      url        = character(),
      html_url   = character(),
      badge_url  = character()
    )
  } else {
    d = purrr::map_dfr(result(res)[["workflows"]], ~ .x)
  }

  if (!full)
    d = dplyr::select(d, "name", "id", "path", "state", "badge_url")

  d
}

#' @rdname local_repo
#' @export
local_repo_branch = function(repo_dir, branch) {
  require_gert()
  arg_is_chr(repo_dir)
  arg_is_chr_scalar(branch)

  repo_dir = repo_dir_helper(repo_dir)

  res = purrr::map(
    repo_dir,
    function(dir) {
      res = purrr::safely(gert::git_branch_create)(
        name = branch, repo = dir
      )

      repo = fs::path_file(dir)
      status_msg(
        res,
        "Added branch {.val {branch}} to {.val {repo}}.",
        "Failed to add branch {.val {branch}} to {.val {repo}}."
      )

      res
    }
  )

  invisible(res)
}

github_api_user_repos = function(owner, type) {
  arg_is_chr_scalar(owner)

  ghclass_api_v3_req(
    endpoint = "GET /users/:owner/repos",
    owner = owner,
    type = type
  )
}

github_api_your_repos = function(type) {
  ghclass_api_v3_req(
    endpoint = "GET /user/repos",
    type = type
  )
}

#' @rdname user
#' @export
#'
user_repos = function(user, type = c("owner", "all", "public", "private", "member"),
                      filter = NULL, exclude = FALSE, full_repo = TRUE) {

  type = match.arg(type)

  arg_is_chr_scalar(user, type)
  arg_is_chr_scalar(filter, allow_null = TRUE)
  arg_is_lgl_scalar(exclude, full_repo)

  me = github_whoami(quiet = TRUE)

  res = purrr::safely(
    function() {
      if (!is.null(me) && tolower(user) == tolower(me)) {
        github_api_your_repos(type)
      } else {
        github_api_user_repos(user, type)
      }
    }
  )()

  status_msg(res, fail = "Failed to retrieve repos for user {.val {user}}.")

  if (failed(res))
    return(invisible(NULL))

  if (full_repo) {
    res = purrr::map_chr(result(res), "full_name")
  } else {
    res = purrr::map_chr(result(res), "name")
  }

  filter_results(res, filter, exclude)
}

github_api_whoami = function() {
  ghclass_api_v3_req(
    endpoint = "GET /user"
  )
}

#' Returns the login of the authenticated user (based on the current PAT).
#'
#' @param quiet Logical. Should status messages be shown. Default `FALSE`.
#'
#' @return Character value containing user login.
#'
#' @examples
#' \dontrun{
#' github_whoami()
#' }
#'
#' @export
#'
github_whoami = function(quiet = FALSE) {
  res = purrr::safely(github_api_whoami)()

  if (!quiet) {
    status_msg(
      res,
      fail = "Failed to retrieve whoami results."
    )
  }

  if (failed(res))
    return(invisible(NULL))

  result(res)[["login"]]
}



github_api_orgs = function() {
  ghclass_api_v3_req(
    endpoint = "GET /user/memberships/orgs"
  )
}

#' Collect details on the authenticated user's GitHub organization memberships
#' (based on the current PAT).
#'
#' @param quiet Logical. Should status messages be shown. Default `FALSE`.
#'
#' @return Returns a tibble with organization details.
#'
#'
#' @examples
#' \dontrun{
#' github_orgs()
#' }
#'
#' @export
#'
github_orgs = function(quiet = FALSE) {
  res = purrr::safely(github_api_orgs)()

  if (!quiet) {
    status_msg(
      res,
      fail = "Failed to retrieve authenticated user's organization memberships."
    )
  }

  res = result(res)

  if (empty_result(res))
    return(tibble::tibble(
      org = character(),
      role = character(),
      state = character()
    ))

  tibble::tibble(
    org = purrr::map_chr(res, c("organization", "login")),
    role = purrr::map_chr(res, "role"),
    state = purrr::map_chr(res, "state")
  )
}

#' @name local_repo
#' @rdname local_repo
#'
#' @title Functions for managing local git repositories
#'
#' @description
#'
#' * `local_repo_clone()` - Clones a GitHub repository to a local directory.
#'
#' * `local_repo_add()` - Equivalent to `git add` - stages a file in a local repository.
#'
#' * `local_repo_commit()` - Equivalent to `git commit` - commits staged files in a local repository.
#'
#' * `local_repo_push()` - Equivalent to `git push` - push a local repository.
#'
#' * `local_repo_pull()` - Equivalent to `git pull` - pull a local repository.
#'
#' * `local_repo_branch()` - Equivalent to `git branch` - create a branch in a local repository.
#'
#' * `local_repo_log()` - Equivalent to `git log` - returns a data frame for git log entries.
#'
#' @param repo_dir Vector of repo directories or a single directory containing one or more repos.
#' @param files Files to be staged
#' @param message Commit message
#' @param remote Repository remote to use.
#' @param branch Repository branch to use.
#' @param mirror Equivalent to `--mirror`
#' @param force Force push?
#' @param prompt Prompt before force push?
#' @param verbose Display verbose output.
#'
#' @return `local_repo_clone()` invisibly returns a character vector of paths for
#' the local repo directories.
#'
#' `local_repo_log()` returns a tibble containing repository details.`
#'
#' All other functions invisibly return a list containing the results of the relevant call
#' to `gert`.
#'
#' @details All `local_repo_*` functions depend on the `gert` library being installed.
#'
#' @examples
#' \dontrun{
#' repo = repo_create("ghclass-test", "local_repo_test")
#'
#' dir = file.path(tempdir(), "repos")
#' local_repo = local_repo_clone(repo, dir)
#'
#' local_repo_log(dir)
#'
#' # Make a local change and push
#' writeLines("Hello World", file.path(local_repo, "hello.txt"))
#'
#' local_repo_add(local_repo, "hello.txt")
#'
#' local_repo_commit(local_repo, "Added hello world")
#'
#' local_repo_push(local_repo)
#'
#' repo_commits(repo)
#'
#' # Pulling remote changes
#' repo_modify_file(repo, "hello.txt", pattern = ".*",  content = "!!!", method = "after")
#'
#' local_repo_pull(local_repo)
#'
#' local_repo_log(dir)
#'
#' repo_delete("ghclass-test/local_repo_test", prompt=FALSE)
#' }
#'
NULL

github_api_download_run_logs = function(repo, run_id, dest) {
  arg_is_chr_scalar(repo)
  run_id = as.double(run_id)

  if (missing(dest))
    dest = tempfile(fileext = ".zip")

  ghclass_api_v3_req(
    endpoint = "GET /repos/{owner}/{repo}/actions/runs/{run_id}/logs",
    owner = get_repo_owner(repo),
    repo = get_repo_name(repo),
    run_id = run_id,
    .destfile = dest,
    .overwrite = TRUE
  )

  dest
}


#' @name action
#' @rdname action
#'
#' @param run_ids Integer or data frame. Run ids for which to download logs.
#' If a data frame is passed then the `run_id` column will be used.
#' Defaults to the most recent run for each repo (i.e. `action_runs(repo)`).
#' @param keep_zip Logical. Should the log zips be saved (`TRUE`) or their contents extracted (`FALSE`). Default `FALSE`.
#' @param overwrite Logical. Should existing files be overwritten. Default `FALSE`.
#'
#' @export
#'
action_run_logs = function(
  repo, dir, run_ids = action_runs(repo),
  keep_zip = FALSE, overwrite = FALSE
) {
  arg_is_chr(repo)
  arg_is_chr_scalar(dir)
  arg_is_lgl_scalar(keep_zip, overwrite)

  dir.create(dir, showWarnings = FALSE, recursive = TRUE)

  if (is.numeric(run_ids))
    run_ids = tibble::tibble(repo = repo, run_id = run_ids)
  arg_is_df(run_ids)

  if (nrow(run_ids) == 0)
    cli_stop("No action runs available for the given repos.")

  df = dplyr::left_join(
    tibble::tibble(repo = repo),
    run_ids,
    by = "repo"
  ) %>%
    dplyr::select("repo", "run_id")

  res = purrr::pmap_chr(
    df,
    function(repo, run_id) {
      res = purrr::safely(github_api_download_run_logs)(repo, run_id)
      file = result(res)

      if (failed(res)) {
        cli::cli_alert_danger(
          "Failed to download logs for run {.val {run_id}} from repo {.val {repo}}.",
          wrap = FALSE
        )
        return(NA_character_)
      }

      if (keep_zip) {
        dest_path = fs::path_norm(glue::glue("{dir}/{get_repo_name(repo)}_{run_id}.zip"))

        if (file.exists(dest_path) & !overwrite) {
          cli::cli_alert_danger(
            "File {.file {dest_path}} already exists, set {.code overwrite = TRUE} to overwrite this file.",
            wrap = FALSE
          )
          return(NA_character_)
        }

        file.rename(file, dest_path)
      } else {
        dest_path = fs::path_norm(glue::glue("{dir}/{get_repo_name(repo)}_{run_id}"))

        if (dir.exists(dest_path) & !overwrite) {
          cli::cli_alert_danger(
            "Directory {.file {dest_path}} already exists, set {.code overwrite = TRUE} to overwrite.",
            wrap = FALSE
          )
          return(NA_character_)
        }

        dir.create(dest_path, showWarnings = FALSE, recursive = TRUE)
        utils::unzip(file, exdir = dest_path, overwrite = overwrite)
        file.remove(file)
      }

      status_msg(
        res,
        "Downloaded logs for run {.val {run_id}} from repo {.val {repo}} to {.val {dest_path}}.",
        NULL
      )

      dest_path
    }
  )

  invisible(res)
}

github_api_repo_delete = function(repo) {
  ghclass_api_v3_req(
    endpoint = "DELETE /repos/:owner/:repo",
    owner = get_repo_owner(repo),
    repo = get_repo_name(repo)
  )
}

#' @rdname repo_core
#' @param prompt Logical. Should the user be prompted before deleting repositories. Default `true`.
#' @export
#'
repo_delete = function(repo, prompt = TRUE) {

  arg_is_chr(repo, allow_null = TRUE)
  arg_is_lgl_scalar(prompt)

  if (prompt) {
    delete = cli_yeah("This command will delete the following repositories permanently: {.val {repo}}.")
    if (!delete) {
      return(invisible())
    }
  }

  res = purrr::map(
    repo,
    function(repo) {
      res = purrr::safely(github_api_repo_delete)(repo)

      status_msg(
        res,
        "Deleted repo {.val {repo}}.",
        "Failed to delete repo {.val {repo}}."
      )
    }
  )

  invisible(res)
}

#' Style repository with styler
#'
#' * `repo_style` implements "non-invasive pretty-printing of R source code" of .R or .Rmd files
#' within a repository using the `styler` package and adhering to `tidyverse` formatting guidelines.
#'
#' @param repo Character. Address of repository in "owner/name" format.
#' @param files Character or vector of characters. Names of .R and/or .Rmd files that styler should be applied to.
#' @param branch Character. Name of new branch to be created or overwritten. Default is "styler".
#' @param base Character. Name of branch that contains the .R and/or .Rmd files to be styled
#' @param create_pull_request Logical. If TRUE, a pull request is created from branch to base. Default `TRUE`.
#' @param draft Logical. Should the pull request be created as a draft pull request? (Draft PRs cannot be merged
#'   until allowed by the author) Default `TRUE`.
#' @param tag_collaborators Logical. If TRUE, a message with the repository collaborators is displayed. Default `TRUE`.
#' @param prompt Logical. Prompt the user before overwriting an existing branch. Default `TRUE`.
#'
#' @return The functions returns NULL invisibly.
#'
#' @examples
#' \dontrun{
#' repo_style("ghclass-test/hw1-team01", base = "main")
#' }
#'
#' @export
#'
repo_style = function(repo, files = c("*.R", "*.Rmd"), branch = "styler", base,
                      create_pull_request = TRUE, draft = TRUE, tag_collaborators = TRUE,
                      prompt = TRUE) {
  require_styler()
  require_gert()

  arg_is_chr(repo, files, branch, base)
  arg_is_lgl(draft)
  arg_is_lgl_scalar(create_pull_request, tag_collaborators)

  dir = file.path(tempdir(),"styler")

  # Make sure the directory is empty
  unlink(dir, recursive = TRUE)
  dir.create(dir, showWarnings = FALSE, recursive = TRUE)

  purrr::pwalk(
    list(repo, base, branch, draft),
    function(repo, base, branch, draft) {
      withr::local_dir(dir)
      path = get_repo_name(repo)
      unlink(path, recursive = TRUE)

      return_on_any_failed( local_repo_clone(repo, local_path = dir, branch = base) )
      return_on_any_failed( local_repo_branch(path, branch = branch) )

      #branch_exists = paste0("origin/",branch) %in% gert::git_branch_list(path)[["name"]]

      file_paths = unlist(
        purrr::map(files, ~ fs::dir_ls(path, recurse = TRUE, glob = .x)),
        use.names = FALSE
      )

      if (length(file_paths) == 0) {
        cli::cli_alert_danger("Found 0 files with the glob {.val {files}} in repo {.val {repo}}")
        return()
      }

      output = utils::capture.output( styler::style_file(file_paths) )
      msg = c("Results of running styler:\n", output)
      msg = paste(msg, collapse = "\n")

      # Check if styler didnt fix anything
      status = gert::git_status(path)
      if (nrow(status) == 0) {
        cli::cli_alert_info("No changes were suggested by styler")
        return()
      }

      return_on_any_failed( local_repo_add(path) )
      return_on_any_failed( local_repo_commit(path, message = msg) )
      return_on_any_failed( local_repo_push(path, branch = branch, force = TRUE, prompt = prompt) )

      if (create_pull_request) {
        msg = paste(c(
          "This pull request contains the results of running the automated R code formating tool styler ",
          "on your repo. Styling is based on the [tidyverse R style guide](http://style.tidyverse.org).\n",
          "\n",
          "Click on the commit below to see details of recommended changes. It is not necessary that your ",
          "code cleanly pass these checks, but if there is a large number of significant changes suggested ",
          "you should review the style guide with an eye towards potentially improving your code formatting."
        ), collapse = "")

        #if (tag_collaborators) {
        #  users = repo_collaborators(repo, include_admins = FALSE)[["username"]]
        #  if (length(users) > 0)
        #    msg = paste0(msg,"\n\n", paste0("@", users, collapse = ", "))
        #}

        open_prs = repo_prs(repo, state="open")
        pr_already_open = any(  (base == open_prs[["base_ref"]])
                              & (branch == open_prs[["head_ref"]]) )

        if (pr_already_open) {
          details = cli_glue("{repo} ({base} <- {branch})")

          cli::cli_alert_info(
            "Pull request {details} already exists. New styler changes should be reflected in this PR."
          )
        } else {
          pr_create(
            repo, title = "styler revisions",
            base = base, head = branch,
            body = paste0(msg, collapse = "\n"),
            draft = draft
          )
        }
      }
    }
  )

  invisible(NULL)
}

github_api_action_artifacts = function(repo) {
  arg_is_chr_scalar(repo)

  ghclass_api_v3_req(
    endpoint = "GET /repos/{owner}/{repo}/actions/artifacts",
    owner = get_repo_owner(repo),
    repo = get_repo_name(repo)
  )
}

#' @name action
#' @rdname action
#'
#' @param repo Character. Address of repositories in `owner/name` format.
#' @param keep_expired Logical. Should expired artifacts be returned. Default `FALSE`.
#' @param which Character. Either `"latest"` to return only the most recent of each
#'   artifact or `"all"` to return all artifacts.
#' @param filter_branch Character. Regex pattern to filter artifacts by branch name.
#' @param exclude_branch Logical. If `TRUE`, exclude branches matching `filter_branch` instead of including them. Default `FALSE`.
#' @param filter Character. Regex pattern to filter artifacts by artifact name.
#' @param exclude Logical. If `TRUE`, exclude artifact names matching `filter` instead of including them. Default `FALSE`.
#' @export
#'
action_artifacts = function(repo, filter = NULL, exclude = FALSE,
                            keep_expired = FALSE, which = c("latest", "all"),
                            filter_branch = NULL, exclude_branch = FALSE) {
  which = match.arg(which)

  arg_is_chr(repo)
  arg_is_chr_scalar(which)
  arg_is_lgl_scalar(keep_expired, exclude_branch, exclude)
  arg_is_chr_scalar(filter_branch, allow_null = TRUE)
  arg_is_chr_scalar(filter, allow_null = TRUE)

  res = purrr::map_dfr(
    repo,
    function(r) {
      res = purrr::safely(github_api_action_artifacts)(r)
      status_msg(
        res,
        fail = "Failed to retrieve artifacts for repo {.val {repo}}."
      )

      if (failed(res) || empty_result(res) || result(res)[["total_count"]] == 0) {
        tibble::tibble(
          repo    = character(),
          branch  = character(),
          name    = character(),
          id      = integer(),
          size    = integer(),
          url     = character(),
          expired = logical(),
          created = lubridate::ymd_hms(),
          updated = lubridate::ymd_hms(),
          expires = lubridate::ymd_hms()
        )
      } else {
        artifacts = result(res)[["artifacts"]]

        tibble::tibble(
          repo    = r,
          branch  = purrr::map_chr(artifacts, c("workflow_run", "head_branch"), .default = NA),
          name    = purrr::map_chr(artifacts, "name", .default = NA),
          id      = purrr::map_dbl(artifacts, "id", .default = NA),
          size    = purrr::map_dbl(artifacts, "size_in_bytes", .default = NA),
          url     = purrr::map_chr(artifacts, "url", .default = NA),
          expired = purrr::map_lgl(artifacts, "expired", .default = NA),
          created = purrr::map_chr(artifacts, "created_at", .default = NA) %>% lubridate::ymd_hms(),
          updated = purrr::map_chr(artifacts, "updated_at", .default = NA) %>% lubridate::ymd_hms(),
          expires = purrr::map_chr(artifacts, "expires_at", .default = NA) %>% lubridate::ymd_hms()
        )
      }
    }
  )

  if (!keep_expired) {
    res = dplyr::filter(res, .data[["expired"]] == FALSE)
  }

  res = filter_results(res, filter_branch, exclude_branch, col = "branch")
  res = filter_results(res, filter,        exclude,        col = "name")

  if (which == "latest") {
    res %>%
      dplyr::group_by(.data[["repo"]], .data[["name"]]) %>%
      dplyr::arrange(dplyr::desc(.data[["created"]])) %>%
      dplyr::slice(1) %>%
      dplyr::ungroup()
  } else {
    res
  }

}

#' Internal ghclass helpers
#'
#' These functions are exported so that companion packages (e.g.
#' ghclasspeer) can reuse them. They are not intended for end users
#' and their interfaces may change without notice.
#'
#' @keywords internal
#' @name ghclass-internal
NULL

#' @rdname repo_core
#' @param status Logical. Should forking be allowed for the repository. Default `TRUE`.
#' @export
#'
repo_set_forking = function(repo, status = TRUE) {

  arg_is_chr(repo)
  arg_is_lgl(status)

  res = purrr::map2(
    repo, status,
    function(repo, status) {
      res = purrr::safely(github_api_repo_edit)(repo, allow_forking = status)
      status_msg(
        res,
        "Changed the forking status of repo {.val {repo}} to {.val {status}}.",
        "Failed to change forking status of repo {.val {repo}}."
      )
    }
  )

  invisible(res)
}

#' @title Add team assignments to a roster
#'
#' @description
#' `team_roster()` adds a team (or repo) name column to a roster data frame by
#' randomly grouping members into teams of a target `size`. Teams may be formed
#' within groups (e.g. lab sections) via `by`, and names are built from a
#' [glue][glue::glue_data] template that can reference any column of `roster`
#' together with the generated `team_id`. When `by` is supplied and no `name` is
#' given, the grouping column(s) are folded into the default name so team names
#' stay unique across groups.
#'
#' @param roster Data frame. Course roster with one row per student.
#' @param size Integer. Target (maximum) team size. Within each group members are
#'   split into `ceiling(n / size)` teams of as-equal-as-possible size. An error
#'   is thrown if a group has fewer than `size` members.
#' @param name Character. A glue template for the team / repo name, used as-is if
#'   supplied. May reference any column of `roster` as well as `team_id` (the
#'   within-group team number). If `NULL` (default) the template is
#'   `"team{team_id}"`, or `"{group}-team{team_id}"` when `by` is set.
#' @param by Character. Optional column name(s) in `roster` to form teams within
#'   (e.g. a lab section). If `NULL` (default) teams are formed across the whole roster.
#' @param col Character. Name of the output column to add. Default `"team"`.
#' @param pad Logical or integer. Zero-padding applied to `team_id` before it is
#'   substituted into `name`. `TRUE` (default) pads to a consistent width based on
#'   the largest team number, `FALSE` leaves it as an integer, or supply an integer
#'   for a fixed width.
#' @param shuffle Logical. Should members be randomly assigned to teams? Default `TRUE`.
#' @param seed Integer. Optional random seed for reproducible assignments.
#'
#' @return The `roster` data frame with `team_id` and `col` columns added.
#'
#' @examples
#' \dontrun{
#' roster = readr::read_csv(system.file("roster.csv", package = "ghclass"))
#'
#' # Teams of 4 across the whole roster -> "team01", "team02", ...
#' team_roster(roster, size = 4)
#'
#' # Teams of 4 within each lab section, group folded into the default name
#' team_roster(roster, size = 4, by = "section", seed = 20250901)
#'
#' # Custom name template (used exactly as written)
#' team_roster(
#'   roster, size = 4, by = "section",
#'   name = "hw1_lab{section}_team{team_id}", seed = 20250901
#' )
#' }
#'
#' @export
#'
team_roster = function(
  roster, size, name = NULL, by = NULL,
  col = "team", pad = TRUE, shuffle = TRUE, seed = NULL
) {
  arg_is_df(roster)
  arg_is_pos_int_scalar(size)
  arg_is_chr_scalar(col)
  arg_is_chr_scalar(name, allow_null = TRUE)
  arg_is_chr(by, allow_null = TRUE)
  arg_is_lgl_scalar(shuffle)
  arg_is_pos_int_scalar(seed, allow_null = TRUE)

  if (is.logical(pad)) arg_is_lgl_scalar(pad) else arg_is_pos_int_scalar(pad)

  missing_cols = setdiff(by, names(roster))
  if (length(missing_cols) > 0)
    cli_stop("Column{?s} {.val {missing_cols}} (from {.arg by}) {?is/are} not present in {.arg roster}.")

  if (is.null(by)) {
    if (nrow(roster) < size)
      cli_stop(
        "Cannot form teams of {.arg size} {.val {size}}: ",
        "the roster has only {nrow(roster)} member{?s}."
      )
  } else {
    counts = dplyr::count(roster, dplyr::across(dplyr::all_of(by)))
    small = counts[counts[["n"]] < size, , drop = FALSE]
    if (nrow(small) > 0) {
      groups = do.call(paste, c(small[by], sep = "/"))
      cli_stop(
        "Cannot form teams of {.arg size} {.val {size}}: ",
        "{cli::qty(groups)} group{?s} ({.val {groups}}) {?has/have} fewer than {size} member{?s}."
      )
    }
  }

  if (is.null(name)) {
    name = if (is.null(by)) {
      "team{team_id}"
    } else {
      paste0(paste0("{", by, "}", collapse = "-"), "-team{team_id}")
    }
  }

  clobber = intersect(c("team_id", col), names(roster))
  if (length(clobber) > 0)
    cli::cli_alert_info("Overwriting existing column{?s}: {.val {clobber}}.")

  assign_ids = function(n) {
    # Build balanced team ids for n members:
    #   ceiling(n / size) -> number of teams needed so none exceeds `size`
    #   rep(., size)      -> repeat the team ids enough times to cover everyone
    #   [seq_len(n)]      -> keep the first n, cycling ids round-robin so team
    #                        sizes never differ by more than one
    ids = rep(seq_len(ceiling(n / size)), size)[seq_len(n)]
    # Shuffle so members are assigned to teams at random, not in roster order
    if (shuffle) ids = sample(ids)
    ids
  }

  do_assign = function() {
    if (is.null(by)) {
      roster[["team_id"]] = assign_ids(nrow(roster))
      roster
    } else {
      roster %>%
        dplyr::group_by(dplyr::across(dplyr::all_of(by))) %>%
        dplyr::mutate(team_id = assign_ids(dplyr::n())) %>%
        dplyr::ungroup()
    }
  }

  out = if (is.null(seed)) do_assign() else withr::with_seed(seed, do_assign())

  if (!isFALSE(pad)) {
    width = if (isTRUE(pad)) max(stringr::str_length(out[["team_id"]])) else pad
    out[["team_id"]] = stringr::str_pad(out[["team_id"]], width = width, side = "left", pad = "0")
  }

  out[[col]] = as.character(glue::glue_data(out, name))

  out
}

#' @title Anonymize a local repo or grading project
#'
#' @description
#' `r lifecycle::badge("experimental")`
#'
#' Makes a best-effort pass over a local git repository, a directory of
#' repositories, or a grading project created by [org_grade_assignment()] and
#' removes student personally identifying information (PII). Two things happen:
#'
#' * **File contents** - every text file is scanned and any roster-derived value
#'   (names, NetIDs, emails, GitHub usernames, ...) is replaced with a stable
#'   per-student token (e.g. `student_07`).
#' * **Git history** - each repository's `.git` directory is either deleted
#'   outright (the default) or kept untouched.
#'
#' @details
#' Git history cannot be reliably anonymized while also being preserved. Even
#' after rewriting commit author and committer identities, the *original file
#' contents* of older commits remain recoverable (`git show <old-sha>:README.md`),
#' and real names routinely survive in commit messages and branch names. Because
#' anyone you hand the repository to receives its full history, there are only two
#' honest options, selected with `git_history`:
#'
#' * `"delete"` (default) - remove each repository's `.git` directory entirely.
#'   The result is the scrubbed working tree with no recoverable history. This is
#'   the safe choice for sharing student work.
#' * `"keep"` - leave `.git` untouched. The working-tree files are still scrubbed,
#'   but the original commit history (with real names and emails) is retained, so
#'   the result is **not** safe to distribute.
#'
#' The function works on a *copy* by default (`<path>_anon`), leaving the original
#' intact, so deleting history is non-destructive to your working copy.
#'
#' Only files whose names match `types` are scanned, so binary and other non-text
#' files are skipped by design. Matching is best effort: word boundaries are used
#' for names / NetIDs to avoid clobbering unrelated text, but short or common
#' names may still over-match, and HTML-entity-encoded names are not detected.
#' Anonymization is roster-driven, so anyone absent from the roster (e.g. a
#' student who later dropped) will not be matched.
#'
#' @param path Character. Path to a git repository, a directory of repositories,
#'   or a grading project folder (one containing a `repos/` subdirectory of
#'   cloned repositories, e.g. as produced by [org_grade_assignment()]).
#' @param roster Either a path to a roster CSV file or a data frame. Each row is
#'   treated as one student.
#' @param cols <[`tidy-select`][dplyr::dplyr_tidy_select]> Roster columns whose
#'   values should be stripped from the contents. Defaults to the `name`,
#'   `first`, `last`, `email`, `netid`, and `github` columns when present.
#' @param output Character. Directory to write the anonymized copy to. Defaults
#'   to `<path>_anon`. Set `output = path` to modify the folder in place (gated by
#'   a confirmation prompt when `prompt = TRUE`).
#' @param types Character. Regular expressions matched against the end of each
#'   file name to decide which files are treated as text and scanned. Defaults to
#'   common text formats found in student work (markdown, R / Python source,
#'   notebooks, html, csv, yaml, ...).
#' @param git_history Character. How each repository's git history is handled.
#'   Either `"delete"` (default, remove the `.git` directory entirely) or `"keep"`
#'   (leave it untouched, retaining the un-anonymized history).
#' @param prompt Logical. Prompt for confirmation before modifying a folder in
#'   place. Defaults to [interactive()].
#'
#' @return Invisibly, a list with two tibbles: `text` (files scrubbed and
#'   replacements made per directory) and `git` (the action taken per repo).
#'
#' @examples
#' \dontrun{
#' local_repo_anonymize(
#'   "grading/hw1",
#'   roster = "rosters/hw1_roster.csv"
#' )
#'
#' # scrub text but keep history, choosing the columns explicitly
#' local_repo_anonymize(
#'   "grading/hw1",
#'   roster = roster_df,
#'   cols = c(name, netid, email),
#'   git_history = "keep"
#' )
#' }
#'
#' @export
#'
local_repo_anonymize = function(
  path,
  roster,
  cols = tidyselect::any_of(c("name", "first", "last", "email", "netid", "github")),
  output = NULL,
  types = c(
    ".md", ".qmd", ".[Rr]md", ".txt", ".csv", ".tsv", ".html?", ".[Rr]",
    ".py", ".ipynb", ".json", ".ya?ml", ".tex", ".bib", ".toml", ".[Rr]proj"
  ),
  git_history = c("delete", "keep"),
  prompt = interactive()
) {
  require_gert()
  arg_is_chr_scalar(path)
  arg_is_chr_scalar(output, allow_null = TRUE)
  arg_is_chr(types)
  arg_is_lgl_scalar(prompt)
  git_history = match.arg(git_history)
  cols = rlang::enquo(cols)

  if (!fs::dir_exists(path))
    cli_stop("Unable to locate {.file {path}}.")
  path = as.character(fs::path_real(path))

  # The roster + map are validated before we copy anything potentially large.
  roster = anon_read_roster(roster)
  map = anon_build_map(roster, cols)
  if (nrow(map) == 0)
    cli_stop("No PII values found in the selected roster columns.")

  if (is.null(output))
    output = paste0(path, "_anon")
  output = fs::path_expand(output)
  in_place = identical(as.character(fs::path_norm(output)), path)

  if (in_place) {
    if (prompt && !cli_yeah("Modify {.file {path}} in place? This rewrites files and deletes git history and cannot be undone."))
      return(invisible(NULL))
    work = path
  } else {
    if (fs::dir_exists(output))
      cli_stop("Output directory {.file {output}} already exists; remove it or choose another {.arg output}.")
    cli::cli_alert_info("Copying {.file {path}} to {.file {output}}.")
    fs::dir_copy(path, output)
    work = as.character(fs::path_real(output))
  }

  layout = anon_detect_layout(work)

  cli::cli_alert_info("Scrubbing text in {.val {length(layout$text_dirs)}} director{?y/ies}.")
  text_summ = purrr::map(
    layout$text_dirs,
    function(d) {
      r = anon_scrub_dir(d, map, types)
      cli::cli_alert_success(
        "Scrubbed {.val {r$replacements}} value{?s} in {.val {r$files}} file{?s} under {.file {fs::path_file(d)}}."
      )
      r
    }
  )

  git_summ = tibble::tibble()
  if (length(layout$repos) > 0) {
    if (git_history == "delete") {
      git_summ = dplyr::bind_rows(purrr::map(
        layout$repos,
        function(repo) {
          res = purrr::safely(anon_git_delete)(repo)
          status_msg(
            res,
            "Deleted git history for {.val {fs::path_file(repo)}}.",
            "Failed to delete git history for {.val {fs::path_file(repo)}}."
          )
          tibble::tibble(
            repo = fs::path_file(repo),
            git_history = git_history,
            deleted = succeeded(res)
          )
        }
      ))
    } else {
      git_summ = tibble::tibble(
        repo = fs::path_file(layout$repos),
        git_history = git_history,
        deleted = FALSE
      )
      cli_warn(
        "Git history was kept; the original commit history (including real names ",
        "and emails) is retained, so the result is not safe to distribute."
      )
    }
  }

  cli::cli_alert_success("Anonymized copy written to {.file {work}}.")

  invisible(list(
    text = tibble::tibble(
      dir = as.character(fs::path_rel(layout$text_dirs, work)),
      files = purrr::map_int(text_summ, "files"),
      replacements = purrr::map_int(text_summ, "replacements")
    ),
    git = git_summ
  ))
}


# Layout detection -------------------------------------------------------------

anon_detect_layout = function(path) {
  path = as.character(fs::path_real(path))

  if (is_git_repo(path))
    return(list(kind = "repo", repos = path, text_dirs = path))

  # A grading project (e.g. from org_grade_assignment()) keeps the student
  # repos under a `repos/` subdirectory alongside artifact folders.
  repos_dir = fs::path(path, "repos")
  repos = character()
  if (fs::dir_exists(repos_dir)) {
    repos = as.character(fs::dir_ls(repos_dir, type = "directory"))
    repos = repos[purrr::map_lgl(repos, is_git_repo)]
  }

  if (length(repos) > 0) {
    top = as.character(fs::dir_ls(path, type = "directory"))
    key = top[purrr::map_lgl(top, is_git_repo)]
    return(list(
      kind = "grading",
      repos = unique(c(repos, key)),
      text_dirs = top
    ))
  }

  dirs = as.character(repo_dir_helper(path))
  list(
    kind = "repo_dir",
    repos = dirs[purrr::map_lgl(dirs, is_git_repo)],
    text_dirs = path
  )
}


# Roster + replacement map -----------------------------------------------------

anon_read_roster = function(roster) {
  if (is.character(roster)) {
    arg_is_chr_scalar(roster)
    if (!fs::file_exists(roster))
      cli_stop("Unable to locate roster file {.file {roster}}.")
    roster = readr::read_csv(roster, show_col_types = FALSE)
  }
  arg_is_df(roster)
  roster
}

anon_build_map = function(roster, cols) {
  loc = tidyselect::eval_select(cols, data = roster)
  if (length(loc) == 0)
    cli_stop("No matching columns were selected from the roster via {.arg cols}.")

  pii = roster[loc]
  n = nrow(pii)
  width = max(2L, nchar(as.character(n)))
  tokens = sprintf(paste0("student_%0", width, "d"), seq_len(n))

  map = tibble::tibble(
    literal = trimws(unlist(lapply(pii, as.character), use.names = FALSE)),
    token = rep(tokens, times = ncol(pii))
  )

  map = map[!is.na(map$literal) & nzchar(map$literal), ]

  short = map$literal[nchar(map$literal) < 2L]
  if (length(short) > 0) {
    cli_warn(
      "Ignoring PII value{?s} shorter than 2 characters: {.val {unique(short)}}."
    )
    map = map[nchar(map$literal) >= 2L, ]
  }

  map = dplyr::distinct(map, .data$literal, .data$token)

  dup = unique(map$literal[duplicated(map$literal)])
  if (length(dup) > 0) {
    cli_warn(
      "Value{?s} {.val {dup}} appear for multiple students and will be replaced ",
      "with a single token; those occurrences cannot be distinguished."
    )
    map = map[!duplicated(map$literal), ]
  }

  map$is_email = grepl("@", map$literal, fixed = TRUE)
  map$regex = anon_build_regex(map$literal, map$is_email)
  map = map[order(-nchar(map$literal), -map$is_email), ]
  map
}

anon_build_regex = function(literal, is_email) {
  quoted = paste0("\\Q", literal, "\\E")
  left = ifelse(grepl("^\\w", literal, perl = TRUE), "\\b", "")
  right = ifelse(grepl("\\w$", literal, perl = TRUE), "\\b", "")
  ifelse(is_email, quoted, paste0(left, quoted, right))
}


# Text scrubbing ---------------------------------------------------------------

anon_match_types = function(files, types) {
  pattern = paste0("(", paste(types, collapse = "|"), ")$")
  grepl(pattern, fs::path_file(files))
}

anon_is_binary = function(file) {
  con = file(file, "rb")
  on.exit(close(con))
  raw = readBin(con, what = "raw", n = 8000L)
  length(raw) > 0L && any(raw == as.raw(0L))
}

anon_scrub_file = function(file, map) {
  txt = readr::read_file(file)
  if (length(txt) == 0L || is.na(txt) || !nzchar(txt))
    return(0L)

  present = vapply(map$literal, function(l) grepl(l, txt, fixed = TRUE), logical(1))
  if (!any(present))
    return(0L)

  n_repl = 0L
  for (i in which(present)) {
    m = gregexpr(map$regex[i], txt, perl = TRUE)[[1]]
    cnt = if (length(m) == 1L && m[1] == -1L) 0L else length(m)
    if (cnt > 0L) {
      txt = gsub(map$regex[i], map$token[i], txt, perl = TRUE)
      n_repl = n_repl + cnt
    }
  }

  if (n_repl > 0L)
    readr::write_file(txt, file)

  n_repl
}

anon_scrub_dir = function(dir, map, types) {
  files = as.character(fs::dir_ls(dir, recurse = TRUE, type = "file", all = TRUE))
  files = files[!grepl("(^|/)\\.git/", files)]
  files = files[anon_match_types(files, types)]

  files_changed = 0L
  total = 0L
  for (f in files) {
    if (anon_is_binary(f))
      next
    res = tryCatch(anon_scrub_file(f, map), error = function(e) NA_integer_)
    if (is.na(res)) {
      cli_warn("Failed to scrub {.file {f}}.")
      next
    }
    if (res > 0L) {
      files_changed = files_changed + 1L
      total = total + res
    }
  }

  list(files = files_changed, replacements = total)
}


# Git history ------------------------------------------------------------------

anon_git_delete = function(repo) {
  git_dir = fs::path(repo, ".git")
  if (fs::dir_exists(git_dir))
    fs::dir_delete(git_dir)
  else if (fs::file_exists(git_dir))
    fs::file_delete(git_dir)
  invisible(TRUE)
}

roster_fixture = function() {
  tibble::tibble(
    last = c("Smith", "Jones"),
    first = c("Alice", "Bob"),
    email = c("alice.smith@duke.edu", "bob.jones@duke.edu"),
    netid = c("as99", "bj12"),
    github = c("alicesmith", "bobby"),
    name = c("Alice Smith", "Bob Jones")
  )
}

# Build a grading-project-style fixture: <root>/repos/<repo> with git history.
grading_fixture = function(dir) {
  repo = fs::path(dir, "repos", "hw1-team01")
  fs::dir_create(repo)

  writeLines(
    c(
      "# Homework 1",
      "* Alice Smith - alice.smith@duke.edu",
      "* Bob Jones (bj12) - bob.jones@duke.edu",
      "",
      "The word smithy must be preserved."
    ),
    fs::path(repo, "README.md")
  )
  writeBin(as.raw(c(1L, 0L, 2L, 0L, 255L)), fs::path(repo, "logo.png"))

  art = fs::path(dir, "html")
  fs::dir_create(art)
  writeLines('<meta name="author" content="Alice Smith">', fs::path(art, "hw1-team01.html"))

  gert::git_init(repo)
  gert::git_add("README.md", repo = repo)
  gert::git_commit(
    "initial commit by Alice", repo = repo,
    author = gert::git_signature("Alice Smith", "as99@machine.stat.duke.edu")
  )
  writeLines("more", fs::path(repo, "notes.txt"))
  gert::git_add("notes.txt", repo = repo)
  gert::git_commit(
    "Bob Jones adds notes", repo = repo,
    author = gert::git_signature("Bob Jones (bj12)", "bob.jones@duke.edu")
  )

  list(root = dir, repo = repo, html = art)
}


test_that("text PII is replaced and non-PII is preserved", {
  skip_if_not_installed("gert")

  root = withr::local_tempdir()
  grading_fixture(root)
  out = paste0(root, "_anon")
  withr::defer(unlink(out, recursive = TRUE))

  suppressWarnings(
    local_repo_anonymize(root, roster = roster_fixture(), git_history = "keep", prompt = FALSE, output = out)
  )

  readme = readr::read_file(fs::path(out, "repos", "hw1-team01", "README.md"))
  expect_false(grepl("Alice|Smith|alice.smith@duke.edu|as99|alicesmith", readme))
  expect_false(grepl("Bob|Jones|bob.jones@duke.edu|bj12", readme))
  expect_match(readme, "student_01")
  expect_match(readme, "student_02")
  expect_match(readme, "smithy")

  html = readr::read_file(fs::path(out, "html", "hw1-team01.html"))
  expect_false(grepl("Alice Smith", html))
  expect_match(html, "student_01")
})

test_that("binary files are left untouched", {
  skip_if_not_installed("gert")

  root = withr::local_tempdir()
  grading_fixture(root)
  out = paste0(root, "_anon")
  withr::defer(unlink(out, recursive = TRUE))

  suppressWarnings(
    local_repo_anonymize(root, roster = roster_fixture(), git_history = "keep", prompt = FALSE, output = out)
  )

  png = readBin(fs::path(out, "repos", "hw1-team01", "logo.png"), "raw", 10)
  expect_identical(png, as.raw(c(1L, 0L, 2L, 0L, 255L)))
})

test_that("only files matching types are scrubbed", {
  skip_if_not_installed("gert")

  root = withr::local_tempdir()
  grading_fixture(root)
  writeLines("contact Alice Smith", fs::path(root, "repos", "hw1-team01", "extra.log"))

  out1 = paste0(root, "_a")
  withr::defer(unlink(out1, recursive = TRUE))
  suppressWarnings(
    local_repo_anonymize(root, roster = roster_fixture(), git_history = "keep", prompt = FALSE, output = out1)
  )
  expect_match(readr::read_file(fs::path(out1, "repos", "hw1-team01", "extra.log")), "Alice Smith")

  out2 = paste0(root, "_b")
  withr::defer(unlink(out2, recursive = TRUE))
  suppressWarnings(
    local_repo_anonymize(root, roster = roster_fixture(), types = ".log", git_history = "keep", prompt = FALSE, output = out2)
  )
  expect_false(grepl("Alice Smith", readr::read_file(fs::path(out2, "repos", "hw1-team01", "extra.log"))))
})

test_that("the original directory is untouched when writing to a copy", {
  skip_if_not_installed("gert")

  root = withr::local_tempdir()
  fx = grading_fixture(root)
  out = paste0(root, "_anon")
  withr::defer(unlink(out, recursive = TRUE))

  local_repo_anonymize(root, roster = roster_fixture(), git_history = "delete", prompt = FALSE, output = out)

  orig = readr::read_file(fs::path(fx$repo, "README.md"))
  expect_match(orig, "Alice Smith")
  expect_true(fs::dir_exists(fs::path(fx$repo, ".git")))
  log = gert::git_log(repo = fx$repo, max = 5)
  expect_true(any(grepl("Alice", log$author)))
})

test_that("delete removes the .git directory entirely", {
  skip_if_not_installed("gert")

  root = withr::local_tempdir()
  grading_fixture(root)
  out = paste0(root, "_anon")
  withr::defer(unlink(out, recursive = TRUE))

  res = local_repo_anonymize(root, roster = roster_fixture(), git_history = "delete", prompt = FALSE, output = out)

  arepo = fs::path(out, "repos", "hw1-team01")
  expect_false(fs::dir_exists(fs::path(arepo, ".git")))
  # working tree is still scrubbed
  expect_false(grepl("Alice Smith", readr::read_file(fs::path(arepo, "README.md"))))
  expect_true(all(res$git$deleted))
})

test_that("keep retains git history and warns", {
  skip_if_not_installed("gert")

  root = withr::local_tempdir()
  grading_fixture(root)
  out = paste0(root, "_anon")
  withr::defer(unlink(out, recursive = TRUE))

  expect_warning(
    local_repo_anonymize(root, roster = roster_fixture(), git_history = "keep", prompt = FALSE, output = out),
    "not safe to distribute"
  )

  arepo = fs::path(out, "repos", "hw1-team01")
  expect_true(fs::dir_exists(fs::path(arepo, ".git")))
  log = gert::git_log(repo = arepo, max = 5)
  expect_true(any(grepl("Alice|Bob", log$author)))
})

test_that("cols accepts tidyselect expressions", {
  map = anon_build_map(roster_fixture(), rlang::quo(c(netid, email)))
  expect_setequal(
    sort(map$literal),
    sort(c("as99", "bj12", "alice.smith@duke.edu", "bob.jones@duke.edu"))
  )
})

test_that("missing default columns are tolerated", {
  roster = roster_fixture()[, c("name", "email")]
  map = anon_build_map(
    roster,
    rlang::quo(tidyselect::any_of(c("name", "first", "last", "email", "netid", "github")))
  )
  expect_setequal(sort(unique(map$token)), c("student_01", "student_02"))
})

test_that("shared values across students warn and collapse to one token", {
  roster = tibble::tibble(
    name = c("Alice Smith", "Bob Smith"),
    first = c("Alice", "Bob"),
    last = c("Smith", "Smith")
  )
  expect_warning(
    anon_build_map(roster, rlang::quo(c(name, first, last))),
    "multiple students"
  )
  map = suppressWarnings(anon_build_map(roster, rlang::quo(c(name, first, last))))
  expect_equal(sum(map$literal == "Smith"), 1L)
})

test_that("short values are dropped with a warning", {
  roster = tibble::tibble(name = c("Al", "X"))
  expect_warning(
    anon_build_map(roster, rlang::quo(name)),
    "shorter than 2"
  )
  map = suppressWarnings(anon_build_map(roster, rlang::quo(name)))
  expect_false("X" %in% map$literal)
})

#' @rdname repo_core
#' @export
#'
repo_allows_forking = function(repo) {

  arg_is_chr(repo)

  purrr::map_lgl(
    repo,
    function(repo) {
      res = purrr::safely(github_api_repo)(repo)
      status_msg(
        res,
        fail = "Failed to retrieve repo {.val {repo}}."
      )

      if (succeeded(res)) {
        result(res)[["allow_forking"]]
      } else {
        NA
      }
    }
  )
}

#' @rdname local_repo
#' @param max Maximum number of log entries to retrieve per repo.
#' @export
#'
local_repo_log = function(repo_dir, max = 100) {
  require_gert()
  arg_is_chr(repo_dir)

  repo_dir = repo_dir_helper(repo_dir)

  purrr::map_dfr(
    repo_dir,
    function(dir) {
      gert::git_log(repo = dir, max = max) %>%
        dplyr::mutate(repo = dir) %>%
        dplyr::relocate(.data$repo)
    }
  )
}

github_api_pr_create = function(repo, head, base, title, body, draft = TRUE){
  ghclass_api_v3_req(
    endpoint = "POST /repos/:owner/:repo/pulls",
    owner = get_repo_owner(repo),
    repo = get_repo_name(repo),
    head = head,
    base = base,
    title = title,
    body = body,
    draft = draft,
    .send_headers = c(Accept = "application/vnd.github.shadow-cat-preview+json")
  )
}

#' @rdname pr
#'
#' @export
#'
pr_create = function(repo, title, head, base, body = "", draft = FALSE) {

  arg_is_chr(repo, title, base, head, body)
  arg_is_lgl(draft)

  res = purrr::pmap(
    list(repo, base, head, title, body, draft),
    function(repo, base, head, title, body, draft) {
      res = purrr::safely(github_api_pr_create)(
        repo, base = base, head = head, title = title, body = body, draft = draft
      )

      details = cli_glue("{repo} ({base} <- {head})")

      status_msg(
        res,
        "Created pull request for {.val {details}}.",
        "Failed create pull request for {.val {details}}."
      )
    }
  )

  invisible(res)
}

github_api_download_file = function(url, dest, max_tries = 3) {
  arg_is_chr_scalar(url, dest)

  for (i in seq_len(max_tries)) {
    req = httr::GET(
      url,
      httr::add_headers(
        Authorization = paste("bearer", github_get_token())
      )
    )

    if (httr::status_code(req) != 202)
      break

    Sys.sleep(2)
  }

  res = httr::content(req, as = "raw")
  code = httr::status_code(req)

  if (code == 202)
    cli_stop("GitHub API Error (202) - the requested archive is still being generated, please retry shortly.")

  if (code >= 300) {
    msg = httr::content(req, as = "parsed")[["message"]]
    if (is.null(msg))
      msg = "request failed"
    cli_stop("GitHub API Error ({code}) - {msg}")
  }

  writeBin(res, dest)

  return(dest)
}

#' @rdname ghclass-internal
#' @export
github_api_repo_tree = function(repo, sha = NULL) {
  arg_is_chr_scalar(repo)
  arg_is_chr_scalar(sha, allow_null = TRUE)

  if (is.null(sha))
    sha = "HEAD"

  ghclass_api_v3_req(
    endpoint = "GET /repos/:owner/:repo/git/trees/:sha?recursive=1",
    owner = get_repo_owner(repo),
    repo = get_repo_name(repo),
    sha = sha
  )
}

#' @rdname ghclass-internal
#' @export
repo_files = function(repo, branch = NULL) {
  arg_is_chr(repo)
  arg_is_chr(branch, allow_null = TRUE)

  if (is.null(branch))
    branch = list(NULL)

  purrr::map2_dfr(
    repo, branch,
    function(repo, branch) {

      res = purrr::safely(github_api_repo_tree)(repo, branch)
      repo_txt = format_repo(repo, branch)

      if (failed(res)) {
        cli::cli_alert_danger("Failed to retrieve files for repo {.val {repo_txt}}")
        return(NULL)
      }

      purrr::map_dfr(result(res)[["tree"]], ~c(repo = repo, .))
    }
  )
}

github_api_org_accept_invite = function(org, token) {
  arg_is_chr_scalar(org, token)

  with_pat(
    token,
    ghclass_api_v3_req(
      endpoint = "PATCH /user/memberships/orgs/:org",
      org = org,
      state = "active"
    )
  )
}

org_accept_invite = function(org, user, pat) {
  arg_is_chr(org, pat)

  purrr::pwalk(
    list(org, user, pat),
    function(org, user, pat) {
      res = purrr::safely(github_api_org_accept_invite)(org, pat)

      status_msg(
        res,
        "Accepted {.val {user}}s invite to org {.val {org}}.",
        "Failed to accept {.val {user}}s invite to org {.val {org}}."
      )
    }
  )
}

# Extracts base64 encoded content from files
extract_content = function(repo, path, file, include_details = TRUE, quiet = FALSE) {
  if (is.null(file) | is.null(file[["content"]])) {
    if (!quiet)
      cli::cli_alert_danger(
        "Unable to retrieve file {.val {path}} from repo {.val {repo}}."
      )
    return(invisible(NULL))
  }

  content = base64enc::base64decode(file[["content"]])
  content = purrr::possibly(rawToChar, content)(content)

  if (include_details) {
    file[["content"]] = NULL
    attributes(content) = file
  }

  content
}

github_api_code_search = function(query) {
  ghclass_api_v3_req(
    endpoint = "GET /search/code",
    q = query
  )
}


find_file = function(repo, file, verbose = TRUE){
  arg_is_chr_scalar(repo)
  arg_is_chr(file)

  purrr::flatten_chr(
    purrr::map(
      file,
      function(file) {

        query = paste0(" path:", ifelse(!(fs::path_dir(file) == "."), fs::path_dir(file), "/"),
                       " repo:", repo,
                       " filename:", fs::path_file(file))
        res = github_api_code_search(query)

        if(res[["total_count"]] > 0){
          purrr::map_chr(res[["items"]], "path")
        } else if (verbose){
          cli::cli_alert_danger("Cannot find file {.val {file}} on {.val {repo}}.")
        }
      }
    )
  )
}


file_exists = function(repo, path, branch = NULL){
  arg_is_chr(repo, path)
  arg_is_chr(branch, allow_null = TRUE)

  files = repo_files(repo, branch)

  purrr::map_lgl(path, ~.x %in% files[["path"]])

}



check_file_modification = function(repo, path, branch){
  arg_is_chr_scalar(repo, path, branch)
  commits = repo_commits(repo, branch = branch, path = path)
  nrow(commits) > 1
}


response_status = function(x) {
  attr(x, "response")[["status"]]
}

# Modified from gh:::gh_set_verb
endpoint_verb = function(x) {
  if (!nzchar(x))
    return(NA)

  if (grepl("^/", x) || grepl("^http", x)) {
     method = "GET"
  } else {
    method = gsub("^([^/ ]+)\\s+.*$", "\\1", x)
  }

  stopifnot(method %in% c("GET", "POST", "PATCH", "PUT", "DELETE"))

  method
}


#' @rdname ghclass-internal
#' @export
ghclass_api_v3_req = function(
    endpoint, ..., .send_headers = character(),
    version = "2022-11-28", limit = github_get_api_limit(),
    max_wait = github_get_max_wait(), max_rate = github_get_max_rate()
) {
  arg_is_chr_scalar(version)
  arg_is_chr(.send_headers)

  method = endpoint_verb(endpoint)

  # Support for GitHub API versioning
  # https://github.blog/2022-11-28-to-infinity-and-beyond-enabling-the-future-of-githubs-rest-api-with-api-versioning/
  .send_headers["X-GitHub-Api-Version"] = version

  if (method != "GET") { # Some non-GET methods don't like having a limit set
    limit = NULL
  }

  gh_args = list(
    endpoint = endpoint,
    ...,
    .limit = limit,
    .token = github_get_token(),
    .send_headers = .send_headers
    # .progress = FALSE # TODO - giving an error for some reason
  )
  if (!is.null(max_wait)) gh_args[[".max_wait"]] = max_wait
  if (!is.null(max_rate)) gh_args[[".max_rate"]] = max_rate

  res = suppressMessages(do.call(gh::gh, gh_args))

  if (method == "GET" && is.null(names(res)) && length(res) == github_get_api_limit()) {
    cli::cli_warn(
      c("The number of results is equal to the limit set by {.fn github_set_api_limit},",
        "consider increasing this limit and rerunning the previous function.")
    )
  }

  res
}


# GitHub's search APIs (REST and GraphQL) return at most 1000 results per query.
# Warn when a search matches more than that, since the results are then a
# truncated subset of the matches.
warn_if_search_capped = function(total_count) {
  if (length(total_count) != 1 || is.na(total_count) || total_count <= 1000)
    return(invisible())

  cli::cli_warn(
    c(
      "This search matched {total_count} repositories but GitHub returns at most 1000 results.",
      i = "The results are truncated - narrow the search to capture the remaining repositories."
    )
  )

  invisible()
}

empty_result = function(res) {
  if (is.null(res)) {
    TRUE
  } else if (is_safely_result(res)) {
    empty_result(result(res))
  } else {
    (length(res) == 1 & all(res == "")) | (is.list(res) & length(res) == 0)
  }
}


is_safely_result = function(x) {
  if (!is.list(x))
    return(FALSE)

  if (!all(c("result", "error") %in% names(x)))
    return(FALSE)

  TRUE
}

check_safely_result = function(x) {
  if (!is_safely_result(x))
    stop(cli_glue("Object is not a {.code purrr::safely} result"), call. = FALSE)
}

#' @rdname ghclass-internal
#' @export
result = function(x) {
  #check_safely_result(x)
  x[["result"]]
}

#' @rdname ghclass-internal
#' @export
error = function(x) {
  #check_safely_result(x)
  x[["error"]]
}

#' @rdname ghclass-internal
#' @export
succeeded = function(x) {
  !is.null(result(x))
}

#' @rdname ghclass-internal
#' @export
failed = function(x) {
  !is.null(error(x))
}

any_failed = function(x) {
  any(purrr::map_lgl(x, failed))
}






return_on_any_failed = function(x) {
  if (any_failed(x)) {
    do.call("return", list(), envir = sys.frame(-1))
  }
}

error_msg = function(x) {
  msg = trimws( error(x)[["message"]] )

  if (grepl("GitHub API error", msg)) {
    msg = gsub("\n+", "\n", msg)
    msg = strsplit(msg, "\n")[[1]]

    sub_replace = function(m, pat) {
      res = gsub(pat, "", m[grepl(pat, m)])

      if(length(res) == 0)
        NULL
      else
        res
    }

    error = msg[1]

    attr(error, "msg") = sub_replace(msg, "Message: ")
    attr(error, "doc") = sub_replace(msg, "Read more at ")
    attr(error, "404") = sub_replace(msg, "URL not found: ")

    error
  } else {
    msg
  }
}

allow_error = function(res, message = NULL, class = NULL, result = "") {

  stopifnot(!is.null(message) | !is.null(class))

  if (succeeded(res))
    return(res)

  message_flag = TRUE
  if (!is.null(message)) {
    message_flag = grepl(message, error_msg(res))
  }

  class_flag = TRUE
  if (!is.null(class)) {
    class_flag = inherits(error(res), class)
  }

  if (message_flag & class_flag) {
    list(
      result = result,
      error = NULL
    )
  } else {
    res
  }
}

ternary = function(check, success, fail) {
  if (check)
    success
  else
    fail
}

has_doc_attr = function(x) {
  !is.null(attr(x, "doc"))
}

has_404_attr = function(x) {
  !is.null(attr(x, "404"))
}

# TODO - fix error_msg processing - doesnt work for PR and some others

#' @rdname ghclass-internal
#' @export
status_msg = function(x, success = NULL, fail = NULL, include_error_msg = TRUE,
                      .envir = parent.frame()) {


  if (succeeded(x) & !is.null(success)) {
    cli::cli_alert_success(success, wrap = FALSE, .envir = .envir)
  }

  if (failed(x) & !is.null(fail)) {
    cli::cli_alert_danger(fail, wrap = FALSE, .envir = .envir)
    if (include_error_msg) {
      msg = error_msg(x)
      cli::cat_line(error_msg_tree(msg))
    }
  }

  invisible(x)
}

error_msg_tree = function(msg) {
  # Only use the attrs that are provided
  attrs = names(attributes(msg))

  d = data.frame(
    id = c("root", "error", "msg", "doc", "404"),
    nodes = I(list("error", attrs, NULL, NULL, NULL)),
    extra = I(list(
      "Error Tree",
      msg,
      paste0(" ", cli_glue('API message: {cli::col_grey(attr(msg,"msg"))}')),
      paste0(" ", cli_glue('API docs: {cli::col_grey(attr(msg,"doc"))}')),
      paste0(" ", cli_glue('Missing page: {cli::col_grey(attr(msg,"404"))}'))
    )),
    stringsAsFactors = FALSE
  )

  cli::tree(d)[-1]
}

# github_api_branch_protect = function(repo, branch) {
#   ghclass_api_v3_req(
#     "PUT /repos/:owner/:repo/branches/:branch/protection",
#     owner = get_repo_owner(repo),
#     repo = get_repo_name(repo),
#     branch = branch,
#     required_status_checks = NA,
#     enforce_admins = NA,
#     required_pull_request_reviews = NA,
#     restrictions = list(
#       users = list(),
#       teams = list()
#     )
#   )
# }
#
# github_api_branch_unprotect = function(repo, branch) {
#   ghclass_api_v3_req(
#     "DELETE /repos/:owner/:repo/branches/:branch/protection",
#     owner = get_repo_owner(repo),
#     repo = get_repo_name(repo),
#     branch = branch
#   )
# }
#
#
#
# branch_protect = function(repo, branch) {
#   arg_is_chr(repo, branch)
#
#   purrr::walk2(
#     repo, branch,
#     function(repo, branch) {
#       res = purrr::safely(github_api_branch_protect)(repo, branch)
#
#       repo_fmt = format_repo(repo, branch)
#
#       status_msg(
#         res,
#         "Protecting branch {.val {repo_fmt}}.",
#         "Failed to protect branch {.val {repo_fmt}}."
#       )
#     }
#   )
# }
#
# branch_unprotect = function(repo, branch) {
#   arg_is_chr(repo, branch)
#
#   purrr::walk2(
#     repo, branch,
#     function(repo, branch) {
#
#       res = purrr::safely(github_api_branch_unprotect)(repo, branch)
#
#       repo_fmt = format_repo(repo, branch)
#
#       status_msg(
#         res,
#         "Removing protection from branch {.val {repo_fmt}}.",
#         "Failed to remove protection from branch {.val {repo_fmt}}."
#       )
#     }
#   )
# }
#

github_api_issue_edit = function(
  repo, number,
  title = NULL, body = NULL, state = NULL, milestone = NULL,
  labels = list(), assignees = list()
) {
  ghclass_api_v3_req(
    endpoint  = "PATCH /repos/:owner/:repo/issues/:issue_number",
    owner     = get_repo_owner(repo),
    repo      = get_repo_name(repo),
    issue_number = number,
    title     = title,
    body      = body,
    assignee  = NULL,
    state     = state,
    milestone = milestone,
    labels    = I(labels),
    assignees = I(assignees)
  )
}

#' @rdname issue
#'
#' @param state	Character. State of the issue. Either "open" or "closed".
#' @param milestone Character. The number of the milestone to associate this issue with.
#' Only users with push access can set the milestone for issues. The milestone is silently dropped otherwise.
#'
#' @export
#'
issue_edit = function(
  repo, number,
  title = NULL, body = NULL,
  state = NULL, milestone = NULL,
  labels = list(), assignees = list()
) {

  arg_is_chr_scalar(repo)
  arg_is_pos_int_scalar(number)
  arg_is_chr_scalar(title, body, state, milestone, allow_null = TRUE)

  if (!is.list(labels))
    labels = list(labels)

  if (!is.list(assignees))
    assignees = list(assignees)

  res = purrr::safely(github_api_issue_edit)(
    repo = repo, number = number, title = title, body = body,
    state = state, milestone = milestone, labels = labels, assignees = assignees
  )

  status_msg(
    res,
    "Edited issue #{.val {number}} for repo {.val {repo}}.",
    "Failed to edit issue #{.val {number}} for repo {.val {repo}}."
  )

  invisible(res)
}

github_api_repo_prs = function(repo, state) {
  arg_is_chr_scalar(repo, state)

  ghclass_api_v3_req(
    endpoint = "GET /repos/:owner/:repo/pulls",
    owner = get_repo_owner(repo),
    repo = get_repo_name(repo),
    state = state
  )
}


#' @rdname repo_details
#'
#' @param state Character. Pull request state.
#' @export
#'
repo_prs = function(repo, state = c("open","closed","all"), quiet = FALSE) {
  state = match.arg(state)
  arg_is_chr(repo)
  arg_is_chr_scalar(state)
  arg_is_lgl_scalar(quiet)

  purrr::map_dfr(
    repo,
    function(repo) {
      res = purrr::safely(github_api_repo_prs)(repo, state)

      if (!quiet) {
        status_msg(
          res,
          fail = "Failed to retrieve pull requests from {.val {repo}}."
        )
      }

      if (empty_result(res)) {
        tibble::tibble(
          repo = character(),
          pr = integer(),
          id = double(),
          title = character(),
          state = character(),
          base_ref = character(),
          head_ref = character()
        )
      } else {
        tibble::tibble(
          repo = repo,
          pr = purrr::map_int(result(res), "number", .default = NA),
          id = purrr::map_dbl(result(res), "id", .default = NA),
          title = purrr::map_chr(result(res), "title", .default = NA),
          state = purrr::map_chr(result(res), "state", .default = NA),
          base_ref = purrr::map_chr(result(res), c("base","ref"), .default = NA),
          head_ref = purrr::map_chr(result(res), c("head","ref"), .default = NA)
        )
      }
    }
  )
}

github_api_org_members = function(org) {
  arg_is_chr_scalar(org)

  ghclass_api_v3_req(
    endpoint = "GET /orgs/:org/members",
    org = org
  )
}

#' @rdname org_members
#' @param include_admins Logical. Should admin users be included in the results. Default `TRUE`.
#' @export
#'
org_members = function(org, filter = NULL, exclude = FALSE,
                      include_admins = TRUE) {
  arg_is_chr_scalar(org)
  arg_is_chr_scalar(filter, allow_null = TRUE)
  arg_is_lgl_scalar(exclude, include_admins)

  res = purrr::safely(github_api_org_members)(org)
  status_msg(
    res,
    fail = "Failed to retrieve members for org {.val {org}}."
  )

  members = purrr::map_chr(result(res), "login")

  if (!include_admins)
    members = setdiff(members, org_admins(org))

  filter_results(members, filter, exclude)
}

#' @name org_perm
#' @rdname org_perm
#'
#' @title Organization permissions
#'
#' @description
#'
#' * `org_sitrep()` - Provides a situation report on a GitHub organization.
#'
#' * `org_set_repo_permission()` - Change the default permission level for org repositories.
#'
#' * `org_workflow_permissions()` - Obtain the current default workflow permission value
#' for the organization.
#'
#' * `org_set_workflow_permissions()` - Change the current default workflow permission value
#' for the organization.
#'
#' * `org_allows_forking()` - returns `TRUE` if members can fork private repositories in the organization.
#'
#' @param org Character. Name of the GitHub organization(s).
#' @param repo_permission Default permission level members have for organization repositories:
#' * read - can pull, but not push to or administer this repository.
#' * write - can pull and push, but not administer this repository.
#' * admin - can pull, push, and administer this repository.
#' * none - no permissions granted by default.
#' @param workflow_permission The default workflow permissions granted to the GITHUB_TOKEN when
#' running workflows in the organization. Accepted values:`"read"` or `"write"`.
#'
#' @return `org_sitep()` invisibly returns the `org` argument.
#'
#' `org_set_repo_permission()` invisibly return a the result of the relevant GitHub API call.
#'
#' `org_workflow_permissions()` returns a character vector with value of either `"read"` or `"write"`.
#'
#' `org_set_workflow_permissions()` invisibly return a the result of the relevant GitHub API call.
#'
#' `org_allows_forking()` returns a logical scalar.
#'
#' @examples
#' \dontrun{
#' org_sitrep("ghclass-test")
#'
#' org_set_repo_permission("ghclass-test", "read")
#'
#' org_workflow_permissions("ghclass-test")
#'
#' org_set_workflow_permissions("ghclass-test", "write")
#'
#' org_sitrep("ghclass-test")
#'
#' # Cleanup
#' org_set_repo_permission("ghclass-test", "none")
#' org_set_workflow_permissions("ghclass-test", "read")
#' }
#'
NULL

#' @name org_members
#' @rdname org_members
#'
#' @title Tools for managing organization membership
#'
#' @description
#'
#' * `org_invite()` - invites user(s) to a GitHub organization.
#'
#' * `org_remove()` - remove user(s) from an organization (and all teams within that organization).
#'
#' * `org_members()` - returns a (filtered) vector of organization members.
#'
#' * `org_pending`() - returns a (filtered) vector of pending organization members.
#'
#' * `org_admins()` - returns a vector of repository administrators. In the case of a
#' non-organization owner (e.g. a user account) returns the owner's login.
#'
#'
#' @param org Character. Name of the GitHub organization(s).
#' @param user Character. GitHub username(s).
#' @param filter Character. Regular expression pattern for matching (or excluding) results
#' @param exclude Logical. Should entries matching the regular expression be excluded or included. Default `FALSE`.
#'
#' @return `org_members()`, `org_pending()`, and `org_admins` all return a character vector
#' of GitHub account names.
#'
#' `org_invite()` and `org_remove()` invisibly return a list containing the results of the relevant GitHub API calls.
#'
#' @examples
#' \dontrun{
#' # Org Details
#' org_admins("ghclass-test")
#'
#' org_admins("rundel") # User, not an organization
#'
#'
#'# Org Membership - Invite, Status, and Remove
#' students = c("ghclass-anya", "ghclass-bruno", "ghclass-celine",
#'              "ghclass-diego", "ghclass-elijah","ghclass-francis")
#'
#' org_invite("ghclass-test", students)
#'
#' org_members("ghclass-test")
#'
#' org_pending("ghclass-test")
#'
#' org_remove("ghclass-test", students, prompt = FALSE)
#'
#' org_pending("ghclass-test")
#' }
#'

NULL




#' @name org_details
#' @rdname org_details
#'
#' @title Obtain details on an organization's repos and teams
#'
#' @description
#'
#' * `org_exists()` - returns `TRUE` if the organization(s) exist on GitHub and `FALSE` otherwise.
#'
#' * `org_teams()` - returns a (filtered) vector of organization teams.
#'
#' * `org_team_details()` - returns a data frame of all organization teams containing identification and permission details.
#'
#' * `org_repos()` - returns a (filtered) vector of organization repositories.
#'
#' * `org_repo_search()` - search for repositories within an organization (preferred for large organizations).
#'
#' * `org_repo_stats()` - returns a tibble of repositories belonging to a GitHub organization along with some
#' basic statistics about those repositories.
#'
#' * `org_repo_forking()` - returns a tibble of private repositories in an organization and whether forking is enabled for each.
#'
#' * `org_user_repos()` - constructs a character vector of per-user repo addresses
#' (`org/prefix<user>suffix`) aligned with `user`, for safely pairing with
#' [repo_add_user()] and similar.
#'
#' @param org Character. Name of the GitHub organization(s).
#' @param filter Character. Regular expression pattern for matching (or excluding) results
#' @param exclude Logical. Should entries matching the regular expression be excluded or included. Default `FALSE`.
#'
#' @return `org_exists()` returns a logical vector.
#'
#' `org_teams()`, `org_repos`, `org_repo_search()`, and `org_user_repos()` return a character vector.
#'
#' `org_team_details()`, `org_repo_stats()`, and `org_repo_forking()` return tibbles.
#'
#' @section Note:
#'
#' The order of `org_repos()` is not guaranteed to match the order of
#' `org_members()`, so do not rely on positional pairing of the two when
#' calling functions like [repo_add_user()]. To pair each user with their own
#' repo deterministically, build the repo names from the user vector with
#' `org_user_repos()`.
#'
#' `org_repo_search()` and `org_repo_stats()` use GitHub's search API, which
#' returns at most 1000 results per query. For organizations with more than 1000
#' matching repositories the results are truncated and a warning is issued; use
#' `filter` (or `name` / `extra`) to narrow the search.
#'
#' @examples
#' \dontrun{
#' # Org repos and teams
#' org_repos("ghclass-test")
#'
#' org_repos("ghclass-test", filter = "hw1-")
#'
#' org_teams("ghclass-test")
#'
#' org_team_details("ghclass-test")
#' }
#'

NULL

#' @name pr
#' @rdname pr
#'
#' @title GitHub Pull Request related tools
#'
#' @description
#'
#' * `pr_create()` - create a pull request GitHub from the `base` branch to the `head` branch.
#'
#' @param repo Character. Address of one or more repositories in "owner/name" format.
#' @param title Character. Title of the pull request.
#' @param head Character. The name of the branch where your changes are implemented.
#' For cross-repository pull requests in the same network, namespace `head` with a user
#' like this: `username:branch`.
#' @param base Character. The name of the branch you want the changes pulled into.
#' This should be an existing branch on the current repository. You cannot submit
#' a pull request to one repository that requests a merge to a base of another repository.
#' @param body Character. The text contents of the pull request.
#' @param draft Logical. Should the pull request be created as a draft pull request
#' (these cannot be merged until allowed by the author). Default `FALSE`.
#'
#' @return `pr_create()` invisibly return a list containing the results of the relevant GitHub API calls.
#'
#' @seealso [repo_issues]
#'
#' @examples
#' \dontrun{
#' repo_create("ghclass-test", "test_pr", auto_init=TRUE)
#'
#' branch_create("ghclass-test/test_pr", branch = "main", new_branch = "test")
#'
#' repo_modify_file("ghclass-test/test_pr", "README.md", pattern = "test_pr",
#'                  content = "Hello", method = "after", branch = "test")
#'
#' pr_create("ghclass-test/test_pr", title = "merge", head = "test", base = "main")
#'
#' repo_delete("ghclass-test/test_pr", prompt = FALSE)
#' }

NULL

github_api_repo_put_file = function(repo, path, content, message, branch = NULL, sha = NULL) {
  ghclass_api_v3_req(
    endpoint = "PUT /repos/:owner/:repo/contents/:path",
    owner = get_repo_owner(repo), repo = get_repo_name(repo),
    path = path,
    content = base64enc::base64encode(content),
    message = message,
    branch = branch,
    sha = sha
  )
}

#' @rdname repo_file
#'
#' @param content Character or raw. Content of the file.
#' @param verbose Logical. Should success / failure messages be printed. Default `TRUE`.
#'
#' @export
#'
repo_put_file = function(repo, path, content, message = NULL, branch = NULL, verbose = TRUE) {

  arg_is_chr_scalar(repo, path)
  arg_is_chr_scalar(message, branch, allow_null = TRUE)

  if (is.null(message))
    message = cli_glue("Adding file {as.character(path)}")

  if (is.character(content))
    content = charToRaw(content)

  # To update an existing file we need its current SHA,
  # if the file does not exist this will be NULL.
  cur_file = repo_get_file(repo, path, branch, quiet=TRUE)
  sha = attr(cur_file, "sha")

  res = purrr::safely(github_api_repo_put_file)(repo, path, content, message, branch, sha)

  if(verbose){
    status_msg(
      res,
      "Added file {.val {path}} to repo {.val {repo}}.",
      "Failed to add file {.val {path}} to repo {.val {repo}}."
    )
  }

  res
}




peer_github_api_repo_put_file = function(repo, path, content, message, branch, sha) {

  args = list(
    endpoint = "PUT /repos/:owner/:repo/contents/:path",
    owner = get_repo_owner(repo), repo = get_repo_name(repo),
    path = path,
    content = base64enc::base64encode(content),
    message = message, branch = branch
  )

  if (!is.null(sha)) {
    args[["sha"]] = sha
  }

  do.call(ghclass_api_v3_req, args)
}



peer_repo_put_file = function(repo, path, content, message = NULL, branch = NULL, sha, verbose = TRUE) {
  arg_is_chr_scalar(repo, path)
  arg_is_chr_scalar(branch, message, sha, allow_null = TRUE)

  if (is.null(message))
    message = cli_glue("Adding file: {path}")

  if (is.character(content))
    content = charToRaw(content)

  res = purrr::safely(peer_github_api_repo_put_file)(
    repo = repo, path = path, content = content,
    message = message, branch = branch, sha = sha
  )

  if (verbose){
    status_msg(
      res,
      "Added file {.val {path}} to repo {.val {repo}}.",
      "Failed to add file {.val {path}} to repo {.val {repo}}."
    )
  }

  res
}

github_api_team_repos = function(org, team_slug) {
  ghclass_api_v3_req(
    endpoint = "GET /orgs/:org/teams/:team_slug/repos",
    org = org,
    team_slug = team_slug
  )
}


#' @rdname team_members
#' @export
#'
team_repos = function(org, team = org_teams(org), team_type = c("name", "slug")) {
  arg_is_chr_scalar(org)
  arg_is_chr(team)
  team_type = match.arg(team_type)

  slug = if (team_type == "name") team_slug_lookup(org, team) else team

  check_team_slug(slug)

  purrr::map2_dfr(
    team, slug,
    function(team, slug) {
      if (is.na(slug))
        res = NULL
      else
        res = purrr::safely(github_api_team_repos)(org, slug)

      repos = if (failed(res) | empty_result(res))
        character()
      else
        purrr::map_chr(result(res), "full_name")

      tibble::tibble(
        team = team,
        slug = slug,
        repo = repos
      )
    }
  )
}

#' @rdname repo_file
#'
#' @param pattern Character. Regex pattern.
#' @param content Character. Content to be added to the file.
#' @param method Character. Should the content `replace` the matched pattern or be inserted `before` or `after` the match.
#' @param all Character. Should all instances of the pattern be modified (`TRUE`) or just the first (`FALSE`).
#'
#' @export
#'
repo_modify_file = function(repo, path, pattern, content, method = c("replace", "before", "after"), all = FALSE,
                            message = "Modified content", branch = NULL) {
  arg_is_chr(repo, path, pattern, content, message)
  arg_is_chr(branch, allow_null=TRUE)

  method = match.arg(method)
  arg_is_chr_scalar(method)
  arg_is_lgl_scalar(all)

  if (is.null(branch))
    branch = list(NULL)

  res = purrr::pmap(
    list(repo, path, pattern, content, message, branch),
    function(repo, path, pattern, content, message, branch) {
      cur_content = repo_get_file(repo = repo, path = path, branch = branch, quiet = TRUE)

      repo_txt = format_repo(repo, branch, path)

      if (is.null(cur_content)) {
        cli::cli_alert_danger("Unable to retrieve {.val {repo_txt}}.")
      } else if (!is.character(cur_content)) {
        cli::cli_alert_danger("Cannot modify non-text file {.val {repo_txt}}.")
      } else {

        if (all) sub_func = gsub
        else     sub_func = sub

        pattern = paste0("(", pattern, ")")

        content = switch(
          method,
          replace = content,
          before  = paste0(content, "\\1"),
          after   = paste0("\\1", content),
          cli_stop("This shouldn't happen")
        )

        new_content = sub_func(pattern, content, cur_content)

        if (cur_content == new_content) {
          cli::cli_alert_danger("Unable to find pattern in {.val {repo_txt}}.", wrap = TRUE)
          return(NULL)
        }

        res = repo_put_file(repo = repo, path = path, content = new_content,
                            message = message, branch = branch, verbose = FALSE)


        status_msg(
          res,
          "Modified file {.val {repo_txt}}.",
          "Failed to modify file {.val {repo_txt}}."
        )
      }
    }
  )

  invisible(res)
}

---
title: "Reviewer feedback form"
output: github_document
params:
  q1_score: [INSERT SCORE]
  q2_score: [INSERT SCORE]
  q3_score: [INSERT SCORE]
---


## Instructions

Enter your feedback for each question below. Please replace `[INSERT SCORE]` in the `q*_score` fields in the YAML with the scores you give the author for each question.

## Feedback

#### 1. Place Question 1 text here. [max. xxx points]

Your response goes here...

#### 2. Place Question 2 text here. [max. xxx points]

Your response goes here...

#### 3. Place Question 3 text here. [max. xxx points]

Your response goes here...

---
title: "Author rating form"
output: github_document
params:
  helpfulness: [INSERT SCORE]
  accuracy: [INSERT SCORE]
  fairness: [INSERT SCORE]
---


## Instructions

Please rate the reviewer's feedback based on the categories below on a scale from "Strongly disagree" to "Strongly agree." Please replace `[INSERT SCORE]` in the `helpfulness`, `accuracy`, and `fairness` fields in the YAML with the scores you give the reviewer for each category.

1. `helpfulness`: "The reviewer's feedback was constructive and helpful."  [max. 4 points]

2. `accuracy`: "The reviewer's assessment accurately describes the quality of my work." [max. 4 points]

3. `fairness`: "The reviewer's assessment was fair." [max. 4 points]

| Score | Rating            |
|-------|-------------------|
| 1     | Strongly disagree |
| 2     | Disagree          |
| 3     | Agree             |
| 4     | Strongly agree    |


You can use the check boxes to help you keep track of your progress.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/local_repo_anonymize.R
\name{local_repo_anonymize}
\alias{local_repo_anonymize}
\title{Anonymize a local repo or grading project}
\usage{
local_repo_anonymize(
  path,
  roster,
  cols = tidyselect::any_of(c("name", "first", "last", "email", "netid", "github")),
  output = NULL,
  types = c(".md", ".qmd", ".[Rr]md", ".txt", ".csv", ".tsv", ".html?", ".[Rr]", ".py",
    ".ipynb", ".json", ".ya?ml", ".tex", ".bib", ".toml", ".[Rr]proj"),
  git_history = c("delete", "keep"),
  prompt = interactive()
)
}
\arguments{
\item{path}{Character. Path to a git repository, a directory of repositories,
or a grading project folder (one containing a \verb{repos/} subdirectory of
cloned repositories, e.g. as produced by \code{\link[=org_grade_assignment]{org_grade_assignment()}}).}

\item{roster}{Either a path to a roster CSV file or a data frame. Each row is
treated as one student.}

\item{cols}{<\code{\link[dplyr:dplyr_tidy_select]{tidy-select}}> Roster columns whose
values should be stripped from the contents. Defaults to the \code{name},
\code{first}, \code{last}, \code{email}, \code{netid}, and \code{github} columns when present.}

\item{output}{Character. Directory to write the anonymized copy to. Defaults
to \verb{<path>_anon}. Set \code{output = path} to modify the folder in place (gated by
a confirmation prompt when \code{prompt = TRUE}).}

\item{types}{Character. Regular expressions matched against the end of each
file name to decide which files are treated as text and scanned. Defaults to
common text formats found in student work (markdown, R / Python source,
notebooks, html, csv, yaml, ...).}

\item{git_history}{Character. How each repository's git history is handled.
Either \code{"delete"} (default, remove the \code{.git} directory entirely) or \code{"keep"}
(leave it untouched, retaining the un-anonymized history).}

\item{prompt}{Logical. Prompt for confirmation before modifying a folder in
place. Defaults to \code{\link[=interactive]{interactive()}}.}
}
\value{
Invisibly, a list with two tibbles: \code{text} (files scrubbed and
replacements made per directory) and \code{git} (the action taken per repo).
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}

Makes a best-effort pass over a local git repository, a directory of
repositories, or a grading project created by \code{\link[=org_grade_assignment]{org_grade_assignment()}} and
removes student personally identifying information (PII). Two things happen:
\itemize{
\item \strong{File contents} - every text file is scanned and any roster-derived value
(names, NetIDs, emails, GitHub usernames, ...) is replaced with a stable
per-student token (e.g. \code{student_07}).
\item \strong{Git history} - each repository's \code{.git} directory is either deleted
outright (the default) or kept untouched.
}
}
\details{
Git history cannot be reliably anonymized while also being preserved. Even
after rewriting commit author and committer identities, the \emph{original file
contents} of older commits remain recoverable (\verb{git show <old-sha>:README.md}),
and real names routinely survive in commit messages and branch names. Because
anyone you hand the repository to receives its full history, there are only two
honest options, selected with \code{git_history}:
\itemize{
\item \code{"delete"} (default) - remove each repository's \code{.git} directory entirely.
The result is the scrubbed working tree with no recoverable history. This is
the safe choice for sharing student work.
\item \code{"keep"} - leave \code{.git} untouched. The working-tree files are still scrubbed,
but the original commit history (with real names and emails) is retained, so
the result is \strong{not} safe to distribute.
}

The function works on a \emph{copy} by default (\verb{<path>_anon}), leaving the original
intact, so deleting history is non-destructive to your working copy.

Only files whose names match \code{types} are scanned, so binary and other non-text
files are skipped by design. Matching is best effort: word boundaries are used
for names / NetIDs to avoid clobbering unrelated text, but short or common
names may still over-match, and HTML-entity-encoded names are not detected.
Anonymization is roster-driven, so anyone absent from the roster (e.g. a
student who later dropped) will not be matched.
}
\examples{
\dontrun{
local_repo_anonymize(
  "grading/hw1",
  roster = "rosters/hw1_roster.csv"
)

# scrub text but keep history, choosing the columns explicitly
local_repo_anonymize(
  "grading/hw1",
  roster = roster_df,
  cols = c(name, netid, email),
  git_history = "keep"
)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/user.R, R/user_exists.R, R/user_repos.R,
%   R/user_type.R
\name{user}
\alias{user}
\alias{user_exists}
\alias{user_repos}
\alias{user_type}
\title{GitHub user related tools}
\usage{
user_exists(user)

user_repos(
  user,
  type = c("owner", "all", "public", "private", "member"),
  filter = NULL,
  exclude = FALSE,
  full_repo = TRUE
)

user_type(user)
}
\arguments{
\item{user}{Character. GitHub username(s).}

\item{type}{Character. Can be one of "all", "owner", "public", "private", "member".}

\item{filter}{Character. Regular expression pattern for matching (or excluding) repositories.}

\item{exclude}{Logical. Should entries matching the regular expression in \code{filter} be excluded or included? Default \code{FALSE}.}

\item{full_repo}{Logical. Should the full repository address be returned
(e.g. \code{owner/repo} instead of just \code{repo})? Default \code{TRUE}.}
}
\value{
\code{user_exists()} returns a logical vector.

\code{user_repos()} and \code{user_type()} retrun a character vector.
}
\description{
\itemize{
\item \code{user_exists()} - returns \code{TRUE} if the username(s) (or organization) exist on GitHub
and \code{FALSE} otherwise. Note that GitHub considers organizations to be a type of user.
\item \code{user_repos()} - returns a (filtered) vector of repositories belonging to the user.
\item \code{user_type()} - returns a vector of the accounts' types.
}
}
\examples{
\dontrun{
user_exists(c("rundel", "ghclass-test", "hopefullydoesnotexist"))

user_repos("rundel", type = "public", filter = "ghclass")

user_repos("ghclass-test")

org_repos("ghclass-test")

user_type(c("rundel", "ghclass-test"))
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pages.R
\name{pages}
\alias{pages}
\alias{pages_enabled}
\alias{pages_status}
\alias{pages_create}
\alias{pages_delete}
\title{Retrieve information about GitHub Pages sites and builds.}
\usage{
pages_enabled(repo)

pages_status(repo)

pages_create(
  repo,
  build_type = c("legacy", "workflow"),
  branch = "main",
  path = "/"
)

pages_delete(repo, prompt = TRUE)
}
\arguments{
\item{repo}{Character. Address of repositories in \code{owner/name} format.}

\item{build_type}{Character. Either \code{"workflow"} or \code{"legacy"} - the former uses GitHub actions to
build and publish the site (requires a workflow file to achieve this).}

\item{branch}{Character. Repository branch to publish.}

\item{path}{Character. Repository path to publish.}

\item{prompt}{Logical. Should the user be prompted before deleting Pages sites. Default \code{TRUE}.}
}
\value{
\code{pages_enabled()} returns a named logical vector - \code{TRUE} if a Pages site exists, \code{FALSE} otherwise.

\code{pages_status()} returns a tibble containing details on Pages sites.

\code{pages_create()} & \code{pages_delete()} return an invisible list containing the API responses.
}
\description{
\itemize{
\item \code{pages_enabled()} - returns \code{TRUE} if a Pages site exists for the repo.
\item \code{pages_status()} - returns more detailed information about a repo's Pages site.
\item \code{pages_create()} - creates a Pages site for the provided repos.
\item \code{pages_delete()} - deletes the Pages site for the provided repos.
}
}
\examples{
\dontrun{
pages_enabled("rundel/ghclass")

pages_status("rundel/ghclass")
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/repo.R, R/repo_unwatch.R, R/repo_watch.R,
%   R/repo_watching.R
\name{repo_notification}
\alias{repo_notification}
\alias{repo_unwatch}
\alias{repo_watch}
\alias{repo_ignore}
\alias{repo_watching}
\title{GitHub Repository tools - notification functions}
\usage{
repo_unwatch(repo)

repo_watch(repo)

repo_ignore(repo)

repo_watching(filter = NULL, exclude = FALSE)
}
\arguments{
\item{repo}{repository address in \code{owner/repo} format}

\item{filter}{character, regex pattern for matching (or excluding) repositories.}

\item{exclude}{logical, should entries matching the regex be excluded or included.}
}
\value{
\code{repo_ignore()}, \code{repo_unwatch()}, and \code{repo_watch()} all invisibly return a
list containing the results of the relevant GitHub API call.

\code{repo_watching()} returns a character vector of watched repos.
}
\description{
\itemize{
\item \code{repo_ignore()} - Ignore a GitHub repository.
\item \code{repo_unwatch()} - Unwatch / unsubscribe from a GitHub repository.
\item \code{repo_watch()} - Watch / subscribe to a GitHub repository.
\item \code{repo_watching()} - Returns a vector of your watched repositories. This should
match the list at \href{https://github.com/watching}{github.com/watching}.
}
}
\examples{
\dontrun{
repo_ignore("Sta323-Sp19/hw1")

repo_unwatch("rundel/ghclass")

repo_watch("rundel/ghclass")
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/team.R, R/team_invite.R, R/team_members.R,
%   R/team_pending.R, R/team_remove.R, R/team_repos.R
\name{team_members}
\alias{team_members}
\alias{team_invite}
\alias{team_pending}
\alias{team_remove}
\alias{team_repos}
\title{Tools for inviting, removing, and managing members of an organization team}
\usage{
team_invite(org, user, team, team_type = c("name", "slug"))

team_members(org, team = org_teams(org), team_type = c("name", "slug"))

team_pending(org, team = org_teams(org), team_type = c("name", "slug"))

team_remove(org, user, team, team_type = c("name", "slug"))

team_repos(org, team = org_teams(org), team_type = c("name", "slug"))
}
\arguments{
\item{org}{Character. Name of the GitHub organization.}

\item{user}{Character. One or more GitHub users to invite.}

\item{team}{Character. Name of teams.}

\item{team_type}{Character. Either "slug" if the team names are slugs or "name" if full team names are provided.}
}
\value{
\code{team_members()}, \code{team_pending()}, and \code{team_repos()} all return a tibble.
The \code{team} column echoes the supplied \code{team} values (names or slugs, per \code{team_type})
and a \code{slug} column gives the corresponding team slug used for the API calls.

\code{team_invite()} and \code{team_remove()} invisibly return a list containing the results
of the relevant GitHub API calls.
}
\description{
\itemize{
\item \code{team_invite()} - add members to team(s).
\item \code{team_remove()} - remove members from team(s).
\item \code{team_members()} - returns a tibble of team members.
\item \code{team_pending()} - returns a tibble of pending team members.
\item \code{team_repos()} - returns a tibble of teams and their repos.
}
}
\examples{
\dontrun{
team_create("ghclass-test",c("hw1-team01","hw1-team02"))

team_invite("ghclass-test", user = "rundel", team = c("hw1-team01", "hw1-team02", "missing_team"))

team_remove("ghclass-test", user = "rundel", team = c("hw1-team01", "missing_team"))

team_members("ghclass-test", org_teams("ghclass-test", "hw1-"))

team_pending("ghclass-test", org_teams("ghclass-test", "hw1-"))

# Add team repo
repo_create("ghclass-test", name = "hw1-team02")
repo_add_team("ghclass-test/hw1-team02", team = "hw1-team02")

team_repos("ghclass-test", org_teams("ghclass-test", "hw1-"))


# Cleanup
repo_delete("ghclass-test/hw1-team02", prompt = FALSE)
team_delete("ghclass-test", org_teams("ghclass-test", "hw1-"), prompt = FALSE)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/local_repo_rename.R
\name{local_repo_rename}
\alias{local_repo_rename}
\title{Rename local directories using a vector of patterns and replacements.}
\usage{
local_repo_rename(repo_dir, pattern, replacement)
}
\arguments{
\item{repo_dir}{Character. A single directory containing one or more repos.}

\item{pattern}{Character. One or more regexp patterns to match to directory names.}

\item{replacement}{Character.  One or more text strings containing the replacement value for matched patterns.}
}
\value{
Returns a character vector of the new repo directory paths, or \code{NA} if the rename
failed.
}
\description{
This function is meant to help with renaming local student repos to include
something more useful like \verb{Last, First name} or a unique identifier
for the purposes of ordering repository folders.
}
\examples{
\dontrun{
# Prefix each cloned repo directory with a roster identifier
local_repo_rename(
  "/path/to/cloned/repos",
  pattern = "hw1-(.*)",
  replacement = "\\\\1"
)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/team_roster.R
\name{team_roster}
\alias{team_roster}
\title{Add team assignments to a roster}
\usage{
team_roster(
  roster,
  size,
  name = NULL,
  by = NULL,
  col = "team",
  pad = TRUE,
  shuffle = TRUE,
  seed = NULL
)
}
\arguments{
\item{roster}{Data frame. Course roster with one row per student.}

\item{size}{Integer. Target (maximum) team size. Within each group members are
split into \code{ceiling(n / size)} teams of as-equal-as-possible size. An error
is thrown if a group has fewer than \code{size} members.}

\item{name}{Character. A glue template for the team / repo name, used as-is if
supplied. May reference any column of \code{roster} as well as \code{team_id} (the
within-group team number). If \code{NULL} (default) the template is
\code{"team{team_id}"}, or \code{"{group}-team{team_id}"} when \code{by} is set.}

\item{by}{Character. Optional column name(s) in \code{roster} to form teams within
(e.g. a lab section). If \code{NULL} (default) teams are formed across the whole roster.}

\item{col}{Character. Name of the output column to add. Default \code{"team"}.}

\item{pad}{Logical or integer. Zero-padding applied to \code{team_id} before it is
substituted into \code{name}. \code{TRUE} (default) pads to a consistent width based on
the largest team number, \code{FALSE} leaves it as an integer, or supply an integer
for a fixed width.}

\item{shuffle}{Logical. Should members be randomly assigned to teams? Default \code{TRUE}.}

\item{seed}{Integer. Optional random seed for reproducible assignments.}
}
\value{
The \code{roster} data frame with \code{team_id} and \code{col} columns added.
}
\description{
\code{team_roster()} adds a team (or repo) name column to a roster data frame by
randomly grouping members into teams of a target \code{size}. Teams may be formed
within groups (e.g. lab sections) via \code{by}, and names are built from a
\link[glue:glue_data]{glue} template that can reference any column of \code{roster}
together with the generated \code{team_id}. When \code{by} is supplied and no \code{name} is
given, the grouping column(s) are folded into the default name so team names
stay unique across groups.
}
\examples{
\dontrun{
roster = readr::read_csv(system.file("roster.csv", package = "ghclass"))

# Teams of 4 across the whole roster -> "team01", "team02", ...
team_roster(roster, size = 4)

# Teams of 4 within each lab section, group folded into the default name
team_roster(roster, size = 4, by = "section", seed = 20250901)

# Custom name template (used exactly as written)
team_roster(
  roster, size = 4, by = "section",
  name = "hw1_lab{section}_team{team_id}", seed = 20250901
)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/action.R, R/action_badge.R
\name{action_badge}
\alias{action_badge}
\alias{action_add_badge}
\alias{action_remove_badge}
\title{Add or remove GitHub Actions badges from a repository}
\usage{
action_add_badge(
  repo,
  workflow = NULL,
  where = "^.",
  line_padding = "\\n\\n\\n",
  file = "README.md"
)

action_remove_badge(repo, workflow_pat = ".*?", file = "README.md")
}
\arguments{
\item{repo}{Character. Address of repository in \code{owner/name} format.}

\item{workflow}{Character. Name of the workflow.}

\item{where}{Character. Regex pattern indicating where to insert the badge, defaults
to the beginning of the target file.}

\item{line_padding}{Character. What text should be added after the badge.}

\item{file}{Character. Target file to be modified, defaults to \code{README.md}.#'}

\item{workflow_pat}{Character. Name of the workflow to be removed, or a regex pattern
that matches the workflow name.}
}
\value{
Both \code{action_add_badge()} and \code{action_remove_badge()} invisibly return a list
containing the results of the relevant GitHub API call.
}
\description{
\itemize{
\item \code{action_add_badge()} - Add a GitHub Actions badge to a file.
\item \code{action_remove_badge()} - Remove one or more GitHub Action badges from a file.
}
}
\note{
\code{action_add_badge()} is not idempotent; re-running it adds another badge
block, so use \code{action_remove_badge()} first if you need to avoid duplicates.
}
\examples{
\dontrun{
action_add_badge("ghclass-test/hw1-team01")

action_remove_badge("ghclass-test/hw1-team01")
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/action.R, R/action_artifacts.R,
%   R/action_artifacts_delete.R, R/action_artifacts_download.R,
%   R/action_run_logs.R, R/action_runs.R, R/action_runtime.R,
%   R/action_workflows.R
\name{action}
\alias{action}
\alias{action_artifacts}
\alias{action_artifact_delete}
\alias{action_artifact_download}
\alias{action_run_logs}
\alias{action_runs}
\alias{action_status}
\alias{action_runtime}
\alias{action_workflows}
\title{Retrieve information about GitHub Actions workflows and their runs.}
\usage{
action_artifacts(
  repo,
  filter = NULL,
  exclude = FALSE,
  keep_expired = FALSE,
  which = c("latest", "all"),
  filter_branch = NULL,
  exclude_branch = FALSE
)

action_artifact_delete(repo, ids)

action_artifact_download(
  repo,
  dir,
  filter = NULL,
  exclude = FALSE,
  filter_branch = NULL,
  exclude_branch = FALSE,
  keep_zip = FALSE,
  nest = FALSE,
  overwrite = FALSE,
  ids = action_artifacts(repo, filter = filter, exclude = exclude, filter_branch =
    filter_branch, exclude_branch = exclude_branch)
)

action_run_logs(
  repo,
  dir,
  run_ids = action_runs(repo),
  keep_zip = FALSE,
  overwrite = FALSE
)

action_runs(
  repo,
  branch = NULL,
  event = NULL,
  status = NULL,
  created = NULL,
  limit = 1
)

action_status(
  repo,
  branch = NULL,
  event = NULL,
  status = NULL,
  created = NULL,
  limit = 1
)

action_runtime(
  repo,
  branch = NULL,
  event = NULL,
  status = NULL,
  created = NULL,
  limit = 1
)

action_workflows(repo, full = FALSE)
}
\arguments{
\item{repo}{Character. Address of repository in \code{owner/name} format.}

\item{filter}{Character. Regex pattern to filter artifacts by artifact name.}

\item{exclude}{Logical. If \code{TRUE}, exclude artifact names matching \code{filter} instead of including them. Default \code{FALSE}.}

\item{keep_expired}{Logical. Should expired artifacts be returned. Default \code{FALSE}.}

\item{which}{Character. Either \code{"latest"} to return only the most recent of each
artifact or \code{"all"} to return all artifacts.}

\item{filter_branch}{Character. Regex pattern to filter artifacts by branch name.}

\item{exclude_branch}{Logical. If \code{TRUE}, exclude branches matching \code{filter_branch} instead of including them. Default \code{FALSE}.}

\item{ids}{Integer or data frame. Artifact ids to be downloaded or deleted.
If a data frame is passed then the \code{id} column will be used.}

\item{dir}{Character. Path to the directory where logs or artifacts will be saved.}

\item{keep_zip}{Logical. Should the log zips be saved (\code{TRUE}) or their contents extracted (\code{FALSE}). Default \code{FALSE}.}

\item{nest}{Logical. If \code{TRUE}, place each repo's artifacts under a
\verb{dir/\{repo_name\}/} subdirectory; otherwise flatten into \verb{dir/} with
repo-prefixed names. Default \code{FALSE}.}

\item{overwrite}{Logical. Should existing files be overwritten. Default \code{FALSE}.}

\item{run_ids}{Integer or data frame. Run ids for which to download logs.
If a data frame is passed then the \code{run_id} column will be used.
Defaults to the most recent run for each repo (i.e. \code{action_runs(repo)}).}

\item{branch}{Character. Filter runs associated with a particular branch.}

\item{event}{Character. Filter runs for triggered by a specific event. See
\href{https://docs.github.com/en/actions/automating-your-workflow-with-github-actions/events-that-trigger-workflows}{here}
for possible event names.}

\item{status}{Character. Filter runs for a particular status or conclusion (e.g. \code{completed} or \code{success}).}

\item{created}{Character. Filter runs for a given creation date.
See \href{https://docs.github.com/en/search-github/getting-started-with-searching-on-github/understanding-the-search-syntax#query-for-dates}{here}
for date query syntax.}

\item{limit}{Numeric. Maximum number of workflow runs to return. Default \code{1}. Note results
are chronologically ordered, so \code{limit = 1} will return the most recent action run for a repository.}

\item{full}{Logical. Should all workflow columns be returned. Default \code{FALSE}.}
}
\value{
\code{action_workflows()}, \code{action_runs()}, \code{action_runtime()}, and \code{action_artifacts}
all return tibbles containing information on requested repos' available workflows,
recent workflow runs, workflow runs runtimes, and generated artifacts
respectively.

\code{action_artifact_download()} returns an invisible character vector containing the
paths of the extraction directories for each successfully downloaded artifact
(\code{NA} entries indicate failures).

\code{action_artifact_delete()} returns an invisible data frame containing repository names and
ids of the deleted artifacts.

\code{action_run_logs()} returns an invisible character vector containing the paths of all
downloaded log files or directories.
}
\description{
\itemize{
\item \code{action_workflows()} - retrieve details on repo workflows.
\item \code{action_runs()} - retrieve details on repo workflow runs.
\item \code{action_status()} - DEPRECATED - retrieve details on most recent workflow runs.
\item \code{action_runtime()} - retrieves runtime durations for workflow runs.
\item \code{action_artifacts()} - retrieve details on available workflow artifacts.
\item \code{action_artifact_download()} - downloads artifact(s) into a local directory.
\item \code{action_artifact_delete()} - deletes artifact(s).
\item \code{action_run_logs()} - downloads workflow run log(s) into a local directory.
}
}
\examples{
\dontrun{
action_workflows("rundel/ghclass")

action_runs("rundel/ghclass")

action_runtime(c("rundel/ghclass", "rundel/parsermd"))

action_artifacts(c("rundel/ghclass", "rundel/parsermd"))
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/org_grade_assignment.R
\name{org_grade_assignment}
\alias{org_grade_assignment}
\title{Setup grading for an assignment}
\usage{
org_grade_assignment(
  path,
  org,
  repo_filter,
  artifacts = character(),
  comment_template = "",
  key_repo = NULL
)
}
\arguments{
\item{path}{Character. Root directory for the grading folder (created if it doesn't exist).}

\item{org}{Character. Name of the GitHub organization.}

\item{repo_filter}{Character. Regex pattern passed to \code{\link[=org_repos]{org_repos()}}'s \code{filter}
argument to select repos (e.g. \code{"hw01_"}).}

\item{artifacts}{Named character vector. Names are subfolder names, values are
regex patterns passed as \code{filter} to \code{\link[=action_artifact_download]{action_artifact_download()}} (matched
against artifact names). E.g. \code{c("html" = "html-output", "report" = "pdf-report")}.}

\item{comment_template}{Character. Template text used to populate each comment
markdown file. Defaults to \code{""}.}

\item{key_repo}{Character. Optional repository address in \code{owner/name} format
to clone into the root of the grading folder as an answer key.}
}
\value{
An invisible list containing:
\itemize{
\item \code{repos} — character vector of matched repo addresses
\item \code{cloned} — result from \code{\link[=local_repo_clone]{local_repo_clone()}}
\item \code{artifacts} — named list of results from each \code{\link[=action_artifact_download]{action_artifact_download()}} call
\item \code{comments} — character vector of comment file paths created
\item \code{key} — result from cloning the key repo (if provided)
}
}
\description{
This is a higher level function that automates the common steps for
setting up grading of an assignment:
\itemize{
\item Find repos matching a filter pattern in an organization
\item Clone all matched repos locally
\item Download GitHub Actions artifacts (e.g. rendered html, md files)
\item Report any missing artifacts
\item Create comment template files for each repo
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/branch.R, R/branch_create.R, R/branch_delete.R
\name{branch}
\alias{branch}
\alias{branch_create}
\alias{branch_delete}
\alias{branch_remove}
\title{Create and delete branches in a repository}
\usage{
branch_create(repo, branch, new_branch)

branch_delete(repo, branch)

branch_remove(repo, branch)
}
\arguments{
\item{repo}{GitHub repository address in \code{owner/repo} format.}

\item{branch}{Repository branch to use.}

\item{new_branch}{Name of branch to create.}
}
\value{
\code{branch_create()} and \code{branch_delete()} invisibly return a list containing
the results of the relevant GitHub API call.
}
\description{
\itemize{
\item \code{branch_create()} - creates a new branch from an existing GitHub repo.
\item \code{branch_delete()} - deletes a branch from an existing GitHub repo.
\item \code{branch_remove()} - previous name of \code{branch_delete}, deprecated.
}
}
\examples{
\dontrun{
repo_create("ghclass-test", "test_branch", auto_init=TRUE)

branch_create("ghclass-test/test_branch", branch = "main", new_branch = "test")
repo_branches("ghclass-test/test_branch")

branch_delete("ghclass-test/test_branch", branch="test")
repo_branches("ghclass-test/test_branch")

repo_delete("ghclass-test/test_branch", prompt = FALSE)
}

}
\seealso{
\link{repo_branches}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/github_whoami.R
\name{github_orgs}
\alias{github_orgs}
\title{Collect details on the authenticated user's GitHub organization memberships
(based on the current PAT).}
\usage{
github_orgs(quiet = FALSE)
}
\arguments{
\item{quiet}{Logical. Should status messages be shown. Default \code{FALSE}.}
}
\value{
Returns a tibble with organization details.
}
\description{
Collect details on the authenticated user's GitHub organization memberships
(based on the current PAT).
}
\examples{
\dontrun{
github_orgs()
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/org.R, R/org_allows_forking.R, R/org_sitrep.R,
%   R/org_update.R, R/org_workflow_permissions.R
\name{org_perm}
\alias{org_perm}
\alias{org_allows_forking}
\alias{org_sitrep}
\alias{org_set_repo_permission}
\alias{org_workflow_permissions}
\alias{org_set_workflow_permissions}
\title{Organization permissions}
\usage{
org_allows_forking(org)

org_sitrep(org)

org_set_repo_permission(
  org,
  repo_permission = c("none", "read", "write", "admin")
)

org_workflow_permissions(org)

org_set_workflow_permissions(org, workflow_permission = c("read", "write"))
}
\arguments{
\item{org}{Character. Name of the GitHub organization(s).}

\item{repo_permission}{Default permission level members have for organization repositories:
\itemize{
\item read - can pull, but not push to or administer this repository.
\item write - can pull and push, but not administer this repository.
\item admin - can pull, push, and administer this repository.
\item none - no permissions granted by default.
}}

\item{workflow_permission}{The default workflow permissions granted to the GITHUB_TOKEN when
running workflows in the organization. Accepted values:\code{"read"} or \code{"write"}.}
}
\value{
\code{org_sitep()} invisibly returns the \code{org} argument.

\code{org_set_repo_permission()} invisibly return a the result of the relevant GitHub API call.

\code{org_workflow_permissions()} returns a character vector with value of either \code{"read"} or \code{"write"}.

\code{org_set_workflow_permissions()} invisibly return a the result of the relevant GitHub API call.

\code{org_allows_forking()} returns a logical scalar.
}
\description{
\itemize{
\item \code{org_sitrep()} - Provides a situation report on a GitHub organization.
\item \code{org_set_repo_permission()} - Change the default permission level for org repositories.
\item \code{org_workflow_permissions()} - Obtain the current default workflow permission value
for the organization.
\item \code{org_set_workflow_permissions()} - Change the current default workflow permission value
for the organization.
\item \code{org_allows_forking()} - returns \code{TRUE} if members can fork private repositories in the organization.
}
}
\examples{
\dontrun{
org_sitrep("ghclass-test")

org_set_repo_permission("ghclass-test", "read")

org_workflow_permissions("ghclass-test")

org_set_workflow_permissions("ghclass-test", "write")

org_sitrep("ghclass-test")

# Cleanup
org_set_repo_permission("ghclass-test", "none")
org_set_workflow_permissions("ghclass-test", "read")
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/issue.R, R/issue_close.R, R/issue_create.R,
%   R/issue_edit.R
\name{issue}
\alias{issue}
\alias{issue_close}
\alias{issue_create}
\alias{issue_edit}
\title{GitHub Issue related tools}
\usage{
issue_close(repo, number, comment = NULL)

issue_create(
  repo,
  title,
  body,
  labels = character(),
  assignees = character(),
  delay = 0
)

issue_edit(
  repo,
  number,
  title = NULL,
  body = NULL,
  state = NULL,
  milestone = NULL,
  labels = list(),
  assignees = list()
)
}
\arguments{
\item{repo}{Character. Address of one or more repositories in \code{owner/name} format.}

\item{number}{Integer. GitHub issue number.}

\item{comment}{Character. Optional comment to post on the issue before
closing it. If posting the comment fails the issue is not closed. Recycled
against \code{repo} / \code{number}.}

\item{title}{Character. Title of the issue.}

\item{body}{Character. Content of the issue.}

\item{labels}{Character. Vector of the labels to associate with this issue}

\item{assignees}{Character. Vector of logins for users assigned to the issue.}

\item{delay}{Numeric. Delay between each API request. Issue creation has a secondary rate limit (~ 20/min).}

\item{state}{Character. State of the issue. Either "open" or "closed".}

\item{milestone}{Character. The number of the milestone to associate this issue with.
Only users with push access can set the milestone for issues. The milestone is silently dropped otherwise.}
}
\value{
All functions invisibly return a list containing the results of the relevant GitHub API call.
}
\description{
\itemize{
\item \code{issue_create} creates a new issue.
\item \code{issue_close} closes an existing issue.
\item \code{issue_edit} edits the properties of an existing issue.
}
}
\examples{
\dontrun{
repo_create("ghclass-test","test_issue")

issue_create(
  "ghclass-test/test_issue",
  title = "Issue 1",
  body = "This is an issue"
 )

issue_create(
  "ghclass-test/test_issue",
  title = "Issue 2", body = "This is also issue",
  label = "Important"
)

issue_create(
  "ghclass-test/test_issue",
  title = "Issue 3", body = "This is also issue",
  label = c("Important", "Super Important"),
  assignees = "rundel"
)

issue_close("ghclass-test/test_issue", 1)

issue_edit(
  "ghclass-test/test_issue", 2,
  title = "New issue 2 title!",
  body = "Replacement body text"
)

ghclass::repo_issues("ghclass-test/test_issue")

repo_delete("ghclass-test/test_issue", prompt=FALSE)
}
}
\seealso{
\link{repo_issues}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/repo.R, R/repo_allows_forking.R,
%   R/repo_create.R, R/repo_delete.R, R/repo_exists.R, R/repo_is_template.R,
%   R/repo_mirror.R, R/repo_mirror_template.R, R/repo_rename.R,
%   R/repo_set_forking.R, R/repo_set_template.R
\name{repo_core}
\alias{repo_core}
\alias{repo_allows_forking}
\alias{repo_create}
\alias{repo_delete}
\alias{repo_exists}
\alias{repo_is_template}
\alias{repo_mirror}
\alias{repo_mirror_template}
\alias{repo_rename}
\alias{repo_set_forking}
\alias{repo_set_template}
\title{GitHub Repository tools - core functions}
\usage{
repo_allows_forking(repo)

repo_create(
  org,
  name,
  prefix = "",
  suffix = "",
  private = TRUE,
  auto_init = FALSE,
  gitignore_template = "R"
)

repo_delete(repo, prompt = TRUE)

repo_exists(repo, strict = FALSE, quiet = FALSE)

repo_is_template(repo)

repo_mirror(
  source_repo,
  target_repo,
  overwrite = FALSE,
  verbose = FALSE,
  warn = TRUE
)

repo_mirror_template(source_repo, target_repo, private = TRUE)

repo_rename(repo, new_repo)

repo_set_forking(repo, status = TRUE)

repo_set_template(repo, status = TRUE)
}
\arguments{
\item{repo}{Character. Address of repository in \code{owner/repo} format.}

\item{org}{Character. GitHub organization that will own the repository}

\item{name}{Character. Repository name}

\item{prefix}{Character. Common repository name prefix}

\item{suffix}{Character. Common repository name suffix}

\item{private}{Logical. Should the new repository be private or public. Default \code{TRUE}.}

\item{auto_init}{Logical. Should the repository be initialized with a \code{README.md}. Default \code{FALSE}.}

\item{gitignore_template}{Character. \code{.gitignore} language template to use.}

\item{prompt}{Logical. Should the user be prompted before deleting repositories. Default \code{true}.}

\item{strict}{Logical. Should the old name of a renamed repositories be allowed. Default \code{FALSE}.}

\item{quiet}{Logical. Should details on renamed repositories be printed. Default \code{FALSE}.}

\item{source_repo}{Character. Address of template repository in \code{owner/name} format.}

\item{target_repo}{Character. One or more repository addresses in \code{owner/name} format.
Note when using template repos these new repositories must \emph{not} exist.}

\item{overwrite}{Logical. Should the target repositories be overwritten. Default \code{FALSE}.}

\item{verbose}{Logical. Display verbose output. Default \code{FALSE}.}

\item{warn}{Logical. Warn the user about the function being deprecated. Default \code{TRUE}.}

\item{new_repo}{Character. New name of repository without the owner.}

\item{status}{Logical. Should the repository be set as a template repository. Default \code{TRUE}.}
}
\value{
\code{repo_create()} returns a character vector of created repos (in \code{owner/repo} format)

\code{repo_exists()}, \code{repo_is_template()}, and \code{repo_allows_forking()} all return a logical vector.

\code{repo_set_forking()} invisibly returns a list containing the results of the relevant GitHub API calls.

All other functions invisibly return a list containing the results of the relevant GitHub API calls.
}
\description{
\itemize{
\item \code{repo_create()} - create a GitHub repository.
\item \code{repo_delete()} - delete a GitHub repository.
\item \code{repo_rename()} - rename a repository, note that renamed repositories retain their
unique identifier and can still be accessed via their old names due to
GitHub re-directing.
\item \code{repo_exists()} - returns \code{TRUE} if the GitHub repository exists. It will also print
a message if a repository has been renamed, unless \code{quiet = TRUE}.
\item \code{repo_mirror()} - mirror the content of a repository to another repository,
the target repo must already exist.
\item \code{repo_mirror_template()} - mirror the content of a source template repository to a new repository,
the target repo must \emph{not} already exist.
\item \code{repo_is_template()} - returns \code{TRUE} if a repository is a template repo.
\item \code{repo_set_template()} - change the template status of a repository.
\item \code{repo_allows_forking()} - returns \code{TRUE} if forking is enabled for a repository.
\item \code{repo_set_forking()} - enable or disable forking for a repository.
}
}
\examples{
\dontrun{
repo_create("ghclass-test", "repo_test")

repo_exists("ghclass-test/repo_test")

repo_rename("ghclass-test/repo_test", "repo_test_new")


# The new repo exists
repo_exists("ghclass-test/repo_test_new")

# The old repo forwards to the new repo
repo_exists("ghclass-test/repo_test")


# Check for the redirect by setting `strict = TRUE`
repo_exists("ghclass-test/repo_test", strict = TRUE)


# The prefered way of copying a repo is by making the source a template
repo_is_template("ghclass-test/repo_test_new")

repo_set_template("ghclass-test/repo_test_new")

repo_is_template("ghclass-test/repo_test_new")


# Given a template repo we can then directly copy the repo on GitHub
repo_mirror_template("ghclass-test/repo_test_new", "ghclass-test/repo_test_copy")

repo_exists("ghclass-test/repo_test_copy")

# Cleanup
repo_delete(
 c("ghclass-test/repo_test_new",
   "ghclass-test/repo_test_copy"),
 prompt = FALSE
)

}


}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/org.R, R/org_exists.R, R/org_repo_forking.R,
%   R/org_repo_search.R, R/org_repo_stats.R, R/org_repos.R,
%   R/org_team_details.R, R/org_teams.R, R/org_user_repos.R
\name{org_details}
\alias{org_details}
\alias{org_exists}
\alias{org_repo_forking}
\alias{org_repo_search}
\alias{org_repo_stats}
\alias{org_repos}
\alias{org_team_details}
\alias{org_teams}
\alias{org_user_repos}
\title{Obtain details on an organization's repos and teams}
\usage{
org_exists(org)

org_repo_forking(org)

org_repo_search(org, name, extra = "", full_repo = TRUE)

org_repo_stats(
  org,
  branch = NULL,
  filter = "",
  filter_type = "in:name",
  inc_commits = TRUE,
  inc_issues = TRUE,
  inc_prs = TRUE
)

org_repos(
  org,
  filter = NULL,
  exclude = FALSE,
  full_repo = TRUE,
  sort = c("full_name", "created", "updated", "pushed"),
  direction = c("asc", "desc"),
  type = c("all", "public", "private", "forks", "sources", "member", "internal")
)

org_team_details(org)

org_teams(org, filter = NULL, exclude = FALSE, team_type = c("name", "slug"))

org_user_repos(org, user, prefix = "", suffix = "")
}
\arguments{
\item{org}{Character. Name of the GitHub organization(s).}

\item{name}{Character. Full or partial repo name to search for within the org}

\item{extra}{Character. Any additional search qualifiers, see
\href{https://docs.github.com/en/free-pro-team@latest/github/searching-for-information-on-github/searching-for-repositories}{Searching for repositories}
for details.}

\item{full_repo}{Logical. Should the full repository address be returned (e.g. \code{owner/repo} instead of just \code{repo}). Default \code{TRUE}.}

\item{branch}{Character. The branch to use for counting commits, if \code{NULL} then each repo's default branch is used.}

\item{filter}{Character. Regular expression pattern for matching (or excluding) results}

\item{filter_type}{Character. One or more GitHub search \verb{in} qualifiers.
See \href{https://help.github.com/en/articles/searching-for-repositories}{documentation}
for more details.}

\item{inc_commits}{Logical. Include commit statistics (\code{branch}, \code{commits}, \code{last_update}). Default \code{TRUE}.}

\item{inc_issues}{Logical. Include issue statistics (\code{open_issues}, \code{closed_issues}). Default \code{TRUE}.}

\item{inc_prs}{Logical. Include pull request statistics (\code{open_prs}, \code{merged_prs}, \code{closed_prs}). Default \code{TRUE}.}

\item{exclude}{Logical. Should entries matching the regular expression be excluded or included. Default \code{FALSE}.}

\item{sort}{Character. Sorting criteria to use, can be one of "created", "updated", "pushed", or "full_name".}

\item{direction}{Character. Sorting order to use.}

\item{type}{Character. Specifies the type of repositories you want, can be one of
"all", "public", "private", "forks", "sources", "member", or "internal".}

\item{team_type}{Character. Either "slug" if the team names are slugs or "name" if full team names are provided.}

\item{user}{Character. One or more GitHub usernames.}

\item{prefix}{Character. String to prepend to each user name.}

\item{suffix}{Character. String to append to each user name.}
}
\value{
\code{org_exists()} returns a logical vector.

\code{org_teams()}, \code{org_repos}, \code{org_repo_search()}, and \code{org_user_repos()} return a character vector.

\code{org_team_details()}, \code{org_repo_stats()}, and \code{org_repo_forking()} return tibbles.
}
\description{
\itemize{
\item \code{org_exists()} - returns \code{TRUE} if the organization(s) exist on GitHub and \code{FALSE} otherwise.
\item \code{org_teams()} - returns a (filtered) vector of organization teams.
\item \code{org_team_details()} - returns a data frame of all organization teams containing identification and permission details.
\item \code{org_repos()} - returns a (filtered) vector of organization repositories.
\item \code{org_repo_search()} - search for repositories within an organization (preferred for large organizations).
\item \code{org_repo_stats()} - returns a tibble of repositories belonging to a GitHub organization along with some
basic statistics about those repositories.
\item \code{org_repo_forking()} - returns a tibble of private repositories in an organization and whether forking is enabled for each.
\item \code{org_user_repos()} - constructs a character vector of per-user repo addresses
(\verb{org/prefix<user>suffix}) aligned with \code{user}, for safely pairing with
\code{\link[=repo_add_user]{repo_add_user()}} and similar.
}
}
\section{Note}{


The order of \code{org_repos()} is not guaranteed to match the order of
\code{org_members()}, so do not rely on positional pairing of the two when
calling functions like \code{\link[=repo_add_user]{repo_add_user()}}. To pair each user with their own
repo deterministically, build the repo names from the user vector with
\code{org_user_repos()}.

\code{org_repo_search()} and \code{org_repo_stats()} use GitHub's search API, which
returns at most 1000 results per query. For organizations with more than 1000
matching repositories the results are truncated and a warning is issued; use
\code{filter} (or \code{name} / \code{extra}) to narrow the search.
}

\examples{
\dontrun{
# Org repos and teams
org_repos("ghclass-test")

org_repos("ghclass-test", filter = "hw1-")

org_teams("ghclass-test")

org_team_details("ghclass-test")
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/org_create_assignment.R
\name{org_create_assignment}
\alias{org_create_assignment}
\title{Create a team or individual assignment}
\usage{
org_create_assignment(
  org,
  repo,
  user,
  team = NULL,
  source_repo = NULL,
  private = TRUE,
  add_badges = FALSE,
  ignore_existing = FALSE
)
}
\arguments{
\item{org}{Character. Name of the GitHub organization.}

\item{repo}{Character. Name of the repo(s) for the assignment.}

\item{user}{Character. GitHub username(s).}

\item{team}{Character. Team names, if not provided an individual assignment will be created.}

\item{source_repo}{Character. Address of the repository to use as a template for all created repos.}

\item{private}{Logical. Should the created repositories be private. Default \code{TRUE}.}

\item{add_badges}{Logical. Should GitHub action badges be added to the README. Default \code{FALSE}.}

\item{ignore_existing}{Logical. If \code{TRUE}, any requested repos that already
exist are skipped (with a warning) and the remaining repos are still
created; useful for resuming a partially failed run. Default \code{FALSE}.}
}
\value{
An invisible list containing the results of each step.
}
\description{
This is a higher level function that combines the following steps:
\itemize{
\item Create repos
\item Create teams and invite students if necessary
\item Add teams or individuals to the repositories
\item Mirror a template repository to assignment repositories
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/github_api_limit.R
\name{github_api_limit}
\alias{github_api_limit}
\alias{github_get_api_limit}
\alias{github_set_api_limit}
\alias{github_get_max_wait}
\alias{github_set_max_wait}
\alias{github_get_max_rate}
\alias{github_set_max_rate}
\title{Tools for limiting gh's GitHub api requests.}
\usage{
github_get_api_limit()

github_set_api_limit(limit = 10000L)

github_get_max_wait()

github_set_max_wait(max_wait = NULL)

github_get_max_rate()

github_set_max_rate(max_rate = NULL)
}
\arguments{
\item{limit}{The maximum number of records to return from an API request.}

\item{max_wait}{The maximum number of seconds to wait when rate limited, or \code{NULL}
to use gh's default. Passed to \code{\link[gh:gh]{gh::gh()}}'s \code{.max_wait}.}

\item{max_rate}{The maximum request rate in requests per second, or \code{NULL} for no
throttling. Passed to \code{\link[gh:gh]{gh::gh()}}'s \code{.max_rate}.}
}
\value{
\code{github_get_api_limit()} returns a single integer value.

\code{github_get_max_wait()} and \code{github_get_max_rate()} return their stored value or \code{NULL}.

\verb{github_set_*()} functions invisibly return the value of their argument.
}
\description{
\itemize{
\item \code{github_get_api_limit()} - returns the current limit on results returned by gh.
\item \code{github_set_api_limit()} - sets a limit on results returned by gh.
\item \code{github_get_max_wait()} / \code{github_set_max_wait()} - get or set the maximum number
of seconds gh will wait when it hits a rate limit. \code{NULL} (the default) uses gh's
own default.
\item \code{github_get_max_rate()} / \code{github_set_max_rate()} - get or set the maximum request
rate, in requests per second, used to proactively throttle requests and avoid
secondary rate limits. \code{NULL} (the default) disables throttling.
}
}
\details{
These values are stored in the \code{"ghclass.api.limit"}, \code{"ghclass.max.wait"},
and \code{"ghclass.max.rate"} options globally.
}
\examples{
github_get_api_limit()

github_set_api_limit(500)

github_get_api_limit()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/github_with_pat.R
\name{github_with_pat}
\alias{github_with_pat}
\alias{with_pat}
\alias{local_pat}
\title{\code{withr}-like functions for temporary personal access token}
\usage{
with_pat(new, code)

local_pat(new, .local_envir = parent.frame())
}
\arguments{
\item{new}{Temporary GitHub access token}

\item{code}{Code to execute with the temporary token}

\item{.local_envir}{The environment to use for scoping.}
}
\value{
The results of the evaluation of the code argument.
}
\description{
Temporarily change the \code{GITHUB_PAT} environmental variable for
GitHub authentication. Based on the \code{withr} interface.
}
\details{
if \code{new = NA} is used the \code{GITHUB_PAT} environment variable will be unset.
}
\examples{
\dontrun{
with_pat("1234", print(github_get_token()))
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pr.R, R/pr_create.R
\name{pr}
\alias{pr}
\alias{pr_create}
\title{GitHub Pull Request related tools}
\usage{
pr_create(repo, title, head, base, body = "", draft = FALSE)
}
\arguments{
\item{repo}{Character. Address of one or more repositories in "owner/name" format.}

\item{title}{Character. Title of the pull request.}

\item{head}{Character. The name of the branch where your changes are implemented.
For cross-repository pull requests in the same network, namespace \code{head} with a user
like this: \code{username:branch}.}

\item{base}{Character. The name of the branch you want the changes pulled into.
This should be an existing branch on the current repository. You cannot submit
a pull request to one repository that requests a merge to a base of another repository.}

\item{body}{Character. The text contents of the pull request.}

\item{draft}{Logical. Should the pull request be created as a draft pull request
(these cannot be merged until allowed by the author). Default \code{FALSE}.}
}
\value{
\code{pr_create()} invisibly return a list containing the results of the relevant GitHub API calls.
}
\description{
\itemize{
\item \code{pr_create()} - create a pull request GitHub from the \code{base} branch to the \code{head} branch.
}
}
\examples{
\dontrun{
repo_create("ghclass-test", "test_pr", auto_init=TRUE)

branch_create("ghclass-test/test_pr", branch = "main", new_branch = "test")

repo_modify_file("ghclass-test/test_pr", "README.md", pattern = "test_pr",
                 content = "Hello", method = "after", branch = "test")

pr_create("ghclass-test/test_pr", title = "merge", head = "test", base = "main")

repo_delete("ghclass-test/test_pr", prompt = FALSE)
}
}
\seealso{
\link{repo_issues}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/team.R, R/team_create.R, R/team_delete.R,
%   R/team_rename.R
\name{team}
\alias{team}
\alias{team_create}
\alias{team_delete}
\alias{team_rename}
\title{Create, delete, and rename teams within an organization}
\usage{
team_create(
  org,
  team,
  prefix = "",
  suffix = "",
  privacy = c("secret", "closed")
)

team_delete(org, team, team_type = c("name", "slug"), prompt = TRUE)

team_rename(org, team, new_team, team_type = c("name", "slug"))
}
\arguments{
\item{org}{Character. Name of the GitHub organization.}

\item{team}{Character. Name of teams.}

\item{prefix}{Character. Shared prefix.}

\item{suffix}{Character. Shared suffix.}

\item{privacy}{Character. Level of privacy for team, "closed" (visible to all
members of the organization) or "secret" (only visible to organization owners
and members of a team), default is "secret"}

\item{team_type}{Character. Either "slug" if the team names are slugs or "name" if full team names are provided.}

\item{prompt}{Logical. Should the user be prompted before deleting team. Default \code{true}.}

\item{new_team}{character, new team name.}
}
\value{
All functions invisibly return a list containing
the results of the relevant GitHub API calls.
}
\description{
\itemize{
\item \code{team_create()} - create teams in a GitHub organization
\item \code{team_delete()} - delete a team from a GitHub organization.
\item \code{team_rename()} - rename an existing team
}
}
\examples{
\dontrun{
team_create("ghclass-test",c("hw1-team01","hw1-team02"))

org_teams("ghclass-test", "hw1-")

team_rename("ghclass-test", "hw1-team02", "hw1-team03")

org_teams("ghclass-test", "hw1-")

team_delete("ghclass-test", "hw1-team01", prompt = FALSE)

org_teams("ghclass-test", "hw1-")


# Cleanup
team_delete("ghclass-test", org_teams("ghclass-test", "hw1-"), prompt = FALSE)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/local_repo.R, R/local_repo_add.R,
%   R/local_repo_branch.R, R/local_repo_clone.R, R/local_repo_commit.R,
%   R/local_repo_log.R, R/local_repo_pull.R, R/local_repo_push.R
\name{local_repo}
\alias{local_repo}
\alias{local_repo_add}
\alias{local_repo_branch}
\alias{local_repo_clone}
\alias{repo_clone}
\alias{local_repo_commit}
\alias{local_repo_log}
\alias{local_repo_pull}
\alias{local_repo_push}
\title{Functions for managing local git repositories}
\usage{
local_repo_add(repo_dir, files = ".")

local_repo_branch(repo_dir, branch)

local_repo_clone(
  repo,
  local_path = ".",
  branch = NULL,
  mirror = FALSE,
  verbose = FALSE
)

local_repo_commit(repo_dir, message)

local_repo_log(repo_dir, max = 100)

local_repo_pull(repo_dir, verbose = FALSE)

local_repo_push(
  repo_dir,
  remote = "origin",
  branch = NULL,
  force = FALSE,
  prompt = TRUE,
  mirror = FALSE,
  verbose = FALSE
)
}
\arguments{
\item{repo_dir}{Vector of repo directories or a single directory containing one or more repos.}

\item{files}{Files to be staged}

\item{branch}{Repository branch to use.}

\item{repo}{GitHub repo address with the form \code{owner/name}.}

\item{local_path}{Local directory to store cloned repos.}

\item{mirror}{Equivalent to \code{--mirror}}

\item{verbose}{Display verbose output.}

\item{message}{Commit message}

\item{max}{Maximum number of log entries to retrieve per repo.}

\item{remote}{Repository remote to use.}

\item{force}{Force push?}

\item{prompt}{Prompt before force push?}
}
\value{
\code{local_repo_clone()} invisibly returns a character vector of paths for
the local repo directories.

\code{local_repo_log()} returns a tibble containing repository details.`

All other functions invisibly return a list containing the results of the relevant call
to \code{gert}.
}
\description{
\itemize{
\item \code{local_repo_clone()} - Clones a GitHub repository to a local directory.
\item \code{local_repo_add()} - Equivalent to \verb{git add} - stages a file in a local repository.
\item \code{local_repo_commit()} - Equivalent to \verb{git commit} - commits staged files in a local repository.
\item \code{local_repo_push()} - Equivalent to \verb{git push} - push a local repository.
\item \code{local_repo_pull()} - Equivalent to \verb{git pull} - pull a local repository.
\item \code{local_repo_branch()} - Equivalent to \verb{git branch} - create a branch in a local repository.
\item \code{local_repo_log()} - Equivalent to \verb{git log} - returns a data frame for git log entries.
}
}
\details{
All \verb{local_repo_*} functions depend on the \code{gert} library being installed.
}
\examples{
\dontrun{
repo = repo_create("ghclass-test", "local_repo_test")

dir = file.path(tempdir(), "repos")
local_repo = local_repo_clone(repo, dir)

local_repo_log(dir)

# Make a local change and push
writeLines("Hello World", file.path(local_repo, "hello.txt"))

local_repo_add(local_repo, "hello.txt")

local_repo_commit(local_repo, "Added hello world")

local_repo_push(local_repo)

repo_commits(repo)

# Pulling remote changes
repo_modify_file(repo, "hello.txt", pattern = ".*",  content = "!!!", method = "after")

local_repo_pull(local_repo)

local_repo_log(dir)

repo_delete("ghclass-test/local_repo_test", prompt=FALSE)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/github_rate_limit.R, R/github_token.R
\name{github_rate_limit}
\alias{github_rate_limit}
\alias{github_graphql_rate_limit}
\alias{github_token}
\alias{github_get_token}
\alias{github_set_token}
\alias{github_reset_token}
\alias{github_test_token}
\alias{github_token_scopes}
\title{Tools for handling GitHub personal access tokens (PAT)}
\usage{
github_rate_limit()

github_graphql_rate_limit()

github_get_token()

github_set_token(token)

github_reset_token()

github_test_token(token = github_get_token())

github_token_scopes(token = github_get_token())
}
\arguments{
\item{token}{Character. Either the literal token, or the path to a file containing the token.}
}
\value{
\code{github_get_token()} returns the current PAT as a character string with the \code{gh_pat}
class. See \code{\link[gh:gh_token]{gh::gh_token()}} for additional details.

\code{github_set_token()} and \code{github_reset_token()} return the result of \code{Sys.setenv()} and
\code{Sys.unsetenv()} respectively.

\code{github_test_token()} invisibly returns a logical value, \code{TRUE} if the test passes,
\code{FALSE} if not.

\code{github_token_scopes()} returns a character vector of granted scopes.
}
\description{
\itemize{
\item \code{github_get_token} - returns the user's GitHub personal access token (PAT).
\item \code{github_set_token} - defines the user's GitHub PAT by setting the \code{GITHUB_PAT} environmental variable.
This value will persist until the session ends or \code{gihub_reset_token()} is called.
\item \code{github_reset_token} - removes the value stored in the \code{GITHUB_PAT} environmental variable.
\item \code{github_test_token} - checks if a PAT is valid by attempting to authenticate with the GitHub API.
\item \code{github_token_scopes} - returns a vector of scopes granted to the token.
}
}
\details{
This package looks for the personal access token (PAT) in the following places (in order):
\itemize{
\item Value of \code{GITHUB_PAT} environmental variable.
\item Any GitHub PAT token(s) stored with \code{gitcreds} via \code{gitcreds_set()}.
}

For additional details on creating a GitHub PAT see the usethis vignette on
\href{https://usethis.r-lib.org/articles/articles/git-credentials.html}{Managing Git(Hub) Credentials}.
For those who do not wish to read the entire article, the quick start method is to use:
\itemize{
\item \code{usethis::create_github_token()} - to create the token and then,
\item \code{gitcreds::gitcreds_set()} - to securely cache the token.
}
}
\examples{
\dontrun{
github_test_token()

github_token_scopes()

(pat = github_get_token())

github_set_token("ghp_BadTokenBadTokenBadTokenBadTokenBadToken")
github_get_token()
github_test_token()

github_set_token(pat)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/repo_style.R
\name{repo_style}
\alias{repo_style}
\title{Style repository with styler}
\usage{
repo_style(
  repo,
  files = c("*.R", "*.Rmd"),
  branch = "styler",
  base,
  create_pull_request = TRUE,
  draft = TRUE,
  tag_collaborators = TRUE,
  prompt = TRUE
)
}
\arguments{
\item{repo}{Character. Address of repository in "owner/name" format.}

\item{files}{Character or vector of characters. Names of .R and/or .Rmd files that styler should be applied to.}

\item{branch}{Character. Name of new branch to be created or overwritten. Default is "styler".}

\item{base}{Character. Name of branch that contains the .R and/or .Rmd files to be styled}

\item{create_pull_request}{Logical. If TRUE, a pull request is created from branch to base. Default \code{TRUE}.}

\item{draft}{Logical. Should the pull request be created as a draft pull request? (Draft PRs cannot be merged
until allowed by the author) Default \code{TRUE}.}

\item{tag_collaborators}{Logical. If TRUE, a message with the repository collaborators is displayed. Default \code{TRUE}.}

\item{prompt}{Logical. Prompt the user before overwriting an existing branch. Default \code{TRUE}.}
}
\value{
The functions returns NULL invisibly.
}
\description{
\itemize{
\item \code{repo_style} implements "non-invasive pretty-printing of R source code" of .R or .Rmd files
within a repository using the \code{styler} package and adhering to \code{tidyverse} formatting guidelines.
}
}
\examples{
\dontrun{
repo_style("ghclass-test/hw1-team01", base = "main")
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/org.R, R/org_admins.R, R/org_invite.R,
%   R/org_members.R, R/org_pending.R, R/org_remove.R
\name{org_members}
\alias{org_members}
\alias{org_admins}
\alias{org_invite}
\alias{org_pending}
\alias{org_remove}
\title{Tools for managing organization membership}
\usage{
org_admins(org)

org_invite(org, user)

org_members(org, filter = NULL, exclude = FALSE, include_admins = TRUE)

org_pending(org, filter = NULL, exclude = FALSE)

org_remove(org, user, prompt = TRUE)
}
\arguments{
\item{org}{Character. Name of the GitHub organization(s).}

\item{user}{Character. GitHub username(s).}

\item{filter}{Character. Regular expression pattern for matching (or excluding) results}

\item{exclude}{Logical. Should entries matching the regular expression be excluded or included. Default \code{FALSE}.}

\item{include_admins}{Logical. Should admin users be included in the results. Default \code{TRUE}.}

\item{prompt}{Logical. Prompt before removing member from organization. Default \code{TRUE}.}
}
\value{
\code{org_members()}, \code{org_pending()}, and \code{org_admins} all return a character vector
of GitHub account names.

\code{org_invite()} and \code{org_remove()} invisibly return a list containing the results of the relevant GitHub API calls.
}
\description{
\itemize{
\item \code{org_invite()} - invites user(s) to a GitHub organization.
\item \code{org_remove()} - remove user(s) from an organization (and all teams within that organization).
\item \code{org_members()} - returns a (filtered) vector of organization members.
\item \code{org_pending}() - returns a (filtered) vector of pending organization members.
\item \code{org_admins()} - returns a vector of repository administrators. In the case of a
non-organization owner (e.g. a user account) returns the owner's login.
}
}
\examples{
\dontrun{
# Org Details
org_admins("ghclass-test")

org_admins("rundel") # User, not an organization


# Org Membership - Invite, Status, and Remove
students = c("ghclass-anya", "ghclass-bruno", "ghclass-celine",
             "ghclass-diego", "ghclass-elijah","ghclass-francis")

org_invite("ghclass-test", students)

org_members("ghclass-test")

org_pending("ghclass-test")

org_remove("ghclass-test", students, prompt = FALSE)

org_pending("ghclass-test")
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/repo.R, R/repo_branches.R, R/repo_clone_url.R,
%   R/repo_commits.R, R/repo_issues.R, R/repo_n_commits.R, R/repo_prs.R,
%   R/repo_pushes.R
\name{repo_details}
\alias{repo_details}
\alias{repo_branches}
\alias{repo_clone_url}
\alias{repo_commits}
\alias{repo_issues}
\alias{repo_n_commits}
\alias{repo_prs}
\alias{repo_pushes}
\title{GitHub Repository tools - repository details}
\usage{
repo_branches(repo)

repo_clone_url(repo, type = c("https", "ssh"))

repo_commits(
  repo,
  branch = NULL,
  sha = branch,
  path = NULL,
  author = NULL,
  since = NULL,
  until = NULL,
  quiet = FALSE
)

repo_issues(
  repo,
  state = c("open", "closed", "all"),
  assignee = NULL,
  creator = NULL,
  mentioned = NULL,
  labels = NULL,
  sort = c("created", "updated", "comments"),
  direction = c("desc", "asc"),
  since = NULL
)

repo_n_commits(repo, quiet = FALSE)

repo_prs(repo, state = c("open", "closed", "all"), quiet = FALSE)

repo_pushes(
  repo,
  branch = NULL,
  author = NULL,
  time_period = c("all time", "day", "week", "month", "quarter", "year"),
  quiet = FALSE
)
}
\arguments{
\item{repo}{Character. Address of repository in \code{owner/name} format.}

\item{type}{Character. Clone url type, either "https" or "ssh".}

\item{branch}{Character.    Branch to list commits from.}

\item{sha}{Character.    SHA to start listing commits from.}

\item{path}{Character.    Only commits containing this file path will be returned.}

\item{author}{Character.    GitHub login or email address by which to filter commit author.}

\item{since}{Character. Only issues updated at or after this time are returned.}

\item{until}{Character.    Only commits before this date will be returned, expects \code{YYYY-MM-DDTHH:MM:SSZ} format. Filtered against the \code{committer_date} column, see Details.}

\item{quiet}{Logical. Should an error message be printed if a repo does not exist. Default \code{FALSE}.}

\item{state}{Character. Pull request state.}

\item{assignee}{Character. Return issues assigned to a particular username.
Pass in "none" for issues with no assigned user, and "*" for issues assigned to any user.}

\item{creator}{Character. Return issues created the by the given username.}

\item{mentioned}{Character. Return issues that mentioned the given username.}

\item{labels}{Character. Return issues labeled with one or more of of the given label names.}

\item{sort}{Character. What to sort results by. Can be either "created", "updated", or "comments".}

\item{direction}{Character. The direction of the sort. Can be either "asc" or "desc".}

\item{time_period}{Character. The time period to filter by.
Options are  "all time", "day", "week", "month", "quarter", "year".}
}
\value{
\code{repo_clone_url()} and \code{repo_branches()} both return a character vector.

\code{repo_commits()}, \code{repo_issues()}, \code{repo_n_commits()}, and \code{repo_prs()} all return a tibble.
}
\description{
\itemize{
\item \code{repo_clone_url()} - Returns the url, for cloning, a GitHub repo (either ssh or https)
\item \code{repo_branches()} - Returns a (filtered) vector of branch names.
\item \code{repo_commits()} - Returns a tibble of commits to a GitHub repository.
\item \code{repo_issues()} - Returns a tibble of issues for a GitHub repository.
\item \code{repo_n_commits()} - Returns a tibble of the number of commits in a GitHub repository (and branch).
\item \code{repo_prs()} - Returns a tibble of pull requests for a GitHub repository.
\item \code{repo_pushes()} - Returns a tibble of push activity to a GitHub repository.
}
}
\details{
For \code{repo_commits()}, GitHub returns two timestamps for each commit: the
author date (\code{date}) and the committer date (\code{committer_date}). These
differ for rebased, cherry-picked, or amended commits. The \code{since} and
\code{until} arguments are filtered by GitHub against the committer date.

The \code{login} column is \code{NA} when the commit's author email is not linked
to a GitHub account. The git author's \code{name} and \code{email} are still
populated in those cases.
}
\examples{
\dontrun{
repo_clone_url("rundel/ghclass")

repo_branches("rundel/ghclass")

repo_commits("rundel/ghclass")

repo_issues("rundel/ghclass")

repo_n_commits("rundel/ghclass", branch = "master")

repo_prs("rundel/ghclass")
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/github_whoami.R
\name{github_whoami}
\alias{github_whoami}
\title{Returns the login of the authenticated user (based on the current PAT).}
\usage{
github_whoami(quiet = FALSE)
}
\arguments{
\item{quiet}{Logical. Should status messages be shown. Default \code{FALSE}.}
}
\value{
Character value containing user login.
}
\description{
Returns the login of the authenticated user (based on the current PAT).
}
\examples{
\dontrun{
github_whoami()
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ghclass-internal.R, R/repo_add_file.R,
%   R/repo_get_file.R, R/util_args.R, R/util_cli.R, R/util_github.R,
%   R/util_repo.R, R/util_safely.R
\name{ghclass-internal}
\alias{ghclass-internal}
\alias{read_bin_file}
\alias{github_api_repo_get_file}
\alias{arg_is_chr_scalar}
\alias{arg_is_lgl_scalar}
\alias{arg_is_pos_int_scalar}
\alias{arg_is_chr}
\alias{arg_is_lgl}
\alias{arg_is_pos_int}
\alias{cli_glue}
\alias{cli_stop}
\alias{cli_warn}
\alias{github_api_repo_tree}
\alias{repo_files}
\alias{ghclass_api_v3_req}
\alias{match_repo}
\alias{get_repo_name}
\alias{get_repo_owner}
\alias{format_repo}
\alias{result}
\alias{error}
\alias{succeeded}
\alias{failed}
\alias{status_msg}
\title{Internal ghclass helpers}
\usage{
read_bin_file(x)

github_api_repo_get_file(repo, path, branch = NULL)

arg_is_chr_scalar(..., allow_null = FALSE, allow_na = FALSE)

arg_is_lgl_scalar(..., allow_null = FALSE, allow_na = FALSE)

arg_is_pos_int_scalar(..., allow_null = FALSE, allow_na = FALSE)

arg_is_chr(..., allow_null = FALSE, allow_na = FALSE, allow_empty = TRUE)

arg_is_lgl(..., allow_null = FALSE, allow_na = FALSE, allow_empty = TRUE)

arg_is_pos_int(..., allow_null = FALSE)

cli_glue(..., .envir = parent.frame())

cli_stop(..., .envir = parent.frame())

cli_warn(..., .envir = parent.frame())

github_api_repo_tree(repo, sha = NULL)

repo_files(repo, branch = NULL)

ghclass_api_v3_req(
  endpoint,
  ...,
  .send_headers = character(),
  version = "2022-11-28",
  limit = github_get_api_limit(),
  max_wait = github_get_max_wait(),
  max_rate = github_get_max_rate()
)

match_repo(repo, index = 1)

get_repo_name(repo)

get_repo_owner(repo)

format_repo(repo, branch = NULL, path = NULL)

result(x)

error(x)

succeeded(x)

failed(x)

status_msg(
  x,
  success = NULL,
  fail = NULL,
  include_error_msg = TRUE,
  .envir = parent.frame()
)
}
\description{
These functions are exported so that companion packages (e.g.
ghclasspeer) can reuse them. They are not intended for end users
and their interfaces may change without notice.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/repo.R, R/repo_add_team.R, R/repo_add_user.R,
%   R/repo_collaborators.R, R/repo_contributors.R, R/repo_remove_team.R,
%   R/repo_remove_user.R
\name{repo_user}
\alias{repo_user}
\alias{repo_add_team}
\alias{repo_team_permission}
\alias{repo_add_user}
\alias{repo_user_permission}
\alias{repo_collaborators}
\alias{repo_contributors}
\alias{repo_remove_team}
\alias{repo_remove_user}
\title{GitHub Repository tools - user functions}
\usage{
repo_add_team(
  repo,
  team,
  permission = c("push", "pull", "admin", "maintain", "triage"),
  team_type = c("name", "slug")
)

repo_team_permission(
  repo,
  team,
  permission = c("push", "pull", "admin", "maintain", "triage"),
  team_type = c("name", "slug")
)

repo_add_user(
  repo,
  user,
  permission = c("push", "pull", "admin", "maintain", "triage")
)

repo_user_permission(
  repo,
  user,
  permission = c("push", "pull", "admin", "maintain", "triage")
)

repo_collaborators(repo, include_admins = TRUE)

repo_contributors(repo)

repo_remove_team(repo, team, team_type = c("name", "slug"))

repo_remove_user(repo, user)
}
\arguments{
\item{repo}{Character. Address of repository in \code{owner/repo} format.}

\item{team}{Character. Slug or name of team to add.}

\item{permission}{Character. Permission to be granted to a user or team for repo, defaults to "push".}

\item{team_type}{Character. Either "slug" if the team names are slugs or "name" if full team names are provided.}

\item{user}{Character. One or more GitHub usernames.}

\item{include_admins}{Logical. If \code{FALSE}, user names of users with Admin rights are not included, defaults to \code{TRUE}.}
}
\value{
\code{repo_collaborators()} and \code{repo_contributors()} return a tibble.

All other functions invisibly return a list containing the results of the relevant GitHub API calls.
}
\description{
\itemize{
\item \code{repo_add_user()} - Add a user to a repository
\item \code{repo_remove_user()} - Remove a user from a repository
\item \code{repo_add_team()} - Add a team to a repository
\item \code{repo_remove_team()} - Remove a team from a repository
\item \code{repo_user_permission()} - Change a collaborator's permissions for a repository
\item \code{repo_team_permission()} - Change a team's permissions for a repository
\item \code{repo_collaborators()} - Returns a data frame of repos, their collaborators, and their permissions.
\item \code{repo_contributors()} - Returns a data frame containing details on repository contributor(s).
}
}
\details{
Permissions can be set to any of the following:
\itemize{
\item \code{"pull"} - can pull, but not push to or administer this repository.
\item \code{"push"} - can pull and push, but not administer this repository.
\item \code{"admin"} - can pull, push and administer this repository.
\item \code{"maintain"} - Recommended for project managers who need to manage the repository without access to sensitive or destructive actions.
\item \code{"triage"} - Recommended for contributors who need to proactively manage issues and pull requests without write access.
}

\code{repo_contributors()} is built on GitHub's \verb{/stats/contributors} endpoint,
which groups commits by GitHub user. Commits whose author email is not
linked to a GitHub account are not included in the result.
}
\examples{
\dontrun{
repo = repo_create("ghclass-test", "hw1")

team_create("ghclass-test", "team_awesome")

repo_add_user(repo, "rundel")

repo_add_team(repo, "team_awesome")

repo_remove_team(repo, "team_awesome")

repo_collaborators(repo)

repo_contributors(repo)
repo_contributors("rundel/ghclass")

# Cleanup
repo_delete(repo, prompt=FALSE)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/repo.R, R/repo_add_file.R,
%   R/repo_delete_file.R, R/repo_get_file.R, R/repo_get_readme.R, R/repo_ls.R,
%   R/repo_modify_file.R, R/repo_put_file.R
\name{repo_file}
\alias{repo_file}
\alias{repo_add_file}
\alias{repo_delete_file}
\alias{repo_get_file}
\alias{repo_get_readme}
\alias{repo_ls}
\alias{repo_modify_file}
\alias{repo_put_file}
\title{GitHub Repository tools - file functions}
\usage{
repo_add_file(
  repo,
  file,
  message = NULL,
  repo_folder = NULL,
  branch = NULL,
  preserve_path = FALSE,
  overwrite = FALSE
)

repo_delete_file(repo, path, message = NULL, branch = NULL)

repo_get_file(repo, path, branch = NULL, quiet = FALSE, include_details = TRUE)

repo_get_readme(repo, branch = NULL, include_details = TRUE)

repo_ls(repo, path = ".", branch = NULL, full_path = FALSE)

repo_modify_file(
  repo,
  path,
  pattern,
  content,
  method = c("replace", "before", "after"),
  all = FALSE,
  message = "Modified content",
  branch = NULL
)

repo_put_file(
  repo,
  path,
  content,
  message = NULL,
  branch = NULL,
  verbose = TRUE
)
}
\arguments{
\item{repo}{Character. Address of repository in \code{owner/name} format.}

\item{file}{Character. Local file path(s) of file or files to be added.}

\item{message}{Character. Commit message.}

\item{repo_folder}{Character. Name of folder on repository to save the file(s) to. If the folder does not exist on the repository, it will be created.}

\item{branch}{Character. Name of branch to use.}

\item{preserve_path}{Logical. Should the local relative path be preserved. Default \code{FALSE}.}

\item{overwrite}{Logical. Should existing file or files with same name be overwritten. Default \code{FALSE}.}

\item{path}{Character. File's path within the repository.}

\item{quiet}{Logical. Should status messages be printed. Default \code{FALSE}.}

\item{include_details}{Logical. Should file details be attached as attributes. Default \code{TRUE}.}

\item{full_path}{Logical. Should the function return the full path of the files and directories. Default \code{FALSE}.}

\item{pattern}{Character. Regex pattern.}

\item{content}{Character or raw. Content of the file.}

\item{method}{Character. Should the content \code{replace} the matched pattern or be inserted \code{before} or \code{after} the match.}

\item{all}{Character. Should all instances of the pattern be modified (\code{TRUE}) or just the first (\code{FALSE}).}

\item{verbose}{Logical. Should success / failure messages be printed. Default \code{TRUE}.}
}
\value{
\code{repo_delete_file()}, \code{repo_modify_file()}, and \code{repo_put_file()} all invisibly
return a list containing the results of the relevant GitHub API calls.

\code{repo_ls()} returns a character vector of repo files in the given path.

\code{repo_get_file()} and \code{repo_get_readme()} return a character vector with API results
attached as attributes if \code{include_details = TRUE}
}
\description{
\itemize{
\item \code{repo_add_file()} - Add / update files in a GitHub repository.
Note that due to delays in caching, files that have been added
very recently might not yet be displayed as existing and might accidentally
be overwritten.
\item \code{repo_delete_file()} - Delete a file from a GitHub repository
\item \code{repo_modify_file()} - Modify an existing file within a GitHub repository.
\item \code{repo_ls()} - Low level function for listing the files in a GitHub Repository
\item \code{repo_put_file()} - Low level function for adding a file to a GitHub repository
\item \code{repo_get_file()} - Low level function for retrieving the content of a file from a GitHub Repository
\item \code{repo_get_readme()} - Low level function for retrieving the content of the \code{README.md} of a GitHub Repository
}
}
\examples{
\dontrun{
repo = repo_create("ghclass-test", "repo_file_test", auto_init=TRUE)

repo_ls(repo, path = ".")

repo_get_readme(repo, include_details = FALSE)

repo_get_file(repo, ".gitignore", include_details = FALSE)

repo_modify_file(
  repo, path = "README.md", pattern = "repo_file_test",
  content = "\n\nHello world!\n", method = "after"
)

repo_get_readme(repo, include_details = FALSE)

repo_add_file(repo, file = system.file("DESCRIPTION", package="ghclass"))

repo_get_file(repo, "DESCRIPTION", include_details = FALSE)

repo_delete_file(repo, "DESCRIPTION")

repo_delete(repo, prompt=FALSE)
}

}
